      program corr_bry_flux

! Reads free-surface and normal velocity components from boundary
! forcing file, computes net volume flux into the model domain, and
! adjusts normal velocities to make sure that net flux integrates to
! zero.  Should be used as
!
!        corr_bry_flux grid_file.nc bry_file.nc
! or
!        corr_bry_flux -n grid_file.nc bry_file.nc
! or
!        corr_bry_flux --nomodify grid_file.nc bry_file.nc
!
! where -n [shorthand for --nomodify] makes it be a purely diagnostic
! tools, i.e., it just prints the total integrated flux coming through
! the open boundaries into the domain normalized by the net cross-
! section area (hence in units of velocity, m/sec) without modifying
! the boundary file.

! Method of correction: (i) separately compute the sum of all fluxes
! incoming into the domain and all the sum of all outgoing fluxes; 
! (ii) compute the correction ratio
!
!                           flx_in-flx_out
!                  alpha = ----------------
!                           flx_in+flx_out 
!
! after which (iii) multiply all incoming velocities by (1-alpha) and
! all outgoing by (1+alpha).  Once this is complete, flx_in and flx_out
! computed from modified velocities add up to zero.  Note that this is
! different from applying a constant-value correction to all the
! velocities: adjusting it this way makes no velocity will ever change
! sign; maximum adjustments are where velocities are at maximum (and
! where the errors are potentially the largest);

      implicit none
      integer nargs, Lm,Mm,N, ncgrd, ncbry, nrecs
      character(len=64) grdfile, bryfile, str
      real(kind=16) barcrss,barflux,crss,flx2,flx3, sum1,sum2,sum3
      real(kind=8) theta_s, theta_b, hc, dA, alpha, ucorr, ubarcorr
      real(kind=8), allocatable, dimension(:) :: Cs_w,Cs_r, dn_u,dm_v,
     &                         hu_west,dn_u_west, hv_south,dm_v_south,
     &                         hu_east,dn_u_east, hv_north,dm_v_north
      real(kind=8), allocatable, dimension(:,:) :: pm,pn, h,
     &                   dhu_west,dhu_east, dhv_south, dhv_north
      real(kind=8), allocatable, dimension(:,:,:) :: z_w,Hz
      real(kind=4), allocatable, dimension(:) ::
     &                zeta_west, ubar_west,    zeta_east,  ubar_east,
     &               zeta_south, vbar_south,   zeta_north, vbar_north
      real(kind=4), allocatable, dimension(:,:) :: u_west, u_east,
     &                                            v_south, v_north
      integer(kind=2), allocatable, dimension(:) :: msku_west,
     &                             msku_east, mskv_south, mskv_north
      integer(kind=2), allocatable, dimension(:,:) :: msk

      logical modify_bry_file, OBC_WEST,OBC_EAST,OBC_SOUTH,OBC_NORTH
      integer ierr, varid, i,j,k,rec, iter, ndims, dimids(4), natts,
     &                             vtype,  lgrd, lbry, lstr, lenstr

#include "netcdf.inc"

      nrecs=0 ;  modify_bry_file=.true.
      lgrd=0  ;  OBC_WEST=.false.  ; OBC_EAST=.false.
      lbry=0  ;  OBC_SOUTH=.false. ; OBC_NORTH=.false.

      nargs=iargc()
      if (nargs == 2 .or. nargs == 3) then
        do i=1,nargs
          call getarg(i,str) ; lstr=lenstr(str)
          if (str(1:1) == '-') then
            if ( (lstr == 10 .and. str(1:lstr) == '--nomodify')
     &      .or. (lstr == 2  .and. str(1:lstr) == '-n') ) then
              modify_bry_file=.false.
            else
              write(*,'(/1x,3A/)') '### ERROR: illegal option ''',
     &                                         str(1:lstr), '''.'
              stop
            endif
          elseif (lgrd == 0) then
            grdfile=str(1:lstr) ; lgrd=lstr
          elseif (lbry == 0) then
            bryfile=str(1:lstr) ; lbry=lstr
          endif
        enddo
      else
        write(*,'(/1x,A/,/8x,A/1x,A/8x,A/1x,A/8x,A/)')
     &          'This command should be used as',
     &          'corr_bry_flux grid_file.nc bry_file.nc',   'or',
     &          'corr_bry_flux -n grid_file.nc bry_file.nc',  'or',
     &          'corr_bry_flux --nomodify grid_file.nc bry_file.nc'
        stop
      endif

      if (modify_bry_file) then
        ierr=nf_open(bryfile, nf_write, ncbry)
      else
        ierr=nf_open(bryfile, nf_nowrite, ncbry)
      endif
      if (ierr == nf_noerr) then
        ierr=nf_inq_varid(ncbry, 'bry_time', varid)
        if (ierr == nf_noerr) then
          ierr=nf_inq_var(ncbry, varid, str, vtype, ndims,
     &                                      dimids, natts)
          if (ierr == nf_noerr) then
            lstr=lenstr(str)
            if (ndims == 1) then
              ierr=nf_inq_dim(ncbry,  dimids(ndims), str, nrecs)
              if (ierr == nf_noerr) then
                write(*,'(/1x,A,I5,1x,4A)') 'Found', nrecs, 'time ',
     &                      'records in ''', bryfile(1:lbry), '''.'
              else
                write(*,'(/1x,2A,I3,1x,3A/12x,A/)')    '### ERROR: ',
     &              'Cannot find size of dimension #', dimids(ndims),
     &              'in ''',bryfile(1:lbry),'''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,3A,I3,A/)')     '### ERROR: variable ''',
     &             str(1:lstr), ''' has wrong number of dimensions,',
     &                         ndims, ', must be just a single one.'
            endif
         else
            write(*,'(/1x,2A,I3,1x,3A/12x,A/)') '### ERROR: Cannot ',
     &         'make general inquiry for variable #', varid, 'in ''',
     &                     bryfile(1:lbry), '''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot find time ',
     &                            'variable in ''', bryfile(1:lbry),
     &                                     '''.', nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) stop

        call roms_find_dims(ncbry, bryfile,  Lm,Mm,N)
        call get_sclr_by_name_double(ncbry, 'theta_s', theta_s)
        call get_sclr_by_name_double(ncbry, 'theta_b', theta_b)
        call get_sclr_by_name_double(ncbry,      'hc',      hc)

        write(*,'(2(2x,A,F12.8),2x,A,F12.6)')
     &          'From BRY file: theta_s =', theta_s,
     &          'theta_b =',  theta_b,   'hc =', hc

! Check which side boundaries are present in BRY file and allocate
! only arrays which are actually needed. Note that these arrays are
! allocated first to avoid memory fragmentation later.

        ierr=nf_inq_varid(ncbry, 'zeta_west', varid)
        if (ierr == nf_noerr) then
          write(*,*) 'found WESTERN boundary'
          OBC_WEST=.true.
          allocate( hu_west(0:Mm+1), dn_u_west(0:Mm+1),
     &            msku_west(0:Mm+1), zeta_west(0:Mm+1),
     &            ubar_west(0:Mm+1),  u_west(0:Mm+1,N),
     &                             dhu_west(0:Mm+1,N))
        endif
        ierr=nf_inq_varid(ncbry, 'zeta_east', varid)
        if (ierr == nf_noerr) then
          write(*,*) 'found EASTERN boundary'
          OBC_EAST=.true.
          allocate( hu_east(0:Mm+1), dn_u_east(0:Mm+1),
     &            msku_east(0:Mm+1), zeta_east(0:Mm+1),
     &            ubar_east(0:Mm+1),  u_east(0:Mm+1,N),
     &                             dhu_east(0:Mm+1,N))
        endif
        ierr=nf_inq_varid(ncbry, 'zeta_south', varid)
        if (ierr == nf_noerr) then
          write(*,*) 'found SOUTHERN boundary'
          OBC_SOUTH=.true.
          allocate( hv_south(0:Lm+1), dm_v_south(0:Lm+1),
     &            mskv_south(0:Lm+1), zeta_south(0:Lm+1),
     &            vbar_south(0:Lm+1),  v_south(0:Lm+1,N),
     &                              dhv_south(0:Lm+1,N))
        endif
        ierr=nf_inq_varid(ncbry, 'zeta_north', varid)
        if (ierr == nf_noerr) then
          write(*,*) 'found NORTHERN boundary'
          OBC_NORTH=.true.
          allocate( hv_north(0:Lm+1), dm_v_north(0:Lm+1),
     &            mskv_north(0:Lm+1), zeta_north(0:Lm+1),
     &            vbar_north(0:Lm+1),  v_north(0:Lm+1,N),
     &                              dhv_north(0:Lm+1,N))
        endif

! Open grid file, check dimensions against what is already found
! above, then allocate all the grid-related arrays and read them;

        ierr=nf_open(grdfile, nf_nowrite, ncgrd)
        if (ierr == nf_noerr) then
          call roms_check_dims(ncgrd, grdfile, Lm,Mm,0)

          allocate(Cs_w(0:N), Cs_r(N), pm(0:Lm+1,0:Mm+1),
     &              pn(0:Lm+1,0:Mm+1),  h(0:Lm+1,0:Mm+1),
     &           msk(0:Lm+1,0:Mm+1), Hz(0:Lm+1,0:Mm+1,N),
     &                           z_w(0:Lm+1,0:Mm+1,0:N))

          call get_var_by_name_double(ncgrd, 'pm', pm)
          call get_var_by_name_double(ncgrd, 'pn', pn)
          ierr=nf_inq_varid(ncgrd, 'mask_rho', varid)
          if (ierr == nf_noerr) then
            ierr=nf_get_var_double(ncgrd, varid, h)
            if (ierr == nf_noerr) then
              do j=0,Mm+1
                do i=0,Lm+1
                  if (h(i,j) > 0.5D0) then
                    msk(i,j)=1
                  else
                    msk(i,j)=0
                  endif
                enddo
              enddo
              write(*,'(9x,2A)')  'read ''mask_rho'', converted to ',
     &                                             'integer(kind=2)'
            else
              write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot read ',
     &              'variable ''mask_rho'' from ''', grdfile(1:lgrd),
     &                                      '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(9x,3A)') 'no mask found in ''',grdfile(1:lgrd),
     &                              ''', assuming water everywhere.'
            msk=1
          endif
          call get_var_by_name_double(ncgrd, 'h',  h)
          ierr=nf_close(ncgrd)
        else
          write(*,'(/1x,4A/)')   '### ERROR: Cannot open grid file ''',
     &                     grdfile(1:lgrd), '''. ',  nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/)') '### ERROR: Cannot open boundary file ''',
     &                     bryfile(1:lbry), '''. ',  nf_strerror(ierr)
      endif
      if (ierr /= nf_noerr) stop

      write(*,*) 'creating 3D coordinate system...'
      call set_scoord(theta_s,theta_b, N, Cs_r,Cs_w)
      call set_depth(Lm,Mm,N, Cs_r,Cs_w,hc, 'SM09', h, Hz,z_w)
      write(*,*) '..................'
      sum2=0.D0
      do j=0,Mm+1                       ! just a sanity check:
        do i=0,Lm+1                     ! make sure that all Hz
          sum1=0.D0                     ! add up to total depth
          do k=N,1,-1
            Hz(i,j,k)=z_w(i,j,k)-z_w(i,j,k-1)
            sum1=sum1+Hz(i,j,k)
          enddo
          sum2=max(sum2, abs(sum1-h(i,j)))
        enddo
      enddo
      write(*,*) 'MAX ERROR =', sum2

! Initialize horizontal cross-sections and layer thickness along
! the open boundaries (they are needed to compute fluxes); compute
! unperturbed (zeta=0) cross-section of the open boundary just for
! checking: it should be the same as as reported by ROMS code. 

      barcrss=0.D0
#define i illegal

      if (OBC_WEST) then
        hu_west=0.D0 ; dn_u_west=0.D0 ; msku_west=0
        zeta_west=0. ; ubar_west=0.   ; u_west=0.
        do j=1,Mm
          if (msk(1,j) == 1 .and. msk(0,j) == 1) then
            msku_west(j)=1
            hu_west(j)=0.5D0*(h(1,j)+h(0,j))
            dn_u_west(j)=2.D0/(pn(1,j)+pn(0,j))
            barcrss=barcrss + hu_west(j)*dn_u_west(j)
          endif
        enddo
        do k=1,N
          do j=0,Mm+1
            dhu_west(j,k)=0.5D0*(Hz(1,j,k)+Hz(0,j,k))
          enddo
        enddo
      endif

      if (OBC_EAST) then
        hu_east=0.D0 ; dn_u_east=0.D0 ; msku_east=0
        zeta_east=0. ; ubar_east=0.   ; u_east=0.
        do j=1,Mm
          if (msk(Lm,j) == 1 .and. msk(Lm+1,j) == 1) then
            msku_east(j)=1
            hu_east(j)=0.5D0*(h(Lm,j)+h(Lm+1,j))
            dn_u_east(j)=2.D0/(pn(Lm,j)+pn(Lm+1,j))
            barcrss=barcrss + hu_east(j)*dn_u_east(j)
          endif
        enddo
        do k=1,N
          do j=0,Mm+1
            dhu_east(j,k)=0.5D0*(Hz(Lm,j,k)+Hz(Lm+1,j,k))
          enddo
        enddo
      endif

#undef i
#define j illegal

      if (OBC_SOUTH) then
        hv_south=0.D0 ; dm_v_south=0.D0 ; mskv_south=0
        zeta_south=0. ; vbar_south=0.   ; v_south=0.
        do i=1,Lm
          if (msk(i,0) == 1 .and. msk(i,1) == 1) then
            mskv_south(i)=1
            hv_south(i)=0.5D0*(h(i,0)+h(i,1))
            dm_v_south(i)=2.D0/(pm(i,0)+pm(i,1))
            barcrss=barcrss + hv_south(i)*dm_v_south(i)
          endif
        enddo
        do k=1,N
          do i=0,Lm+1
            dhv_south(i,k)=0.5D0*(Hz(i,0,k)+Hz(i,1,k))
          enddo
        enddo
      endif

      if (OBC_NORTH) then
        hv_north=0.D0 ; dm_v_north=0.D0 ; mskv_north=0
        zeta_north=0. ; vbar_north=0.   ; v_north=0.
        do i=1,Lm
          if (msk(i,Mm) == 1 .and. msk(i,Mm+1) == 1) then
            mskv_north(i)=1
            hv_north(i)=0.5D0*(h(i,Mm)+h(i,Mm+1))
            dm_v_north(i)=2.D0/(pm(i,Mm)+pm(i,Mm+1))
            barcrss=barcrss + hv_north(i)*dm_v_north(i)
          endif
        enddo
        do k=1,N
          do i=0,Lm+1
            dhv_north(i,k)=0.5D0*(Hz(i,Mm,k)+Hz(i,Mm+1,k))
          enddo
        enddo
      endif
#undef j

      write(*,*) 'unpeturbed barcrss =', barcrss
      deallocate(Hz,z_w, msk, h, pn, pm, Cs_r, Cs_w)
      allocate(dn_u(0:Mm+1), dm_v(0:Lm+1))
      write(*,'(/1x,A)') 'initialization complete'




      do rec=1,nrecs
        if (OBC_WEST) then
          call get_rec_by_name_real(ncbry, bryfile, 'zeta_west',
     &                                0,Mm+2,0,rec,  zeta_west)
          call get_rec_by_name_real(ncbry, bryfile, 'ubar_west',
     &                                0,Mm+2,0,rec,  ubar_west)
          call get_rec_by_name_real(ncbry, bryfile, 'u_west',
     &                                0,Mm+2,N,rec,  u_west)
        endif
        if (OBC_SOUTH) then
          call get_rec_by_name_real(ncbry, bryfile, 'zeta_south',
     &                                Lm+2,0,0,rec,  zeta_south)
          call get_rec_by_name_real(ncbry, bryfile, 'vbar_south',
     &                                Lm+2,0,0,rec,  vbar_south)
          call get_rec_by_name_real(ncbry, bryfile, 'v_south',
     &                                Lm+2,0,N,rec,  v_south)
        endif
        if (OBC_EAST) then
          call get_rec_by_name_real(ncbry, bryfile, 'zeta_east',
     &                                0,Mm+2,0,rec,  zeta_east)
          call get_rec_by_name_real(ncbry, bryfile, 'ubar_east',
     &                                0,Mm+2,0,rec,  ubar_east)
          call get_rec_by_name_real(ncbry, bryfile, 'u_east',
     &                                0,Mm+2,N,rec,  u_east)
        endif
        if (OBC_NORTH) then
          call get_rec_by_name_real(ncbry, bryfile, 'zeta_north',
     &                                Lm+2,0,0,rec,  zeta_north)
          call get_rec_by_name_real(ncbry, bryfile, 'vbar_north',
     &                                Lm+2,0,0,rec,  vbar_north)
          call get_rec_by_name_real(ncbry, bryfile, 'v_north',
     &                                Lm+2,0,N,rec,  v_north)
        endif

        do iter=1,3

! Compute total integrated finite-volume flux going into the model
! domain through open boundaries. This can be done in two ways: from
! barotropic variables as well as from full 3D velocities.  Normally
! the two should agree, but...  compute both ways and compare.
! In the code below dn_u(j) and dm_v(i) are the horizontal width of
! velocity facets masked and scaled by the factor of (1+zeta/h), so
! its multiplication by unperturbed depth "h" or unperturbed "Hz"
! yields their free-surface perturbed versions.  To mitigate roundoff
! errors 3D summation is done in 2 stages: first vertically in each
! vertical column; then horizontally.

                                   ! Note that dn_u(j) computed just
          barcrss=0.D0 ; crss=0.D0 ! below and dm_v(i) further down are
          barflux=0.D0 ; flx2=0.D0 ! actually HALF of what they should
                         flx3=0.D0 ! be by their mathematical meaning:
                                   ! this is to cancel the factor of
          if (OBC_WEST) then       ! TWO when computing (Hz+Hz)*dn_u.
#define i illegal
            do j=1,Mm
              if (msku_west(j) == 1) then
                dn_u(j)=dn_u_west(j)*(1.D0+zeta_west(j)/hu_west(j))
                dA=hu_west(j)*dn_u(j)
                barcrss=barcrss+dA ; barflux=barflux+dA*ubar_west(j)
              endif
            enddo
            do j=1,Mm
              if (msku_west(j) == 1) then
                sum1=0.D0 ; sum2=0.D0 ; sum3=0.D0
                do k=N,1,-1
                  dA=dhu_west(j,k)*dn_u(j)
                  sum1=sum1+dA
                  if (u_west(j,k) > 0.) then
                    sum2=sum2 +dA*u_west(j,k)
                  else
                    sum3=sum3 -dA*u_west(j,k)
                  endif
                enddo
                crss=crss+sum1 ; flx2=flx2+sum2 ; flx3=flx3+sum3
              endif
            enddo
          endif

          if (OBC_EAST) then
            do j=1,Mm
              if (msku_east(j) == 1) then
                dn_u(j)=dn_u_east(j)*(1.D0+zeta_east(j)/hu_east(j))
                dA=hu_east(j)*dn_u(j)
                barcrss=barcrss+dA ; barflux=barflux-dA*ubar_east(j)
              endif
            enddo
            do j=1,Mm
              if (msku_east(j) == 1) then
                sum1=0.D0 ; sum2=0.D0 ; sum3=0.D0
                do k=N,1,-1
                  dA=dhu_east(j,k)*dn_u(j)
                  sum1=sum1+dA
                  if (u_east(j,k) < 0.) then
                    sum2=sum2 -dA*u_east(j,k)
                  else
                    sum3=sum3 +dA*u_east(j,k)
                  endif
                enddo
                crss=crss+sum1 ; flx2=flx2+sum2 ; flx3=flx3+sum3
              endif
            enddo
          endif
#undef i
#define j illegal

          if (OBC_SOUTH) then
            do i=1,Lm
              if (mskv_south(i) == 1) then
                dm_v(i)=dm_v_south(i)*(1.D0+zeta_south(i)/hv_south(i))
                dA=hv_south(i)*dm_v(i)
                barcrss=barcrss+dA ;  barflux=barflux+dA*vbar_south(i)
              endif
            enddo
            do i=1,Lm
              if (mskv_south(i) == 1) then
                sum1=0.D0 ; sum2=0.D0 ; sum3=0.D0
                do k=N,1,-1
                  dA=dhv_south(i,k)*dm_v(i)
                  sum1=sum1+dA
                  if (v_south(i,k) > 0.) then
                    sum2=sum2 +dA*v_south(i,k)
                  else
                    sum3=sum3 -dA*v_south(i,k)
                  endif
                enddo
                crss=crss+sum1 ; flx2=flx2+sum2 ; flx3=flx3+sum3
              endif
            enddo
          endif

          if (OBC_NORTH) then
            do i=1,Lm
              if (mskv_north(i) == 1) then
                dm_v(i)=dm_v_south(i)*(1.D0+zeta_north(i)/hv_north(i))
                dA=hv_north(i)*dm_v(i)
                barcrss=barcrss+dA ;  barflux=barflux-dA*vbar_north(i)
              endif
            enddo
            do i=1,Lm
              if (mskv_north(i) == 1) then
                sum1=0.D0 ; sum2=0.D0 ; sum3=0.D0
                do k=N,1,-1
                  dA=dhv_north(i,k)*dm_v(i)
                  sum1=sum1+dA
                  if (v_north(i,k) < 0.) then
                    sum2=sum2 -dA*v_north(i,k)
                  else
                    sum3=sum3 +dA*v_north(i,k)
                  endif
                enddo
                crss=crss+sum1 ; flx2=flx2+sum2 ; flx3=flx3+sum3
              endif
            enddo
          endif
#undef j

          alpha=(flx2-flx3)/(flx2+flx3)
          ubarcorr=barflux/barcrss ;  ucorr=(flx2-flx3)/crss

          write(*,'(1x,A,3(ES22.14))') 'BAR',barcrss,barflux,ubarcorr
          write(*,'(1x,A,4(ES22.14)/)') 'E3D', crss, flx2+flx3, ucorr,
     &                                                          alpha

#define i illegal
          if (OBC_WEST) then
            do j=0,Mm+1
              ubar_west(j)=ubar_west(j)-ubarcorr
            enddo
            do k=1,N
              do j=0,Mm+1
c               u_west(j,k)=u_west(j,k)-ucorr
                if (u_west(j,k) > 0.) then
                  u_west(j,k)=u_west(j,k)*(1.D0-alpha)
                else
                  u_west(j,k)=u_west(j,k)*(1.D0+alpha)
                endif
              enddo
            enddo
          endif

          if (OBC_EAST) then
            do j=0,Mm+1
              ubar_east(j)=ubar_east(j)+ubarcorr
            enddo
            do k=1,N
              do j=0,Mm+1
c                u_east(j,k)=u_east(j,k)+ucorr
                if (u_east(j,k) < 0.) then
                  u_east(j,k)=u_east(j,k)*(1.D0-alpha)
                else
                  u_east(j,k)=u_east(j,k)*(1.D0+alpha)
                endif
              enddo
            enddo
          endif
#undef i
#define j illegal

          if (OBC_SOUTH) then
            do i=0,Lm+1
              vbar_south(i)=vbar_south(i)-ubarcorr
            enddo
            do k=1,N
              do i=0,Lm+1
c               v_south(i,k)=v_south(i,k)-ucorr
                if (v_south(i,k) > 0.) then
                  v_south(i,k)=v_south(i,k)*(1.D0-alpha)
                else
                  v_south(i,k)=v_south(i,k)*(1.D0+alpha)
                endif
              enddo
            enddo
          endif

          if (OBC_NORTH) then
            do i=0,Lm+1
              vbar_north(i)=vbar_north(i)+ubarcorr
            enddo
            do k=1,N
              do i=1,Lm
c               v_north(i,k)=v_north(i,k)+ucorr
                if (v_north(i,k) < 0.) then
                  v_north(i,k)=v_north(i,k)*(1.D0-alpha)
                else
                  v_north(i,k)=v_north(i,k)*(1.D0+alpha)
                endif
              enddo
            enddo
          endif
#undef j


! Force vertical integral of 3D velocity component be the same
! as barotropic

#define FORCE_2D_TO_3D
#define i illegal

          if (OBC_WEST) then
            do j=0,Mm+1
              sum1=0.D0 ; sum2=0.D0
              do k=N,1,-1
                sum1=sum1+dhu_west(j,k)
                sum2=sum2+dhu_west(j,k)*u_west(j,k)
              enddo
#ifdef FORCE_2D_TO_3D
              ubar_west(j)=sum2/sum1
#else
              dA=ubar_west(j)-sum2/sum1
              do k=1,N
                u_west(j,k)=u_west(j,k)+dA
              enddo
#endif
            enddo
          endif

          if (OBC_EAST) then
            do j=0,Mm+1
              sum1=0.D0 ; sum2=0.D0
              do k=N,1,-1
                sum1=sum1+dhu_east(j,k)
                sum2=sum2+dhu_east(j,k)*u_east(j,k)
              enddo
#ifdef FORCE_2D_TO_3D
              ubar_east(j)=sum2/sum1
#else
              dA=ubar_east(j)-sum2/sum1
              do k=1,N
                u_east(j,k)=u_east(j,k)+dA
              enddo
#endif
            enddo
          endif
#undef i
#define j illegal

          if (OBC_SOUTH) then
            do i=0,Lm+1
              sum1=0.D0 ; sum2=0.D0
              do k=N,1,-1
                sum1=sum1+dhv_south(i,k)
                sum2=sum2+dhv_south(i,k)*v_south(i,k)
              enddo
#ifdef FORCE_2D_TO_3D
              vbar_south(i)=sum2/sum1
#else
              dA=vbar_south(i)-sum2/sum1
              do k=1,N
               v_south(i,k)=v_south(i,k)+dA
              enddo
#endif
            enddo
          endif

          if (OBC_NORTH) then
            do i=0,Lm+1
              sum1=0.D0 ; sum2=0.D0
              do k=N,1,-1
                sum1=sum1+dhv_north(i,k)
                sum2=sum2+dhv_north(i,k)*v_north(i,k)
              enddo
#ifdef FORCE_2D_TO_3D
              vbar_north(i)=sum2/sum1
#else
              dA=vbar_north(i)-sum2/sum1
              do k=1,N
                v_north(i,k)=v_north(i,k)+dA
              enddo
#endif
            enddo
          endif
#undef j
        enddo !<-- iter

! Write back...

        if (modify_bry_file) then
          if (OBC_WEST) then
            call put_rec_by_name_real(ncbry, bryfile, 'ubar_west',
     &                                  0,Mm+2,0,rec,  ubar_west)
            call put_rec_by_name_real(ncbry, bryfile, 'u_west',
     &                                  0,Mm+2,N,rec,  u_west)
          endif
          if (OBC_SOUTH) then
            call put_rec_by_name_real(ncbry, bryfile, 'vbar_south',
     &                                  Lm+2,0,0,rec,  vbar_south)
            call put_rec_by_name_real(ncbry, bryfile, 'v_south',
     &                                  Lm+2,0,N,rec,  v_south)
          endif
          if (OBC_EAST) then
            call put_rec_by_name_real(ncbry, bryfile, 'ubar_east',
     &                                  0,Mm+2,0,rec,  ubar_east)
            call put_rec_by_name_real(ncbry, bryfile, 'u_east',
     &                                  0,Mm+2,N,rec,  u_east)
          endif
          if (OBC_NORTH) then
            call put_rec_by_name_real(ncbry, bryfile, 'vbar_north',
     &                                  Lm+2,0,0,rec,  vbar_north)
            call put_rec_by_name_real(ncbry, bryfile, 'v_north',
     &                                  Lm+2,0,N,rec,  v_north)
          endif
        endif
      enddo !<-- rec
      ierr=nf_close(ncbry)
      end
