      program courant

! Compute advective Courant numbers, horizontal and vertical, and write
! them into a special netCDF file.  Both numbers are defined in finite-
! volume sense as the fraction of water inside control volume around
! density point replaced during one time step. In essence this is sum
! of outgoing horizontal or vertical fluxes multiplied by time step and
! divided by the control volume.  The numbers are computed as 3D, then
! the maxima over each vertical column are takes and saved as 2D fields
! producing horizontal maps.

! CPP switch NO_OMEGA makes this code to compute S-coordinate "omega"
! vertical velocity instead of reading it from the file (hence needing
! only "u" and "v").

! CPP switch PUT_CW3D causes saving vertical Courant number as 3D field
! into the file in addition to just saving the maximum over vertical
! columns.

! CPP switch HOT_SPOTS cause program to take the maximum of "c2x" and
! "c2w" over all time records.  It should be normally activated in
! Makefile rather than here.

#define NO_OMEGA
#define MAX_VELOCITY
c--#define PUT_CW3D
c--#define VERBOSE

c--#define HOT_SPOTS

      implicit none
      integer iarg,nargs, Lm,Mm,N, ncid, nctarg, max_recs,irec, ierr,
     &        vid_u,vid_v,vid_t_in, vtype, ndims,natts, file_to_start,
     &        dimids(8), r2dgrd(4), r3dgrd(4), i,j, lvar, lsrc, ltrg
#ifndef HOT_SPOTS
     &      , vid_t_out, unlm_id_out, irec_out
#endif
#ifndef NO_OMEGA
     &      , vid_w
#endif
      integer  iargc, lenstr     !<-- these are functions

      real(kind=8) time, dt, hc
      real(kind=8), allocatable, dimension(:) :: Cs_r, Cs_w
      real(kind=8), allocatable, dimension(:,:) :: h, pm,pn,rmask,
     &                                             dn_u,dm_v,iA_r
      real(kind=8), allocatable, dimension(:,:,:) :: Hz,z_w
      real(kind=4), allocatable, dimension(:,:) :: c2x,c2w
#ifdef MAX_VELOCITY
     &                                           , uv_amp
#endif
      real(kind=4), allocatable, dimension(:,:,:) :: u,v,W, cw
      real(kind=4) FillValue
      logical found_FillValue

      character(len=128) fname, trg_fname
      character(len=16) VertCoordType, vname
      character(len=128), allocatable, dimension(:) :: args
      character(len=10), parameter :: tname='ocean_time'
      character(len=1), parameter :: u_name='u', v_name='v'
#ifndef NO_OMEGA
      character(len=5), parameter :: w_name='omega'
#endif
#include "netcdf.inc"
                                       ! Check the number of arguments
      nargs=iargc()                    ! given in command line and save
      if (nargs > 0) then              ! the actual arguments (names of
        allocate(args(nargs))          ! netCDF files) into array.
        do iarg=1,nargs
          call getarg(iarg,args(iarg))
          lsrc=lenstr(args(iarg))
        enddo
      else
        write(*,'(/1x,A/ /12x,A/)') 'Usage should be: ',
#ifdef HOT_SPOTS
     &       'hotspots grid_file.nc file1.nc [file2.nc ...]'
#else
     &       'courant grid_file.nc file1.nc [file2.nc ...]'
#endif
        stop
      endif                            ! Reset everything:
                                       !------ --------------
      FillValue=0.                     ! horizontal and vertical
      found_FillValue=.false.          ! dimensions, special value
      Lm=0 ; Mm=0 ; N=0                ! for input velocities, etc

#ifdef HOT_SPOTS
      trg_fname='hotspots.nc'
#else
      trg_fname='courant.nc'
      irec_out=0                       ! record counter for output.
#endif

! Open the first file in the sequence, interpret it as grid file,
! determine horizontal dimensions (vertical may or may not be found at
! this stage), allocate 2D arrays arrays, read geometric data, finally
! check whether the file contains variables u,v,w: if it does, include
! it into the sequence, if not start the sequence from the second file
! (the first one is purely grid data).

      fname=args(1) ; lsrc=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        write(*,*) 'Processing ''', fname(1:lsrc), ''' as grid file.'
        call roms_find_dims(ncid, fname, Lm,Mm,N)

        allocate(h(0:Lm+1,0:Mm+1))  ; allocate(rmask(0:Lm+1,0:Mm+1))
        allocate(pm(0:Lm+1,0:Mm+1)) ;  allocate(dn_u(1:Lm+1,0:Mm+1))
        allocate(pn(0:Lm+1,0:Mm+1)) ;  allocate(dm_v(0:Lm+1,1:Mm+1))
                                       allocate(iA_r(0:Lm+1,0:Mm+1))

        call get_var_by_name_double(ncid, 'h',  h)
        call get_var_by_name_double(ncid, 'pm', pm)
        call get_var_by_name_double(ncid, 'pn', pn)

        ierr=nf_inq_varid(ncid, 'mask_rho', i)
        if (ierr == nf_noerr) then
          ierr=nf_get_var_double(ncid, i, rmask)
          if (ierr == nf_noerr) then
            write(*,'(8x,A)') 'retrieved ''rmask'''
          endif
        else
          rmask=1.D0
          write(*,'(/1x,4A/14x,2A/)')    '### WARNING: Cannot find ',
     &            'land mask from file ''',   fname(1:lsrc),   '''.',
     &            'Assuming rmask=1 everywhere. ', nf_strerror(ierr)
          ierr=nf_noerr
        endif
        file_to_start=min(2,nargs)
        ierr=nf_inq_varid (ncid, u_name, vid_u)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varid (ncid, v_name, vid_v)
          if (ierr == nf_noerr) then
#ifndef NO_OMEGA
            ierr=nf_inq_varid (ncid, w_name, vid_w)
            if (ierr == nf_noerr) then
              file_to_start=1
            endif
#else
            file_to_start=1
#endif
          endif
        endif
        ierr=nf_close (ncid)
      else
        write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open netCDF ',
     &   'file ''',fname(1:lsrc),''' for reading.', nf_strerror(ierr)
      endif

! Examine the starting file of the sequence: verify that all necessary
! variables are available.

      fname=args(file_to_start) ; lsrc=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        write(*,*) 'Processing ''', fname(1:lsrc), '''...'
        if (N == 0) then
          call roms_find_dims(ncid, fname, i,j,N)
        endif
        call roms_check_dims(ncid, fname, Lm,Mm,N)
        ierr=nf_inq_varid(ncid, u_name, vid_u)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varid(ncid, v_name, vid_v)
          if (ierr == nf_noerr) then
#ifndef NO_OMEGA
            ierr=nf_inq_varid(ncid, w_name, vid_w)
            if (ierr == nf_noerr) then
#endif
              ierr=nf_get_att_real(ncid, vid_u, '_FillValue',
     &                                            FillValue)
              if (ierr == nf_noerr) then
                write(*,*) 'retrieved _FillValue =', FillValue,
     &                                ' for ''', u_name, '''.'
                found_FillValue=.true.
              else
                write(*,*) 'no _FillValue is specified'
                ierr=nf_noerr
              endif
#ifndef NO_OMEGA
            else
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot find ',
     &         'variable ''', w_name, ''' in ''', fname(1:lsrc), '''.',
     &                                               nf_strerror(ierr)
            endif
#endif
          else
            write(*,'(/1x,6A/12x,A/)')       '### ERROR: Cannot find ',
     &         'variable ''', v_name, ''' in ''', fname(1:lsrc), '''.',
     &                                               nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,6A/12x,A/)')        '### ERROR: Cannot find ',
     &        'variable ''', u_name, ''' in ''', fname(1:lsrc), '''.',
     &                                               nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open  netCDF ',
     &  'file ''', fname(1:lsrc), ''' for reading.', nf_strerror(ierr)
      endif

!      *****    *********    ******   *******    *********
!    ***   ***  *  ***  *   *   ***   ***   ***  *  ***  *
!    ***           ***     **   ***   ***   ***     ***
!      *****       ***    ***   ***   ***   **      ***
!          ***     ***    *********   ******        ***
!    ***   ***     ***    ***   ***   ***  **       ***
!      *****       ***    ***   ***   ***   ***     ***

! Process the entire sequence:
!-------- --- ------ ---------

      do iarg=file_to_start,nargs
        fname=args(iarg) ; lsrc=lenstr(fname)
        ierr=nf_open(fname, nf_nowrite, ncid)
        if (ierr == nf_noerr) then
          write(*,*) 'Processing ''', fname(1:lsrc), ''' iarg=', iarg

          call roms_check_dims(ncid, fname, Lm,Mm,N)
          if (.not.allocated(Cs_r)) then
            allocate(Cs_r(N)) ; allocate(Cs_w(0:N))
            allocate(c2x(0:Lm+1,0:Mm+1))
#ifdef MAX_VELOCITY
            allocate(uv_amp(0:Lm+1,0:Mm+1))
#endif
            allocate(c2w(0:Lm+1,0:Mm+1))
            allocate(cw(0:Lm+1,0:Mm+1,N))
            allocate(u(1:Lm+1,0:Mm+1,N))
            allocate(v(0:Lm+1,1:Mm+1,N))
            allocate(W(0:Lm+1,0:Mm+1,0:N))
            allocate(Hz(0:Lm+1,0:Mm+1,N))
            allocate(z_w(0:Lm+1,0:Mm+1,0:N))

! Create output file.

            ltrg=lenstr(trg_fname)
            ierr=nf_create(trg_fname, nf_64bit_offset, nctarg)
            if (ierr == nf_noerr) then
              ierr=nf_def_dim(nctarg, 'xi_rho',  Lm+2,  r3dgrd(1))
              ierr=nf_def_dim(nctarg, 'eta_rho', Mm+2,  r3dgrd(2))
              ierr=nf_def_dim(nctarg, 's_rho',   N,     r3dgrd(3))
#ifndef HOT_SPOTS
              ierr=nf_def_dim(nctarg, 'ocean_time',  nf_unlimited,
     &                                                  r3dgrd(4))
              ndims=4 ; unlm_id_out=r3dgrd(4)

              if (unlm_id_out > 0) then
                ierr=nf_def_var(nctarg, tname, nf_real, 1,
     &                                     unlm_id_out, vid_t_out)
                ierr=nf_put_att_text(nctarg, vid_t_out, 'units',
     &                                                  5, 'days')
              endif

              r2dgrd=r3dgrd ; r2dgrd(3)=unlm_id_out
#else
              ndims=3 ; r2dgrd=r3dgrd
#endif

              ierr=nf_def_var(nctarg, 'c2x', nf_real,ndims-1,r2dgrd, i)
              ierr=nf_put_att_text(nctarg, i, 'long_name', 52,
     &          'maximum horizontal Courant number in vertical column')
              if (found_FillValue) then
                ierr=nf_put_att_real(nctarg, i, '_FillValue',
     &                                           nf_real, 1, FillValue)
              endif

              ierr=nf_def_var(nctarg, 'c2w', nf_real,ndims-1,r2dgrd, i)
              ierr=nf_put_att_text(nctarg, i, 'long_name', 50,
     &            'maximum vertical Courant number in vertical column')
              if (found_FillValue) then
                ierr=nf_put_att_real(nctarg, i, '_FillValue',
     &                                          nf_real, 1, FillValue)
              endif
#ifdef MAX_VELOCITY
              ierr=nf_def_var(nctarg, 'vel_max', nf_real,ndims-1,
     &                                                 r2dgrd, i)
              ierr=nf_put_att_text(nctarg, i, 'long_name', 46,
     &          'horizontal velocity maximum in vertical column')
              if (found_FillValue) then
                ierr=nf_put_att_real(nctarg, i, '_FillValue',
     &                                     nf_real, 1, FillValue)
              endif
#endif
#ifdef PUT_CW3D
              ierr=nf_def_var(nctarg, 'cw', nf_real, ndims, r3dgrd,i)
              ierr=nf_put_att_text(nctarg, i, 'long_name', 14,
     &                                              'Courant Number')
              if (found_FillValue) then
                ierr=nf_put_att_real(nctarg, i, '_FillValue',
     &                                        nf_real, 1, FillValue)
              endif
#endif
              ierr=nf_enddef(nctarg)
            else
              write(*,'(/1x,4A/12x,A)')  '### ERROR: Cannot create ',
     &                            'netCDF file ''', trg_fname(1:ltrg),
     &                                      '''.', nf_strerror(ierr)
              stop
            endif

            ierr=nf_get_att_double(ncid, nf_global, 'dt', dt)
            if (ierr == nf_noerr) then
              write(*,*) 'Found dt =', dt
              call read_scoord(ncid, N, Cs_r,Cs_w, hc, VertCoordType)

! Note: set_depth computes "z_r" and "z_w". Since "z_r" is not needed
! anywhere later in this code, it is placed into array "Hz", which is
! then overwritten by  init_cu.

              call set_depth(Lm,Mm,N, Cs_r,Cs_w,hc, VertCoordType,
     &                                                  h, Hz,z_w)

C$OMP PARALLEL SHARED(Lm,Mm,N, z_w, Hz,rmask,pm,pn, dn_u,dm_v,iA_r,
C$OMP&                                                    c2x,c2w)

              call init_cu_thread(Lm,Mm,N, z_w, Hz,rmask,pm,pn,
     &                                 dn_u,dm_v,iA_r, c2x,c2w)
C$OMP END PARALLEL

              deallocate(z_w)    !--> no longer needed
            else
              write(*,*) '### ERROR :: Cannot retrieve time step dt.'
              stop
            endif
          endif  !<-- allocated(Cs_r)

! WARNING: The number of records to be processed is determined by
! the length of dimension of the timing variable. However the timing
! variable may not have dimension at all if the file contains is just
! a single snapshot.   If the actual variable in the file is just a
! scalar, an attempt to read it by nf_get_vara(..,start,count,..)
! returns NO ERROR status, retrieving its only scalar value regardless
! of the value of argument "start" (i.e., nf_get_vara behaves like
! nf_get_var in this case).  Similarly, if u,v variables in the file
! do not have time dimension, an attempt to retrieve them by
! nf_get_vara(..., start,count, ...) with start at the index
! corresponding to would be time dimension set to non-existing record
! nevertheless returns NO_ERR status because start,count for indices
! beyond the actual number of dimensions of the variable in the file
! are simply ignored.

! Because of this behavior of netCDF, the previous version of this
! algorithm relying solely on non-zero ERROR status to detect when
! running out of records occurs ends up executing endless loop
! (reading the same data again and again returning NO_ERR, despite
! the fact that record number "irec" is incremented at every iteration
! of the loop) if there is no time dimension in the file.
! The following code segment is designed to positively identify the
! number of records, assigning it to one, in the case of no time
! dimension.

          ierr=nf_inq_varid(ncid, tname, vid_t_in)
          if (ierr == nf_noerr) then
            ierr=nf_inq_var(ncid, vid_t_in, vname, vtype,  ndims,
     &                                             dimids, natts)
            if (ierr == nf_noerr) then
              max_recs=0
              if (ndims == 0) then
                max_recs=1    ! <-- no records, just a single snapshot
              elseif (ndims == 1) then
                ierr=nf_inq_dimlen(ncid, dimids, max_recs)
                if (ierr/=nf_noerr) then
                   lvar=lenstr(tname)
                   write(*,'(/1x,5A/)') '### ERROR: Cannot determine ',
     &                        'number of records for time variable ''',
     &             tname(1:lvar), ''' in ''', fname(1:lsrc), '''.'
                endif
              else
                lvar=lenstr(tname)
                write(*,'(/1x,5A/)')     '### ERROR: Time variable ''',
     &                    tname(1:lvar), ''' in ''', fname(1:lsrc),
     &                               ''' has more than one dimension.'
              endif
            else
              lvar=lenstr(tname)
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot make ',
     &                        'general inquiry about time variable ''',
     &                    tname(1:lvar), ''' in ''', fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
            endif

            do irec=1,max_recs
              ierr=nf_get_vara_double(ncid,vid_t_in, irec,1, time)
              if (ierr == nf_noerr) then
                write(*,*)  tname(1:lenstr(tname)), ' =', time,
     &                                         '  rec =', irec
                call get_rec_by_name_real(ncid, fname, u_name,
     &                                   Lm+1,Mm+2,N, irec, u)
                call get_rec_by_name_real(ncid, fname, v_name,
     &                                   Lm+2,Mm+1,N, irec, v)
#ifndef NO_OMEGA
                call get_rec_by_name_real(ncid, fname, w_name,
     &                                 Lm+2,Mm+2,N+1, irec, W)
#endif

C$OMP PARALLEL SHARED(Lm,Mm,N, Hz,dn_u,dm_v,iA_r, rmask, u,v,W,
C$OMP&                                               c2x,c2w,cw,
#ifdef MAX_VELOCITY
C$OMP&                                                   uv_amp,
#endif
C$OMP&                                           dt, FillValue)
                call comp_cu_thread(Lm,Mm,N, Hz,dn_u,dm_v,iA_r,
     &                                 rmask, u,v,W, c2x,c2w,cw,
#ifdef MAX_VELOCITY
     &                                                   uv_amp,
#endif
     &                                           dt, FillValue)
C$OMP END PARALLEL
#ifndef HOT_SPOTS
                irec_out=irec_out+1
                ierr=nf_put_vara_double(nctarg, vid_t_out,
     &                                      irec_out, 1, time)
                call put_rec_by_name_real(nctarg, trg_fname, 'c2x',
     &                                 Lm+2,Mm+2,1, irec_out, c2x)
                call put_rec_by_name_real(nctarg, trg_fname, 'c2w',
     &                                 Lm+2,Mm+2,1, irec_out, c2w)
#ifdef MAX_VELOCITY
                call put_rec_by_name_real(nctarg, trg_fname,
     &                   'vel_max', Lm+2,Mm+2,1, irec_out, uv_amp)
#endif
# ifdef PUT_CW3D
                call put_rec_by_name_real(nctarg, trg_fname, 'cw',
     &                             Lm+2,Mm+2,N, irec_out, cw)
# endif
                ierr=nf_sync(nctarg)
#endif
              endif  !<-- ierr==nf_noerr
            enddo
          else
            write(*,'(1x,6A/12x,A/)')       '### ERROR: Cannot find ',
     &         'variable ''', tname(1:lenstr(tname)), ''' in file ''',
     &                        fname(1:lsrc), '''.', nf_strerror(ierr)
          endif
          ierr=nf_close(ncid)
        else
          write(*,'(1x,3A/12x,A/)')  '### ERROR: Cannot open file ''',
     &            fname(1:lsrc), ''' for reading.', nf_strerror(ierr)
        endif
        ierr=nf_sync(nctarg)
      enddo !<-- iarg

#ifdef HOT_SPOTS
      call put_rec_by_name_real(nctarg, trg_fname, 'c2x',
     &                              Lm+2,Mm+2, 0,0, c2x)
      call put_rec_by_name_real(nctarg, trg_fname, 'c2w',
     &                              Lm+2,Mm+2, 0,0, c2w)
# ifdef PUT_CW3D
      call put_rec_by_name_real(nctarg, trg_fname, 'cw',
     &                              Lm+2,Mm+2,N, 0, cw)
# endif
#endif
      ierr=nf_close(nctarg)
      stop
      end

      subroutine init_cu_thread(Lm,Mm,N, z_w,Hz, rmask,pm,pn,
     &                              dn_u,dm_v,iA_r, c2x,c2w)
      implicit none
      integer Lm,Mm,N
      real(kind=8) z_w(0:Lm+1,0:Mm+1,0:N), rmask(0:Lm+1,0:Mm+1),
     &              Hz(0:Lm+1,0:Mm+1,N),    iA_r(0:Lm+1,0:Mm+1),
     &              pm(0:Lm+1,0:Mm+1),      dn_u(1:Lm+1,0:Mm+1),
     &              pn(0:Lm+1,0:Mm+1),      dm_v(0:Lm+1,1:Mm+1)
      real(kind=4) c2x(0:Lm+1,0:Mm+1),       c2w(0:Lm+1,0:Mm+1)

!>
      integer numthreads,trd, nsub_x,nsub_y, my_first,my_last,
     &                       range, tile, istr,iend,jstr,jend
C$    integer omp_get_num_threads, omp_get_thread_num
      numthreads=1 ; trd=0
C$    numthreads=omp_get_num_threads()
C$    trd=omp_get_thread_num()

      call set_tiles(Lm,Mm, nsub_x,nsub_y)
      range=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*range
      my_last=min(my_first + range-1, nsub_x*nsub_y-1)

      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                     istr,iend,jstr,jend)
        call init_cu_tile(istr,iend,jstr,jend, Lm,Mm,N, z_w,Hz,
     &                    rmask,pm,pn, dn_u,dm_v,iA_r, c2x,c2w)
      enddo
      end



      subroutine init_cu_tile(istr,iend,jstr,jend, Lm,Mm,N, z_w,Hz,
     &                        rmask,pm,pn, dn_u,dm_v,iA_r, c2x,c2w)
      implicit none
      integer istr,iend,jstr,jend, Lm,Mm,N,  i,j,k
      real(kind=8) z_w(0:Lm+1,0:Mm+1,0:N), rmask(0:Lm+1,0:Mm+1),
     &              Hz(0:Lm+1,0:Mm+1,N),    iA_r(0:Lm+1,0:Mm+1),
     &              pm(0:Lm+1,0:Mm+1),      dn_u(1:Lm+1,0:Mm+1),
     &              pn(0:Lm+1,0:Mm+1),      dm_v(0:Lm+1,1:Mm+1)
      real(kind=4) c2x(0:Lm+1,0:Mm+1),       c2w(0:Lm+1,0:Mm+1)

#include "compute_extended_bounds.h"

      do j=jstrR,jendR                ! Note: the logic for mask at
        do i=istrR,iendR              ! U- and V-points is essentially
          iA_r(i,j)=pm(i,j)*pn(i,j)   ! .OR., rmask(i,.)+rmask(i-1,.),
        enddo                         ! so having "land" on both sides
      enddo                           ! is required for the point to
      do j=jstrR,jendR                ! become "land".  This is to
        do i=istr,iendR
          if (rmask(i,j)+rmask(i-1,j) > 0.1D0) then
            dn_u(i,j)=2.D0/(pn(i,j)+ pn(i-1,j))
          else
            dn_u(i,j)=-1.D0           ! distinguish between zeroing
          endif                       ! out normal velocities at the
        enddo                         ! coastline and setting special
      enddo                           ! values inside land.
      do j=jstr,jendR
        do i=istrR,iendR
          if (rmask(i,j)+rmask(i,j-1) > 0.1D0) then
            dm_v(i,j)=2.D0/(pm(i,j)+ pm(i,j-1))
          else
            dm_v(i,j)=-1.D0
          endif
        enddo
      enddo
      do k=1,N
        do j=jstrR,jendR
          do i=istrR,iendR
            Hz(i,j,k)=z_w(i,j,k)-z_w(i,j,k-1)
          enddo
        enddo
      enddo
      do j=jstrR,jendR
        do i=istrR,iendR                ! initialize computation of
          c2x(i,j)=0.                   ! maxima of Courant numbers
          c2w(i,j)=0.                   ! throughout vertical columns
        enddo                           ! (these arrays are shared)
      enddo
      end


      subroutine comp_cu_thread(Lm,Mm,N, Hz,dn_u,dm_v,iA_r, rmask,
     &                                          u,v,W, c2x,c2w,cw,
#ifdef MAX_VELOCITY
     &                                                     uv_amp,
#endif
     &                                             dt, FillValue)
      implicit none
      integer Lm,Mm,N
      real(kind=8) dt, Hz(0:Lm+1,0:Mm+1,N),
     &              rmask(0:Lm+1,0:Mm+1), dn_u(1:Lm+1,0:Mm+1),
     &               iA_r(0:Lm+1,0:Mm+1), dm_v(0:Lm+1,1:Mm+1)

      real(kind=4) u(1:Lm+1,0:Mm+1,N),  W(0:Lm+1,0:Mm+1,0:N),
     &             v(0:Lm+1,1:Mm+1,N), cw(0:Lm+1,0:Mm+1,  N),
     &           c2x(0:Lm+1,0:Mm+1),  c2w(0:Lm+1,0:Mm+1)
#ifdef MAX_VELOCITY
     &                             ,  uv_amp(0:Lm+1,0:Mm+1)
#endif
      real(kind=4) FillValue

      integer numthreads,trd, nsub_x,nsub_y, my_first,my_last,
     &                       range, tile, istr,iend,jstr,jend
C$    integer omp_get_num_threads, omp_get_thread_num
      numthreads=1 ; trd=0
C$    numthreads=omp_get_num_threads()
C$    trd=omp_get_thread_num()

      call set_tiles(Lm,Mm, nsub_x,nsub_y)
      range=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*range
      my_last=min(my_first + range-1, nsub_x*nsub_y-1)

      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)

        call comp_cu_tile(istr,iend,jstr,jend, Lm,Mm,N, Hz,
     &            dn_u,dm_v,iA_r, rmask, u,v,W, c2x,c2w, cw,
#ifdef MAX_VELOCITY
     &                                               uv_amp,
#endif
     &                                       dt, FillValue)
      enddo
      end


      subroutine comp_cu_tile(istr,iend,jstr,jend, Lm,Mm,N,
     &          Hz,dn_u,dm_v,iA_r, rmask, u,v,W, c2x,c2w,cw,
#ifdef MAX_VELOCITY
     &                                               uv_amp,
#endif
     &                                       dt, FillValue)
      implicit none
      integer istr,iend,jstr,jend, Lm,Mm,N, i,j,k
      real(kind=8) dt, Hz(0:Lm+1,0:Mm+1,N),
     &              rmask(0:Lm+1,0:Mm+1), dn_u(1:Lm+1,0:Mm+1),
     &               iA_r(0:Lm+1,0:Mm+1), dm_v(0:Lm+1,1:Mm+1)

      real(kind=4) u(1:Lm+1,0:Mm+1,N),  W(0:Lm+1,0:Mm+1,0:N),
     &             v(0:Lm+1,1:Mm+1,N), cw(0:Lm+1,0:Mm+1,  N),
     &           c2x(0:Lm+1,0:Mm+1),  c2w(0:Lm+1,0:Mm+1)
#ifdef MAX_VELOCITY
     &                             ,  uv_amp(0:Lm+1,0:Mm+1)
#endif

      real(kind=4) FillValue , cff,FX
      real(kind=4) cx(istr-1:iend+1,jstr-1:jend+1)
#ifdef MAX_VELOCITY
     &           , uv2(istr-1:iend+1,jstr-1:jend+1)
#endif
#ifdef NO_OMEGA
     &           ,  wm(istr-1:iend+1,jstr-1:jend+1)
     &           ,   dh(istr-1:iend+1,         0:N)
#endif
#include "compute_extended_bounds.h"

#ifndef HOT_SPOTS
      do j=jstrR,jendR
        do i=istrR,iendR                ! initialize computation of
          c2x(i,j)=0.                   ! maxima of Courant numbers
          c2w(i,j)=0.                   ! throughout vertical columns
#ifdef MAX_VELOCITY
          uv_amp(i,j)=0.                ! (these arrays are shared)
#endif
        enddo
      enddo
#endif
#ifdef NO_OMEGA
      do j=jstr-1,jend+1
        do i=istr-1,iend+1              ! initialize computation of
          wm(i,j)=0.                    ! vertical velocity (this is
        enddo                           ! private array)
      enddo
#endif

      do k=1,N
        do j=jstr-1,jend+1
          do i=istr-1,iend+1
            cx(i,j)=0.  !<-- private array
#ifdef MAX_VELOCITY
            uv2(i,j)=0. !<-- private array
#endif
          enddo
        enddo
        do j=jstr,jend
          do i=istr,iend+1
            if (dn_u(i,j) > 0.D0) then
              FX=0.5*(Hz(i,j,k)+Hz(i-1,j,k))*dn_u(i,j)*u(i,j,k)
#ifdef NO_OMEGA
              wm(i  ,j)=wm(i  ,j)+FX
              wm(i-1,j)=wm(i-1,j)-FX
#endif
              if (FX > 0.) then
                cx(i,j)=cx(i,j)+FX
              else
                cx(i-1,j)=cx(i-1,j)-FX
              endif
#ifdef MAX_VELOCITY
              uv2(i  ,j)=uv2(i  ,j)+u(i,j,k)*u(i,j,k)
              uv2(i-1,j)=uv2(i-1,j)+u(i,j,k)*u(i,j,k)
#endif
            endif
          enddo
        enddo
        do j=jstr,jend+1
          do i=istr,iend
            if (dm_v(i,j) > 0.D0) then
              FX=0.5*(Hz(i,j,k)+Hz(i,j-1,k))*dm_v(i,j)*v(i,j,k)
#ifdef NO_OMEGA
              wm(i,j  )=wm(i,  j)+FX
              wm(i,j-1)=wm(i,j-1)-FX
#endif
              if (FX > 0.) then
                cx(i,j)=cx(i,j)+FX
              else
                cx(i,j-1)=cx(i,j-1)-FX
              endif
#ifdef MAX_VELOCITY
              uv2(i,j  )=uv2(i,j  )+v(i,j,k)*v(i,j,k)
              uv2(i,j-1)=uv2(i,j-1)+v(i,j,k)*v(i,j,k)
#endif
            endif
          enddo
        enddo
        do j=jstr,jend
          do i=istr,iend
            if (rmask(i,j) > 0.5) then
              cff=dt/Hz(i,j,k)
              cx(i,j)=cx(i,j)*iA_r(i,j)*cff
              c2x(i,j)=max(c2x(i,j), cx(i,j))
#ifdef MAX_VELOCITY
              uv_amp(i,j)=max(uv_amp(i,j), sqrt(0.5*uv2(i,j)))
#endif
#ifdef NO_OMEGA
              W(i,j,k)=wm(i,j)
#else
              if (W(i,j,k) > 0.) then
                cw(i,j,k)=W(i,j,k)
              else
                cw(i,j,k)=0.
              endif
              if (W(i,j,k-1) < 0.) then
                cw(i,j,k)=cw(i,j,k)-W(i,j,k-1)
              endif

              cw(i,j,k)=cw(i,j,k)*cff
              c2w(i,j)=max(c2w(i,j),cw(i,j,k))
#endif
            else
              c2x(i,j)=FillValue
#ifdef MAX_VELOCITY
              uv_amp(i,j)=FillValue
#endif
#ifndef NO_OMEGA
              cw(i,j,k)=FillValue
              c2w(i,j)=FillValue
#endif
            endif
          enddo
        enddo
      enddo !<-- k, discard cx, wm, keep c2w

#ifdef NO_OMEGA
      do j=jstr,jend
        do i=istr,iend
          dh(i,0)=0.
        enddo
        do k=1,N,+1
          do i=istr,iend
            dh(i,k)=dh(i,k-1)+Hz(i,j,k)  ! <-- distance from the bottom
          enddo
        enddo
        do i=istr,iend
          cx(i,j)=W(i,j,N)/dh(i,N)       !<-- free-surface correction
          cw(i,j,N)=0.
        enddo
        do k=N-1,1,-1       !--> irreversible     ! Note that in this
          do i=istr,iend    !    because of "cw"  ! algorithm W(:,:,0)
            cw(i,j,k)=0.                          ! is never assigned
            if (rmask(i,j) > 0.5) then            ! values and never
              FX=W(i,j,k)-dh(i,k)*cx(i,j)         ! accessed.
              if (FX > 0.) then
                cw(i,j,k)=FX
              else
                cw(i,j,k+1)=cw(i,j,k+1)-FX   !<-- add to previous "k"
              endif
            endif
          enddo
        enddo                  !--> discard cx
        do k=1,N
          do i=istr,iend
            if (rmask(i,j) > 0.5) then
              cw(i,j,k)=cw(i,j,k) *dt*iA_r(i,j)/Hz(i,j,k)
              c2w(i,j)=max(c2w(i,j),cw(i,j,k))
            else
              cw(i,j,k)=FillValue
              c2w(i,j)=FillValue
            endif
          enddo
        enddo
      enddo ! <-- j
#endif
      end
