      program drakkar

! Purpose: interpolate scalar forcing fields of DRAKKAR dataset (or,
! in fact, any data with similar file structure) into ROMS grid. 

! Usage 
!        drakkar roms_grid_file.nc file1.nc [file2.nc file3.nc ...]

! where file1.nc file2.nc ...  are assumed to

!  (1) contain a single 2D scalar field accompanied by a timing
!      variable (same pair for all files in the sequence), and

!  (2) the file format is expected to be "CF-compliant", that is
!      time variable has the same name as time dimension.

!  (3) file names in the sequence a labelled by year, thus having
!      something like "_y1234_" inserted somewhere in the middle
!      (if not, the program is still applicable, but only for a
!      single file rather than sequence).  

! The output created by this program is a sequence of netCDF files
! named after netCDF name of the 2D field they contain, and a year
! label, so one input file corresponds to one output.

! See "drakkar_winds.F" for explanation of CPP-macros below.



#define TIME_SHIFT 2000
c-#define TIME_OVERLAP

#define SPLINE_INTER

      use nc_file_coords
      implicit none
      real(kind=8) time
      integer(kind=2), allocatable, dimension(:) :: msk
      real(kind=8), allocatable, dimension(:) :: lon_r,lat_r, xlon,ylat
      real(kind=4), allocatable, dimension(:) :: bffr, targ

      integer Lm,Mm, roms_grid_size, lon,lat, src_grid_size, nargs,
     &        ncgrd, ncsrc, nctarg, ndims, nvars, natts, vid, vid_fld,
     &        vid_time, varid, vtype, length, irec, tgtrec, i, icase,
     &        vdimids(4), tdimids(4), r2dgrd(4), start(4),count(4),
     &        iarg, ierr, lgrd, lsrc,ltrg, lvar, ltvr, iargc, lenstr 

      character(len=64), allocatable, dimension(:) :: fname
      character(len=64) grid_file, src_file, targ_file
      character(len=32) varname, timevar, vname, dname, tname

#include "netcdf.inc"

      nargs=iargc(); roms_grid_size=0; lvar=0
      if (nargs > 1) then
        allocate(fname(nargs))
        do iarg=1,nargs
          call getarg(iarg,fname(iarg))
        enddo
      else
        write(*,'(/1x,A/ /12x,2A/)') 'Usage should be: ',
     &         'drakkar read_roms_grid_file.nc file.nc ',
     &                         '[file2.nc file3.nc ...]'
        stop
      endif

! Initialize ROMS grid...

      grid_file=fname(1) ;  lgrd=lenstr(grid_file)
      ierr=nf_open(grid_file, nf_nowrite, ncgrd)
      if (ierr == nf_noerr) then
        call roms_find_dims(ncgrd, grid_file, Lm,Mm, i)
        roms_grid_size=(Lm+2)*(Mm+2)
        allocate(lon_r(roms_grid_size), lat_r(roms_grid_size),
     &                                    msk(roms_grid_size))
        call get_var_by_name_double(ncgrd, 'mask_rho', lon_r)
        do i=1,roms_grid_size
          if (lon_r(i) > 0.5D0) then    ! temporarily use "lon_r"
            msk(i)=1                    ! to read file: thereafter
          else                          ! this program uses only
            msk(i)=0                    ! integer(kind=2) mask
          endif
        enddo
        call get_var_by_name_double(ncgrd, 'lon_rho', lon_r)
        call get_var_by_name_double(ncgrd, 'lat_rho', lat_r)
        ierr=nf_close(ncgrd)
      else
        write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &               'file ''', grid_file(1:lgrd), ''' for reading.',
     &                                           nf_strerror(ierr)
      endif


! Start processing: note that "time_size" is assigned value by
! "nc_find_dims" and contains the number of time records of the
! source file.

      do iarg=2,nargs
        src_file=fname(iarg) ; lsrc=lenstr(src_file)
        call nc_find_dims(src_file, lon,lat,i)
        src_grid_size=lon*lat
        write(*,*) 'time_size =', time_size

        ierr=nf_open(src_file, nf_nowrite, ncsrc)
        if (ierr == nf_noerr) then
          if (.not.allocated(bffr)) then
            allocate( xlon(lon),ylat(lat), bffr(src_grid_size),
     &                                     targ(roms_grid_size))

            call get_var_by_name_double(ncsrc, 'lon', xlon)
            call get_var_by_name_double(ncsrc, 'lat', ylat)

            call init_interp_2d(lon,lat, xlon,ylat, Lm,Mm, lon_r,lat_r)
          endif

! Search for the 2D field (must be only one in the file) and its
! timing variable.    Note that "lvar" initially set to zero plays
! the role of logical trap to check for uniqueness of the 2D field.
! The outcome of the following block are names and netCDF variable
! IDs for the field and timing variable.

          lvar=0

          ierr=nf_inq_nvars(ncsrc, nvars)
          if (ierr == nf_noerr) then
            do vid=1,nvars
              ierr=nf_inq_var(ncsrc, vid, vname, vtype,
     &                           ndims, vdimids, natts)
              if (ierr == nf_noerr) then
                if (ndims == 3) then
                  ierr=nf_inq_dim(ncsrc, vdimids(3), dname, length)
                  if (ierr == nf_noerr) then
                    ierr=nf_inq_varid(ncsrc, dname, varid)
                    if (ierr == nf_noerr) then
                      ierr=nf_inq_var(ncsrc, varid, tname,   vtype,
     &                                       ndims, tdimids, natts)
                      if (ierr == nf_noerr) then
                        if (tname == dname .and. ndims == 1 .and.
     &                             tdimids(1) == vdimids(3)) then
                          if (lvar == 0) then
                            vid_fld=vid   ; vid_time=varid
                            varname=vname ; lvar=lenstr(varname)
                            timevar=tname ; ltvr=lenstr(timevar)
                            write(*,'(3(1x,A))') 'found',
     &                          varname(1:lvar), timevar(1:ltvr)
                          else
                            write(*,*)  '### ERROR: Ambiguous: more ',
     &                          'than one field and timing variable ',
     &                          'found in ''', src_file(1:lsrc),'''.'
                            stop
                          endif
                        endif
                      endif
                    endif
                  endif
                endif
              endif
            enddo
          else
            write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot determine ',
     &              'number of variables in file ''', src_file(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &                  'file ''', src_file(1:lsrc), ''' for reading.',
     &                                               nf_strerror(ierr)
        endif
        if (lvar == 0) stop


! Generate target file name:  Capture year label like "_y1234" from
! the source file name and insert it into the target (verify that all
! symbols following '_y' are digits).

        i=1
        do while(src_file(i:i+1) /= '_y' .and. i < lsrc-1)
          i=i+1
        enddo
        if (src_file(i:i+1) == '_y' .and.
     &   '0' <= src_file(i+2:i+2) .and. src_file(i+2:i+2) <= '9' .and.
     &   '0' <= src_file(i+3:i+3) .and. src_file(i+3:i+3) <= '9'  .and.
     &   '0' <= src_file(i+4:i+4) .and. src_file(i+4:i+4) <= '9'  .and.
     &   '0' <= src_file(i+5:i+5) .and. src_file(i+5:i+5) <= '9') then

          targ_file=varname(1:lvar)/ /src_file(i:i+5)/ /'.nc'
        else
          targ_file=varname(1:lvar)/ /'.nc'
        endif
        ltrg=lenstr(targ_file)


! Create target file: define dimensions and variables. Copy their
!------- ------- ---- attributes from the source to the target.

        ierr=nf_create(targ_file(1:ltrg), nf_64bit_offset, nctarg)
        ierr=nf_def_dim(nctarg, 'xi_rho',  Lm+2, r2dgrd(1))
        ierr=nf_def_dim(nctarg, 'eta_rho', Mm+2, r2dgrd(2))
        ierr=nf_def_dim(nctarg,timevar(1:ltvr),nf_unlimited,r2dgrd(3))

        do icase=1,2
          if (icase == 1) then
            ierr=nf_def_var(nctarg, timevar(1:ltvr), nf_double, 1,
     &                                             r2dgrd(3), vid)
            vname=timevar(1:ltvr)
          else
            ierr=nf_def_var(nctarg, varname(1:lvar),  nf_float, 3,
     &                                               r2dgrd, vid)
            vname=varname(1:lvar)
          endif
          if (ierr == nf_noerr) then
            ierr=nf_inq_varid(ncsrc, vname, varid)
            if (ierr == nf_noerr) then
              ierr=nf_inq_varnatts(ncsrc, varid, natts)
              if (ierr == nf_noerr) then
                do i=1,natts
                  ierr=nf_inq_attname(ncsrc, varid, i, tname)
                  if (ierr == nf_noerr) then
#ifdef TIME_SHIFT
                    if (icase == 1 .and. tname(1:5) == 'units') then
                      ierr=nf_put_att_text( nctarg,  vid,  tname,  30,
# if TIME_SHIFT == 2000
     &                               'days since 2000-01-01 00:00:00')
# elif TIME_SHIFT == 1958
     &                               'days since 1958-01-01 00:00:00')
# endif
                    else
#endif
                      ierr=nf_copy_att(ncsrc,varid, tname, nctarg,vid)
#ifdef TIME_SHIFT
                    endif
#endif
                  endif
                enddo
              endif
            endif
          endif
        enddo
        ierr=nf_enddef(nctarg)
        ierr=nf_sync(nctarg)
        tgtrec=0  !<-- initialization of the target file complete. 
      
! Repeat the last record from the previous file to ensure one record
! overlap between consecutive files (only if TIME_OVERLAP is set):

#ifdef TIME_OVERLAP
        if (iarg > 2) then
          tgtrec=tgtrec+1
          call put_sclr_rec_by_name_double(nctarg, targ_file,
     &                         timevar(1:ltvr), tgtrec, time)

          call put_rec_by_name_real( nctarg, targ_file,
     &            varname(1:lvar), Lm+2,Mm+2,0, tgtrec, targ)
        endif
#endif

! Note that input record "irec" and output record "tgtrec" below are
! not necessarily the same because of an extra output record associated
! with TIME_OVERLAP option.

        write(*,'(/1x,3A/)') 'Start processing',src_file(1:lsrc),'...'

        do irec=1,time_size
          start(1)=1; count(1)=lon ; start(3)=irec
          start(2)=1; count(2)=lat ; count(3)=1
          ierr=nf_get_vara_double(ncsrc, vid_time, start(3),count(3),
     &                                                         time)

! Redefine time from days "days since 1900-01-01 00:00:00" into
! "days since 1958-01-01 00:00:00" which is the starting time of
! Drakkar dataset;

#if TIME_SHIFT == 2000
                          time=time-36524.0D0

#elif TIME_SHIFT == 1958
                          time=time-21184.0D0
#endif
          if (ierr == nf_noerr) then
            ierr=nf_get_vara_real(ncsrc, vid_fld, start,count, bffr)
            if (ierr == nf_noerr) then
              write(*,*) 'read record =',irec,' out of',time_size
#ifdef SPLINE_INTER
               call extend_on_land(lon,lat, bffr)
               call spline_interp_r2d(lon,lat, bffr, Lm,Mm, targ)
#else
               call interp_r2d(lon,lat, bffr, Lm,Mm, msk, targ)
#endif
               tgtrec=tgtrec+1
               call put_sclr_rec_by_name_double(nctarg, targ_file,
     &                             timevar(1:ltvr), tgtrec, time)

               call put_rec_by_name_real( nctarg, targ_file,
     &                 varname(1:lvar), Lm+2,Mm+2,0, tgtrec, targ)
            else
              write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot read ',
     &                   'variable ''', varname(1:lvar), ''' from ''',
     &                    src_file(1:lsrc), '''.', nf_strerror(ierr)
            endif
          else
             write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot read ',
     &                  'variable ''', timevar(1:ltvr), ''' from ''',
     &                   src_file(1:lsrc), '''.', nf_strerror(ierr)
          endif
          if (ierr /= nf_noerr) stop
        enddo
        ierr=nf_close(nctarg)
        ierr=nf_close(ncsrc)
      enddo !<-- iarg=2,nargs
      end

