      program drakkar_winds

! Purpose: interpolate DRAKKAR winds into roms grid. 
! Notes:

!   (1) this program requires no arguments as everything is hard-coded
!       inside;  parameter "grid_file" should point to the actual ROMS
!       grid file;  and "prefix" should point to the correct directory
!       with Drakkar wind data files and contain the common starting
!       portion of all file names there;

!   (2) CPP-macro TIME_SHIFT changes definition of time from
!                 'days since 1900-01-01 00:00:00'
!       into
!                 'days since 1958-01-01 00:00:00' which is
!       the initial date of Drakkar reanalysis;

!   (3) TIME_OVERLAP makes the last record in a file of the sequence
!       be repeated as the starting record in the next file;

!   (4) Bulk flux requires both WIND_VELOCITY and WIND_AT_RHO_POINTS;

!   (5) CPP-macro ROTATE rotates velocity vector from lon-lat
!       (X-component points eastward, Y- to the north) to the
!       directions of ROMS grid.  Not needed for non-rotated
!       Mercator grid.

!   (6) CPP-macro SPLINE_INTER selects bi-cubic spline interpolation
!       using Hermite polynomials;  (bi-linear interpolation is used
!       if this CPP-macro is not defined).



#define TIME_SHIFT 2000
c--#define TIME_OVERLAP

#define WIND_VELOCITY
#define WIND_AT_RHO_POINTS
#define SPLINE_INTER

#define ROTATE

! These are dimensions of output arrays written into in netCDF files;

#ifdef WIND_AT_RHO_POINTS
# define NCDF_U_GRID_TYPE  Lm+2,Mm+2,0
# define NCDF_V_GRID_TYPE  Lm+2,Mm+2,0
#else
# define NCDF_U_GRID_TYPE  Lm+1,Mm+2,0
# define NCDF_V_GRID_TYPE  Lm+2,Mm+1,0
#endif

! The following CPP-macros are  not directly used within this file,
! but are needed by "def_wind_file.F" included at the bottom below.
! They specify the names of corresponding variables in netCDF files
! created by this program.

#define WND_TIME 'wnd_time'
#define UWIND 'uwnd'
#define VWIND 'vwnd'

      use nc_file_coords
      implicit none

      integer lon,lat, Lm,Mm, i, src_grid_size, roms_grid_size
      real(kind=8) time
      real(kind=8), allocatable, dimension(:) :: xlon,ylat
      real(kind=4), allocatable, dimension(:) :: usrc,vsrc, utrg,vtrg

      logical curv_grid
      integer(kind=2), allocatable, dimension(:) :: msk
      real(kind=8), allocatable, dimension(:) :: lon_r,lat_r,csA,snA

      character(len=80), parameter ::

c    &     grid_file='grid.nc',
c    &     prefix='/woossee/DFS5.2/drowned_',

c    &     grid_file='/maya/alex/pac23_input/pac23_grid.nc',
     &     grid_file='/maya/alex/black_sea/black_sea_grd.nc',

     &     prefix='/home/alex/DFS5.2/drowned_',
     &     middle='10_DFS5.2_y',
     &     suffix='_ROMS_version.nc'

      character(len=64) ufile, vfile, targ_file
      character(len=4) str4
      integer ncgrd, nuid,nvid, nctarg, vid_T,vid_U,vid_V, year,
     &        irec, trgrec, file_in_sequence, start(4),count(4),
     &        ierr, lgrd, lprfx,lmddl,lsffx, lufl,lvfl, lenstr

#include "netcdf.inc"

      file_in_sequence=0   ; lprfx=lenstr(prefix)
      lmddl=lenstr(middle) ; lsffx=lenstr(suffix)

! Open ROMS grid file and read coordinates, angle, and mask. 

      lgrd=lenstr(grid_file)
      ierr=nf_open(grid_file, nf_nowrite, ncgrd)
      if (ierr == nf_noerr) then
        call roms_find_dims(ncgrd, grid_file, Lm,Mm, i)
        roms_grid_size=(Lm+2)*(Mm+2)
        allocate( lon_r(roms_grid_size), csA(roms_grid_size),
     &            lat_r(roms_grid_size), snA(roms_grid_size),
     &                                  msk(roms_grid_size))
        call get_var_by_name_double(ncgrd, 'lon_rho', lon_r)
        call get_var_by_name_double(ncgrd, 'lat_rho', lat_r)
        call get_var_by_name_double(ncgrd, 'angle',    csA)
        call get_var_by_name_double(ncgrd, 'mask_rho', snA)
        ierr=nf_close(ncgrd)
        curv_grid=.false.             ! temporarily use "snA"
        do i=1,roms_grid_size         ! to read file: thereafter
          if (snA(i) > 0.5D0) then    ! this program uses only
            msk(i)=1                  ! integer(kind=2) version
          else                        ! of land mask
            msk(i)=0
          endif
          if (abs(csA(i)) > 1.D-12) curv_grid=.true.
          snA(i)=sin(csA(i))
          snA(i)=cos(csA(i))
        enddo
      else
        write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &               'file ''', grid_file(1:lgrd), ''' for reading.',
     &                                           nf_strerror(ierr)
        stop
      endif


c     do year=1958,2012
      do year=2000,2012

        write(str4,'(I4)') year
        ufile=prefix(1:lprfx)/ /'u'/ /middle(1:lmddl)/ /str4/
     &                                       /suffix(1:lsffx)
        vfile=prefix(1:lprfx)/ /'v'/ /middle(1:lmddl)/ /str4/
     &                                       /suffix(1:lsffx)
        targ_file='wind_y'/ /str4/ /'.nc'


        lufl=lenstr(ufile) ; lvfl=lenstr(vfile)
        call nc_find_dims(ufile, lon,lat,i)
        src_grid_size=lon*lat
        write(*,*) 'time_size =', time_size


        call def_wind_file(targ_file, nctarg, Lm,Mm)
        trgrec=0 ; file_in_sequence=file_in_sequence+1

! Repeat the last record from the previous file to ensure
! one record overlap between consecitive files:

#ifdef TIME_OVERLAP
        if (file_in_sequence > 1) then
         trgrec=trgrec+1
         call put_sclr_rec_by_name_double(nctarg, targ_file,
     &                              WND_TIME, trgrec, time)

         call put_rec_by_name_real(nctarg, targ_file, UWIND,
     &                      NCDF_U_GRID_TYPE, trgrec, utrg)

         call put_rec_by_name_real(nctarg, targ_file, VWIND,
     &                      NCDF_V_GRID_TYPE, trgrec, vtrg)
        endif
#endif
        ierr=nf_open(ufile, nf_nowrite, nuid)
        if (ierr == nf_noerr) then
          ierr=nf_open(vfile, nf_nowrite, nvid)
          if (ierr == nf_noerr) then
            ierr=nf_inq_varid(nuid, 'wind_time', vid_T)
            if (ierr == nf_noerr) then
              ierr=nf_inq_varid(nuid, 'Uwind', vid_U)
              if (ierr == nf_noerr) then
                ierr=nf_inq_varid(nvid, 'Vwind', vid_V)
                if (ierr == nf_noerr) then
                  write(*,*) 'Start processing ', ufile(1:lufl), '...'
                  if (.not.allocated(usrc)) then
                    allocate( xlon(lon), ylat(lat),
     &                  usrc(src_grid_size), vsrc(src_grid_size),
     &                 utrg(roms_grid_size), vtrg(roms_grid_size))

                    call get_var_by_name_double(nuid, 'lon', xlon)
                    call get_var_by_name_double(nuid, 'lat', ylat)

                    call init_interp_2d(lon,lat, xlon,ylat, Lm,Mm,
     &                                                lon_r,lat_r)
                  endif
                  do irec=1,time_size
                    start(1)=1; count(1)=lon ; start(3)=irec
                    start(2)=1; count(2)=lat ; count(3)=1
                    ierr=nf_get_vara_double(nuid, vid_T, start(3),
     &                                            count(3), time)

! Redefine time from days "days since 1900-01-01 00:00:00" into
! "days since 1958-01-01 00:00:00" which is the starting time of
! Drakkar dataset;

#if TIME_SHIFT == 2000
                          time=time-36524.0D0

#elif TIME_SHIFT == 1958
                          time=time-21184.0D0
#endif

                    if (ierr == nf_noerr) then
                      ierr=nf_get_vara_real(nuid, vid_U, start,count,
     &                                                          usrc)
                      if (ierr == nf_noerr) then
                        ierr=nf_get_vara_real(nvid, vid_V, start,count,
     &                                                           vsrc)
                        if (ierr == nf_noerr) then


! Computational part....

              write(*,*) 'read rec =',irec, ' out of', time_size
#ifdef SPLINE_INTER
              call extend_on_land(lon,lat, usrc)
              call extend_on_land(lon,lat, vsrc)
# ifdef WIND_AT_RHO_POINTS
              call spline_interp_r2d(lon,lat, usrc, Lm,Mm, utrg)
              call spline_interp_r2d(lon,lat, vsrc, Lm,Mm, vtrg)
# else
              call spline_interp_u2d(lon,lat, usrc, Lm,Mm, utrg)
              call spline_interp_v2d(lon,lat, vsrc, Lm,Mm, vtrg)
# endif
#else
# ifdef WIND_AT_RHO_POINTS
              call interp_r2d(lon,lat, usrc, Lm,Mm, msk, utrg)
              call interp_r2d(lon,lat, vsrc, Lm,Mm, msk, vtrg)
# else
              call interp_u2d(lon,lat, usrc, Lm,Mm, msk, utrg)
              call interp_v2d(lon,lat, vsrc, Lm,Mm, msk, vtrg)
# endif
#endif

              if (curv_grid) call rotate_in_place(Lm,Mm, csA,snA,
     &                                           msk, utrg,vtrg)


! Write everything into target file...

                          trgrec=trgrec+1
                          call put_sclr_rec_by_name_double( nctarg,
     &                             targ_file, WND_TIME, trgrec, time)

                          call put_rec_by_name_real(nctarg, targ_file,
     &                         UWIND, NCDF_U_GRID_TYPE, trgrec, utrg)

                          call put_rec_by_name_real(nctarg, targ_file,
     &                         VWIND, NCDF_V_GRID_TYPE, trgrec, vtrg)
                        else
                          write(*,'(/1x,4A/12x,A/)')    '### ERROR: ',
     &                         'Cannot read variable ''Vwind'' from ',
     &                          vfile(1:lvfl),'.', nf_strerror(ierr)
                        endif
                      else
                        write(*,'(/1x,4A/12x,A/)')      '### ERROR: ',
     &                         'Cannot read variable ''Uwind'' from ',
     &                          ufile(1:lufl),'.', nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot ',
     &                             'read variable ''wind_time'' from ',
     &                           ufile(1:lufl),'.', nf_strerror(ierr)
                    endif
                  enddo !<-- irec=1,time_size
                else
                  write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot find ',
     &                     'variable ''Vwind'' inside ', vfile(1:lvfl),
     &                                          '.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot find ',
     &                     'variable ''Uwind'' inside ', ufile(1:lufl),
     &                                          '.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot find ',
     &                 'variable ''wind_time'' inside ', ufile(1:lufl),
     &                                          '.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,4A/12x,A/)')      '### ERROR: Cannot open ',
     &          'netCDF file ', vfile(1:lvfl), '.', nf_strerror(ierr)
          endif
        else
         write(*,'(/1x,4A/12x,A/)')        '### ERROR: Cannot open ',
     &         'netCDF file ', ufile(1:lufl), '.', nf_strerror(ierr)
        endif
        ierr=nf_close(nctarg)
        ierr=nf_close(nuid)
        ierr=nf_close(nvid)
        if (ierr /= nf_noerr) stop
      enddo  !<-- year=..., loop over files 
      end

#include "def_wind_file.F"
