      program etopo

! Purpose: Read "raw" topography from specified source of data (ETOPO,
!          etc...) and interpolate or coarsen it onto ROMS model grid.
!          The outcome is written as netCDF variable "hraw".

! NOTES: (1) the content of this file is just a driver: the actual
!            interpolation/coarsening routine is in "compute_hraw.F".

!        (2) nothing is done here about setting mask, which needs to
!            be processed separately via "rmask.F".   This is avoid
!            accidental overwriting of land mask field "rmask" which
!            may be subject to elaborate hand editing.

! Created and maintained by Alexander Shchepetkin, old_galaxy@yahoo.com

      use roms_grid_vars
      use comm_vars_hraw

      implicit none
      real(kind=8) radius,  dx, ptest, epsil
      real(kind=8), allocatable :: xlon(:),ylat(:), xtmp(:),hraw(:,:)
      real(kind=4), allocatable :: htopo(:,:), bffr(:,:)
      integer nargs, Lm,Mm, lon,lat, iwest,ieast,jsouth,jnorth, period,
     &     shft_west,shft_east,ishft, nx_lon,ny_lat, start(2),count(2),
     &       i,j, ncid, ndims, varid, id, size, ierr, lsrc, lgrd, ldim,
     &                                            lenstr, indx_bound
      character(len=80) srcfile, grid
      character(len=16) str, dname, lon_name, lat_name

#include "netcdf.inc"

      nargs=iargc()
      if (nargs==3) then
        call getarg(1,str)     ; read(str,*) radius
        call getarg(2,srcfile) ; lsrc=lenstr(srcfile)
        call getarg(3,grid)    ; lgrd=lenstr(grid)

        write(*,'(/1x,5A/1x,A,F8.5/)') 'Processing ''',srcfile(1:lsrc),
     &   ''' --> ''', grid(1:lgrd), '''', 'smoothing radius =', radius

      else
        write(*,'(/1x,A/ /8x,A/)')   'usage of "etopo" should be:',
     &                  'etopo smoothing_radius source.nc grid.nc'
        stop
      endif

      lon=0 ; shft_west=0  ! reset everything
      lat=0 ; shft_east=0

! 1. Open and read ROMS grid file. Find geographical limits of the grid.

      call read_roms_grid(grid, Lm,Mm)
      call roms_grid_geo_bounds(lon_r,lat_r, Lm,Mm, radius)

      write(*,'(1x,2A,2F16.8/22x,A,2F16.8/)')  'roms grid extremes: ',
     &   'longitude:', lon_min, lon_max, 'latitude:', lat_min,lat_max

! 2. Open and read topography data file....

      ierr=nf_open (srcfile, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        ierr=nf_inq_ndims (ncid, ndims)
        if (ierr == nf_noerr) then
          do id=1,ndims
            dname='       '
            ierr=nf_inq_dim (ncid, id, dname, size)
            if (ierr == nf_noerr) then
              ldim=lenstr(dname)
              if( (ldim==1 .and. dname(1:ldim)=='x')  .or.
     &            (ldim==3 .and. dname(1:ldim)=='lon')  .or.
     &            (ldim==9 .and. dname(1:ldim)=='longitude') ) then
                lon_name=dname(1:ldim)
                lon=size
              elseif( (ldim==1 .and. dname(1:ldim)=='y') .or.
     &                (ldim==3 .and. dname(1:ldim)=='lat') .or.
     &                (ldim==8 .and. dname(1:ldim)=='latitude') ) then
                lat_name=dname(1:ldim)
                lat=size
              endif
            else
              write(*,*) '### ERROR: dimension id =', id, '?'
            endif
          enddo
        else
          write(*,*) '### ERROR: nf_inq_ndims?'
        endif
        write(*,'(3(1x,A),I6,2x,2(1x,A),I6)') 'Found dimensions:',
     &                     lon_name(1:lenstr(lon_name)), '=', lon,
     &                     lat_name(1:lenstr(lat_name)), '=', lat

                                        ! NOTE: Since the data file is
                                        ! expected to be what is called
        if (lon > 0 .and. lat > 0) then ! "CF-compliant", its coordinate
          allocate(xlon(lon))           ! variables should have the same
          allocate(ylat(lat))           ! names as their corresponding
                                        ! dimensions.

          ierr=nf_inq_varid (ncid, lon_name, varid)
          if (ierr == nf_noerr) then
            ierr=nf_get_var_double (ncid, varid, xlon)
            if (ierr == nf_noerr) then
                                              ! this is needed to fix
              do i=2,lon,+1                   ! etopo5 longitude which
                if (xlon(i)<xlon(i-1)) then   ! goes from 0 to +180,
                  xlon(i)=xlon(i)+360.D0      ! then jumps to -180 in
                endif                         ! the middle, and after
              enddo                           ! that goes -180 to 0.

              ierr=nf_inq_varid(ncid, lat_name, varid)
              if (ierr == nf_noerr) then
                ierr=nf_get_var_double(ncid, varid, ylat)
                if (ierr == nf_noerr) then
                  write(*,'(4(1x,A))')     'retrieved',
     &                  lon_name(1:lenstr(lon_name)),
     &                  lat_name(1:lenstr(lat_name)), 'coordinates.'
                  ierr=nf_inq_varid (ncid, 'topo', varid)
                  if (ierr == nf_noerr) then
                    write(*,*) 'Found variable ''topo''.'
                  else
                    ierr=nf_inq_varid (ncid, 'elevation', varid)
                    if (ierr == nf_noerr) then
                      write(*,*) 'Found variable ''elevation''.'
                    else
                      ierr=nf_inq_varid (ncid, 'z', varid)
                      if (ierr == nf_noerr) then
                        write(*,*) 'Found variable ''z''.'
                      endif
                    endif
                  endif
                  if (ierr == nf_noerr) then

! Determine bounding indices for the relevant portion of topographic
! data (this is the only part of data which will be actually extracted
! from the data set. Compute dimensions of array to hold this portion
! of data, and move the relevant parts of coordinate arrays to the
! beginning. Then allocate the array and read the data.

                    jsouth=indx_bound(ylat,lat, lat_min)
                    jnorth=indx_bound(ylat,lat, lat_max)

                    if ( 0 < jsouth .and. jsouth < lat .and.
     &                   0 < jnorth .and. jnorth < lat ) then
                      if (jsouth > 1) jsouth=jsouth-1
                      jnorth=jnorth+1
                    else
                      write(*,'(/1x,2A,2F12.6/12x,2A/)') '### ERROR: ',
     &                   'North-south extents of the dataset', ylat(1),
     &                       ylat(lat), 'are not sufficient to cover ',
     &                                        'the entire model grid.'
                      stop
                    endif

                    iwest=indx_bound(xlon,lon, lon_min)
                    ieast=indx_bound(xlon,lon, lon_max)

                    if ( 0 < iwest .and. iwest < lon .and.
     &                   0 < ieast .and. ieast < lon ) then
                      if (iwest > 1) iwest=iwest-1
                      ieast=ieast+1

                      write(*,'(1x,2A,2I5,2x,A,2F12.6)')   'Single ',
     &                 'region dataset, iwest,ieast =',  iwest,ieast,
     &                 'xlon(iwest:ieast) =',xlon(iwest),xlon(ieast)
                    else

! Model grid does not fit directly into the east-west extents of the
! dataset. Check whether the data set longitude covers the entire 360
! degrees, so it will be possible to treat is as periodic.
! Two variants are of datasets are known to exist: either the end
! points may repeated,

!                      x(1)+360 = x(lon)
! or may be not,
!                      x(1)+360 = x(lon)+dx.

! Identify which is the case and find periodicity range accordingly.

                      ptest=xlon(lon)-xlon(1)-360.D0
                      dx=(xlon(lon)-xlon(1))/dble(lon-1)
                      epsil=0.01D0*abs(dx)
                      if (abs(ptest) < epsil) then
                        period=lon-1
                      elseif (abs(ptest+dx) < epsil) then
                        period=lon
                      else
                        period=0
                      endif

                      if (period > 0) then
                        write(*,'(/1x,2A,I6,1x,A,I6/)')  'Identified ',
     &                       'periodicity domain in data of ', period,
     &                                           'points out of', lon
                      else
                        write(*,'(/1x,2A/)')  '### ERROR: The data ',
     &                             'does not cover the entire grid.'
                        stop
                      endif

                      shft_west=0
                      if (iwest==0) then
                        shft_west=-1
                        iwest=indx_bound(xlon,lon, lon_min+360.D0)
                      elseif (iwest==lon) then
                        shft_west=+1
                        iwest=indx_bound(xlon,lon, lon_min-360.D0)
                      endif

                      shft_east=0
                      if (ieast == 0) then
                        shft_east=-1
                        ieast=indx_bound(xlon,lon, lon_max+360.D0)
                      elseif (ieast == lon) then
                        shft_east=+1
                        ieast=indx_bound(xlon,lon, lon_max-360.D0)
                      endif

                      if ( 0 < iwest .and. iwest < lon .and.
     &                     0 < ieast .and. ieast < lon ) then
                        if (iwest > 1) iwest=iwest-1
                        ieast=ieast+1
                      else
                         write(*,'(/1x,2A/12x,A/)') '### ERROR: Data ',
     &                      'longitude covers 360 degrees, but still ',
     &                      'cannot find  starting and ending indices.'
                        stop
                      endif
                    endif

                    write(*,'(/1x,2A/2(1x,A,I5,I6,1x,A,I6))')
     &             'Bounding indices of the relevant part to be ',
     &                       'extracted from the entire dataset:',
     &             'iwest,ieast =',  iwest,ieast,   'out of', lon,
     &             'jsouth,jnorth =',jsouth,jnorth, 'out of', lat

                    ny_lat=jnorth-jsouth+1
                    start(2)=jsouth  ; count(2)=ny_lat
                    do j=1,ny_lat,+1
                      ylat(j)=ylat(j+jsouth-1)
                    enddo

! Once we made it up to here, there are two possibilities: either the
! topographic data covering the entire model grid can be retrieved as
! a single rectangular subregion within the dataset, or it has to be
! pieced together from two parts using 360-degree periodicity.

                    if (iwest < ieast) then
                      nx_lon=ieast-iwest+1
                      start(1)=iwest ; count(1)=nx_lon

                      write(*,'(1x,4A)',advance='no')     'Reading ',
     &                'topography from ''', srcfile(1:lsrc), '''...'
                      allocate(htopo(nx_lon,ny_lat))
                      ierr=nf_get_vara_real (ncid,varid, start,count,
     &                                                         htopo)
                      write(*,*)   '...done.'
                      if (ierr /= nf_noerr) then
                        write(*,'(/1x,4A/12x,A/)')       '### ERROR: ',
     &                                'Cannot read topography from ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                      endif

                      ishft=iwest-1
                      if (shft_west > 0 .and. shft_east > 0) then
                        do i=1,nx_lon,+1
                          xlon(i)=xlon(i+ishft) +360.D0
                        enddo
                      elseif (shft_west < 0 .and. shft_east < 0) then
                        do i=1,nx_lon,+1
                          xlon(i)=xlon(i+ishft) -360.D0
                        enddo
                      elseif (shft_west == 0 .and. shft_east == 0) then
                        do i=1,nx_lon,+1
                          xlon(i)=xlon(i+ishft)
                        enddo
                      else
                        write(*,'(/1x,3A/12x,2A/)')      '### ERROR: ',
     &                   'Algoritm failure finding proper 360-degree ',
     &                 'shift in longitude','when reading topography ',
     &                         'data as a single geographical region.'
                        stop
                      endif

                    elseif (iwest>ieast) then
                     write(*,'(/1x,3A)') 'Reading topography from ''',
     &                                           srcfile(1:lsrc),
     &                            ''' in two separate parts adjacent'
                      write(6,'(1x,A)',advance='no')
     &                    'through 360-degree periodicity, first...'
                      flush(unit=6)

                      nx_lon=ieast + period-iwest+1
                      allocate(htopo(nx_lon,ny_lat))
                      allocate(xtmp(nx_lon))

                      start(1)=1 ; count(1)=ieast
                      allocate(bffr(count(1),count(2)))
                      ierr=nf_get_vara_real (ncid,varid, start,count,
     &                                                         bffr)
                      if (ierr == nf_noerr) then
                        do j=1,count(2)
                          do i=1,count(1)
                            htopo(nx_lon-ieast+i,j)=bffr(i,j)
                          enddo
                        enddo
                        deallocate(bffr)

                        ishft=nx_lon-count(1)
                        if (shft_east > 0) then
                          do i=1,count(1)
                            xtmp(i+ishft)=xlon(i) +360.D0
                          enddo
                        elseif (shft_east < 0) then
                          do i=1,count(1)
                            xtmp(i+ishft)=xlon(i) -360.D0
                          enddo
                        else
                          do i=1,count(1)
                            xtmp(i+ishft)=xlon(i)
                          enddo
                        endif

                        write(6,'(A)',advance='no') 'second...'
                        flush(unit=6)

                        start(1)=iwest ; count(1)=period-iwest+1
                        allocate(bffr(count(1),count(2)))
                        ierr=nf_get_vara_real (ncid,varid, start,
     &                                               count, bffr)
                        if (ierr == nf_noerr) then
                          do j=1,count(2)
                            do i=1,count(1)
                              htopo(i,j)=bffr(i,j)
                            enddo
                          enddo
                          deallocate(bffr)

                          ishft=iwest-1
                          if (shft_west > 0) then
                            do i=1,count(1)
                              xtmp(i)=xlon(i+ishft) +360.D0
                            enddo
                          elseif (shft_west < 0) then
                            do i=1,count(1)
                              xtmp(i)=xlon(i+ishft) -360.D0
                            enddo
                          else
                            do i=1,count(1)
                              xtmp(i)=xlon(i+ishft)
                            enddo
                          endif
                          do i=1,nx_lon
                            xlon(i)=xtmp(i)
                          enddo
                          deallocate(xtmp)
                          write(6,'(A)')  'done reading.'
                          flush(unit=6)
                        else
                          write(*,'(/1x,4A/12x,A/)')     '### ERROR: ',
     &                           'Cannot read topography data from ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                          stop
                        endif
                      else
                        write(*,'(/1x,4A/12x,A/)')       '### ERROR: ',
     &                           'Cannot read topography data from ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                        stop
                      endif

! Check whether the patched together longitude is equidistant with some
! tollerance [merely to verify that there is no algorithmic failure].

                      dx=(xlon(nx_lon)-xlon(1))/dble(nx_lon-1)
                      epsil=0.D0
                      do i=2,nx_lon
                        ptest=abs(xlon(i)-xlon(i-1)-dx)
                        if (ptest > epsil) epsil=ptest
                      enddo
                      if (epsil > 0.01D0*dx) then
                        write(*,'(/1x,2A/12x,A,2x,A/12x,A,F12.8/)')
     &                 '### ERROR: Algorithm failure in constructing ',
     &                              'monotone longitudinal coordinate',
     &                    'when reading topography as two subregions.',
     &                                        'Maximum relative error',
     &                    'in uniformity of grid interval =', epsil/dx
                        stop
                      endif
                    endif
                    ierr=nf_close(ncid)
                  else
                    write(*,'(/1x,4A/12x,A/)')    '### ERROR: Cannot ',
     &                   'get netCDF ID for variable ''topo'' from ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,4A/12x,A/)')      '### ERROR: Cannot ',
     &                'read variable ''', lat_name(1:lenstr(lat_name)),
     &         ''' from ''', srcfile(1:lsrc), '''.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot get ',
     &        'netCDF variable ID for ''',lat_name(1:lenstr(lat_name)),
     &        ''' from  ''', srcfile(1:lsrc), '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot read ',
     &                     'variable ''', lon_name(1:lenstr(lon_name)),
     &         ''' from ''', srcfile(1:lsrc), '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,6A/12x,A/)') '### ERROR: Cannot get netCDF ',
     &              'variable ID for ''', lon_name(1:lenstr(lon_name)),
     &         ''' from ''', srcfile(1:lsrc), '''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,4A/)')    '### ERROR: Cannot determine sizes ',
     &             'of lon-lat dimensions for topography data file ''',
     &                                          srcfile(1:lsrc), '''.'
        endif
      else
        write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open netCDF ',
     &            'file ''', srcfile(1:lsrc), ''' in read-only mode.',
     &                                              nf_strerror(ierr)
      endif


! 3. Once both the input topography data and the target roms gridfile
! are successfully opened and the relevant data is read, do the actual
! interpolation:

      if (ierr == nf_noerr) then
        allocate (hraw(0:Lm+1,0:Mm+1)) ; ierr=0

C$OMP PARALLEL SHARED(nx_lon,ny_lat, xlon,ylat, htopo, Lm,Mm,
C$OMP&                        lon_r,lat_r, pm,pn, hraw, radius, ierr)
        call compute_hraw_thread(nx_lon,ny_lat,  xlon,ylat,  htopo,
     &                  Lm,Mm, lon_r,lat_r, pm,pn, hraw, radius, ierr)
C$OMP END PARALLEL
        if (ierr==0) call write_hraw(grid, srcfile,radius, Lm,Mm,hraw)
      endif
      end
