! Replace land values of data "qfld" with values extrapolated from
! the sea. The extrapolation rule adopted here is a nearest-point rule
! modified by inclusion of gradient of the field one-and-half point
! away from the extrapolated point.  In the case when there is a single
! missing point surrounded by two rows of valid points on each side,
! this procedure is equivalent to cubic interpolation onto the location
! missing point.  In this case the domain of dependency contains 20
! surrounding points (5x5 stencil, excluding corners an the point
! itself).  The procedure is iterative and normally goes through as
! many steps as necessary to fill in the land entirely.  CPP switch
! SINGLE_STEP_ONLY terminates it just after one step, so the progress
! of filling in can be visualized for testing purposes.

#define DIAGONAL
c--#define VERBOSE
c--#define SINGLE_STEP_ONLY

#define WESTERN_EDGE istr==1
#define EASTERN_EDGE iend==nx
#define SOUTHERN_EDGE jstr==1
#define NORTHERN_EDGE jend==ny

! Special value "spv" is for logical checks, while "spv_set" is to
! flag that the value of the field at that point is masked. Both of
! them must be large NEGATIVE values, and "spv" should always be
! smaller by absolute value than "spv_set".  For some reason the
! logic below in sensitive to setting of special value "spv": while
! _FillValue = -9.99e+33f is used in SODA files, setting "spv" to
! -9.99E+31 makes "extend_on_land" not work, while -9.9E+31
! is still working.


c#define spv_set -1.E+33
c#define spv -9.9E+30
c-#define spv -9.9E+31 /* works */
c-#define spv -9.99E+31 /*does not */

#define spv -9.9E+9
#define spv_set -1.D+10

      module ext_on_land_vars
       integer, save :: msk_count=0, alloc_wrk_size=0
       real(kind=4), allocatable, dimension(:) :: wrk1,wrk2
       integer, save :: alloc_FX_size=0
C$OMP THREADPRIVATE(alloc_FX_size)
       real(kind=4), allocatable, dimension(:):: FX,FY
C$OMP THREADPRIVATE(FX,FY)
      end module ext_on_land_vars

                                                 ! Note: this driver
      subroutine extend_on_land(nx,ny, qfld)     ! routine is used only
      implicit none                              ! by older versions of
      integer nx,ny                              ! forcing programs.
      integer(kind=2) mask(nx,ny)                ! New ones directly
      real(kind=4) qfld(nx,ny)                   ! call its thread.
      mask(:,:)=1  !<-- backward compatibility   
C$OMP PARALLEL SHARED(nx,ny, mask, qfld)
      call extend_on_land_thread(nx,ny, mask, qfld)
C$OMP END PARALLEL
      end


      subroutine extend_on_land_thread(nx,ny, mask, qfld)
      use ext_on_land_vars
      implicit none
      integer nx,ny
      integer(kind=2) mask(nx,ny)
      real(kind=4) qfld(nx,ny)

      integer nsub_x,nsub_y, isize,jsize, istr,iend,jstr,jend,
     &       trd, ntrds,ntls, tile, my_first,my_last, inc_count
#ifndef SINGLE_STEP_ONLY
     &                                     , iter, msk_bak_cnt
# ifdef VERBOSE
      integer lenstr
      character(len=16) msk_str, inc_str
# endif
#endif
C$    integer omp_get_thread_num, omp_get_num_threads
      ntrds=1 ; trd=0
C$    ntrds=omp_get_num_threads() ; trd=omp_get_thread_num()

      if (alloc_wrk_size < (nx+2)*(ny+2)) then
C$OMP BARRIER
        if (trd == 0) then
          alloc_wrk_size=(nx+2)*(ny+2)
          if (allocated(wrk1)) deallocate(wrk1,wrk2)
          allocate(wrk1(alloc_wrk_size),wrk2(alloc_wrk_size))
          write(*,'(1x,2A,F8.3,1x,A)') 'extend_on_land_thread: ',
     &        'allocated', float(2*alloc_wrk_size)/float(262144),
     &                              'MBytes of shared workspace'
          write(*,*) 'spv, spv_set =', spv, spv_set
        endif
C$OMP BARRIER
      endif

      call set_tiles(nx,ny, nsub_x,nsub_y)

      isize=(nx+nsub_x-1)/nsub_x ; jsize=(ny+nsub_y-1)/nsub_y
      if (alloc_FX_size < (isize+4)*(jsize+4)) then
        alloc_FX_size=(isize+4)*(jsize+4)
        if (allocated(FX)) deallocate(FX,FY)
        allocate(FX(alloc_FX_size),FY(alloc_FX_size))
C$OMP CRITICAL(ext_wr_rgn)
        write(*,'(1x,2A,F8.4,1x,A,I3)')  'extend_on_land_thread: ',
     &           'allocated', float(2*alloc_FX_size)/float(262144),
     &                   'MBytes of private workspace, trd =', trd
#ifdef VERBOSE
        write(*,'(4x,2(A,I5,1x),2(1x,A,I4),1x,2(1x,A,I4))')
     &  'nx =',nx, 'ny =',ny, 'nsub_x =',nsub_x, 'nsub_y =',nsub_y,
     &                         'isize =',isize,   'jsize =',jsize
#endif
C$OMP END CRITICAL(ext_wr_rgn)
      endif

      ntls=(nsub_x*nsub_y+ntrds-1)/ntrds
      my_first=trd*ntls -(ntls*ntrds-nsub_x*nsub_y)/2
      my_last=min(my_first+ntls-1, nsub_x*nsub_y-1)
      my_first=max(my_first, 0)

      do tile=my_first,my_last,+1
        call comp_tile_bounds(tile, nx,ny, nsub_x,nsub_y,
     &                                 istr,iend,jstr,jend)
        call copy_ext_array_tile(istr,iend,jstr,jend, nx,ny,
     &                                    mask, qfld, wrk1)
      enddo
      if (trd == 0) msk_count=0  !<-- shared variable
C$OMP BARRIER

! NOTE: The termination mechanism for while loop below is when the
! global counter of filled in points "msk_count" (a shared variable)
! stops growing.  So its previous value is saved at the beginning of
! each iteration, after which the number of newly filled points is
! added in.   Because "msk_count" is a shared variable, it requires
! a BARRIER between it is being written into and is read from (i.e.,
! write events must have BARRIERS both BEFORE and AFTER to eliminate
! the possibility of reading its value within the same synchronization
! interval). In the code below it is achieved by combining increments
! of two successive "ext_on_land_tile" steps, so an additional BARRIER
! needed solely to synchronize "msk_count" is avoided, thus having
! only one BARRIER per each "ext_on_land_tile" step.  This design
! choice however leaves a possibility of one extra passive step
! before the condition to terminate iterations is detected.

#ifndef SINGLE_STEP_ONLY
      msk_bak_cnt=msk_count-1  !<-- to make while go at least once
      iter=0
      do while(msk_count > msk_bak_cnt)
        msk_bak_cnt=msk_count ; iter=iter+1
# ifdef VERBOSE
C$OMP CRITICAL(ext_wr_rgn)
        write(msk_str,*) msk_count ; write(inc_str,*) inc_count
        write(*,'(8x,I5,2(2x,A,1x,A),2x,A,I3)')  iter,
     &               'msk_count =', msk_str(1:lenstr(msk_str)),
     &        'inc =', inc_str(1:lenstr(inc_str)), 'trd =', trd
C$OMP END CRITICAL(ext_wr_rgn)
# endif
#endif
        inc_count=0
        do tile=my_last,my_first-1
          call comp_tile_bounds(tile, nx,ny, nsub_x,nsub_y,
     &                                  istr,iend,jstr,jend)
          call ext_on_land_tile(istr,iend,jstr,jend, nx,ny,
     &                          wrk1,wrk2, inc_count, FX,FY)
        enddo
#ifndef SINGLE_STEP_ONLY
C$OMP BARRIER

        do tile=my_first,my_last,+1
          call comp_tile_bounds(tile, nx,ny, nsub_x,nsub_y,
     &                                  istr,iend,jstr,jend)
          call ext_on_land_tile(istr,iend,jstr,jend, nx,ny,
     &                          wrk2,wrk1, inc_count, FX,FY)
        enddo
#endif
        if (inc_count > 0) then
C$OMP CRITICAL(ext_msk_rgn)
          msk_count=msk_count+inc_count
C$OMP END CRITICAL(ext_msk_rgn)
        endif
C$OMP BARRIER
#ifndef SINGLE_STEP_ONLY
      enddo
#endif

      do tile=my_first,my_last,+1
        call comp_tile_bounds(tile, nx,ny, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)
        call copy_back_tile(istr,iend,jstr,jend, nx,ny,
#ifndef SINGLE_STEP_ONLY
     &                                         wrk1,qfld)
#else
     &                                         wrk2,qfld)
#endif
      enddo
c**   if (trd == 0) write(*,*) 'leaving ext_on_land, iter =',iter
C$OMP BARRIER
      end


      subroutine copy_ext_array_tile(istr,iend,jstr,jend, nx,ny,
     &                                            mask, src,targ)

! Copy array "src" into extended array with one row of ghost points
! all around, while setting land points to special values.  The logic
! here is designed to work both ways: either there a non-trivial land
! mask array, or masked data is already set to some special value,
! while mask(i,j) is trivially set to all-water mask(:,:)=1 status
! and makes no effect.

      implicit none
      integer istr,iend,jstr,jend, nx,ny, i,j
      integer(kind=2) mask(nx,ny)
      real(kind=4) src(nx,ny), targ(0:nx+1,0:ny+1)
      do j=jstr,jend
        do i=istr,iend
          if (mask(i,j) > 0 .and. abs(src(i,j)) < abs(spv)) then
            targ(i,j)=src(i,j)
          else
            targ(i,j)=spv_set
          endif
        enddo
      enddo
      if (WESTERN_EDGE) then
        do j=jstr,jend
          targ(istr-1,j)=targ(istr,j)
        enddo
      endif
      if (EASTERN_EDGE) then
        do j=jstr,jend
          targ(iend+1,j)=targ(iend,j)
        enddo
      endif
      if (SOUTHERN_EDGE) then
        do i=istr,iend
          targ(i,jstr-1)=targ(i,jstr)
        enddo
      endif
      if (NORTHERN_EDGE) then
        do i=istr,iend
          targ(i,jend+1)=targ(i,jend)
        enddo
      endif
      if (WESTERN_EDGE .and. SOUTHERN_EDGE) then
        targ(istr-1,jstr-1)=targ(istr,jstr)
      endif
      if (WESTERN_EDGE .and. NORTHERN_EDGE) then
        targ(istr-1,jend+1)=targ(istr,jend)
      endif
      if (EASTERN_EDGE .and. SOUTHERN_EDGE) then
        targ(iend+1,jstr-1)=targ(iend,jstr)
      endif
      if (EASTERN_EDGE .and. NORTHERN_EDGE) then
        targ(iend+1,jend+1)=targ(iend,jend)
      endif
      end


      subroutine copy_back_tile(istr,iend,jstr,jend, nx,ny, wrk,qfld)
      implicit none
      integer istr,iend,jstr,jend, nx,ny, i,j
      real(kind=4) qfld(nx,ny), wrk(0:nx+1,0:ny+1)
      do j=jstr,jend
        do i=istr,iend
#ifndef SHOW_PROGRESS
          qfld(i,j)=wrk(i,j)
#else
          if (wrk(i,j,nnew) > spv) then
            qfld(i,j)=wrk(i,j) ; land(i,j)=0.
          else
            qfld(i,j)=0.       ; land(i,j)=1.
          endif
#endif
        enddo
      enddo
      end


! Perform the actual filling of masked data.  The algorithm is similar
! to the nearest-point extrapolation, however, combined with attempting
! to do gradient extrapolation, if enough points behind are available.
! The code below also increments counter "inc_count" by the number of
! points actually being filled within the tile. This counter is private
! to each thread variable and is expected to be initialized at entry.

! Setting parameter "threshold" controls which points are filled up
! during the current sweep, and which are delayed until the next one.
! Depending of the shape of surrounding land, possible values of
! threshold = 0.7, 0.9, 1.1, 1.7, 2.4, 2.7, 3.4, 4.1, ... and up to
! 6.8;  Straight fronts oriented at 90-degrees need 2.4 or less to
! propagate, while 45-degree fronts are able to move for up to 2.7.
! Values above 3 correspond to filling convex corners only, while below
! 2 to concave bays, so that threshold 2 results in deferring
! extrapolation of data from bays to surrounding land, while waiting
! until the data is filled up from the sides.  The least restrictive
! setting of threshold 0.9 just prohibits uni-directional diagonal
! extrapolation, while keeping all other possibilities.


      subroutine ext_on_land_tile(istr,iend,jstr,jend, nx,ny, src,targ,
     &                                               inc_count, FX,FY)
      implicit none
      integer  istr,iend,jstr,jend, nx,ny, inc_count
      real(kind=4), dimension(0:nx+1,0:ny+1) :: src,targ
      real(kind=4), dimension(istr-2:iend+2,jstr-2:jend+2) :: FX,FY

      integer imin,imax,jmin,jmax, i,j, indx, ncrs
      real(kind=4) vlu, weight(0:24), invwgt(24)
      real(kind=4), parameter :: corn=0.707106781186547,
     &                           grad=0.333333333333333,

c     &        threshold=1.5
     &   threshold=2.4

      ncrs=0
      imin=max(istr-1,1); imax=min(iend+2,nx+1) ! Restricted extended
      jmin=max(jstr-1,1); jmax=min(jend+2,ny+1) ! ranges for FX and FE

      do j=jstr-1,jend+1
        do i=imin,imax
          if (src(i,j) > spv .and. src(i-1,j) > spv) then
            FX(i,j)=grad*(src(i,j)-src(i-1,j))
          else
            FX(i,j)=0.
          endif
        enddo
      enddo
      if (WESTERN_EDGE) then
        do j=jstr-1,jend+1
          FX(istr-1,j)=FX(istr,j)
        enddo
      endif
      if (EASTERN_EDGE) then
        do j=jstr-1,jend+1
          FX(iend+2,j)=FX(iend+1,j)
        enddo
      endif

      do j=jmin,jmax
        do i=istr-1,iend+1
          if (src(i,j) > spv .and. src(i,j-1) > spv) then
            FY(i,j)=grad*(src(i,j)-src(i,j-1))
          else
            FY(i,j)=0.
          endif
        enddo
      enddo
      if (SOUTHERN_EDGE) then
        do i=istr-1,iend+1
          FY(i,jstr-1)=FY(i,jstr)
        enddo
      endif
      if (NORTHERN_EDGE) then
        do i=istr-1,iend+1
          FY(i,jend+2)=FY(i,jend+1)
        enddo
      endif

      weight(0)=0. !<-- to simplify logic
      do j=0,4     !<-- number of diagonal neighbors
        do i=0,4    !<-- number of straight neighbors
          if (i+j > 0) then
            weight(i+5*j)=float(i)+corn*float(j)
            invwgt(i+5*j)=1./weight(i+5*j)
          endif
        enddo
      enddo

      do j=jstr,jend
        do i=istr,iend
          if (src(i,j) > spv) then
            targ(i,j)=src(i,j)
          else
            indx=0 ; vlu=0.
#ifdef CIRC
! Check whether the point (i,j) belongs to a strip of land separating
! two unconnected bodies of water: go counterclockwise around it and
! count transitions from land to water: if more than one transition
! occur then this is separatrix; one might defer wetting this point.

            ncrs=0
            if (src(i+1,j) < spv .and. src(i+1,j+1) > spv) ncrs=ncrs+1
            if (src(i+1,j+1) < spv .and. src(i,j+1) > spv) ncrs=ncrs+1
            if (src(i,j+1) < spv .and. src(i-1,j+1) > spv) ncrs=ncrs+1
            if (src(i-1,j+1) < spv .and. src(i-1,j) > spv) ncrs=ncrs+1
            if (src(i-1,j) < spv .and. src(i-1,j-1) > spv) ncrs=ncrs+1
            if (src(i-1,j-1) < spv .and. src(i,j-1) > spv) ncrs=ncrs+1
            if (src(i,j-1) < spv .and. src(i+1,j-1) > spv) ncrs=ncrs+1
            if (src(i+1,j-1) < spv .and. src(i+1,j) > spv) ncrs=ncrs+1
#endif

            if (src(i-1,j) > spv) then
              indx=indx+1 ;  vlu=vlu + src(i-1,j)+FX(i-1,j)
            endif
            if (src(i+1,j) > spv) then
              indx=indx+1 ;  vlu=vlu + src(i+1,j)-FX(i+2,j)
            endif
            if (src(i,j-1) > spv) then
              indx=indx+1 ;  vlu=vlu + src(i,j-1)+FY(i,j-1)
            endif
            if (src(i,j+1) > spv) then
              indx=indx+1 ;  vlu=vlu + src(i,j+1)-FY(i,j+2)
            endif
#ifdef DIAGONAL
            if (src(i-1,j-1) > spv) then
              indx=indx+5  ;   vlu=vlu + corn*( src(i-1,j-1)
     &                       +0.5*(FX(i-1,j-1)+FY(i-1,j-1)) )
            endif
            if (src(i-1,j+1) > spv) then
              indx=indx+5  ;  vlu=vlu + corn*( src(i-1,j+1)
     &                       +0.5*(FX(i-1,j+1)-FY(i-1,j+2)) )
            endif
            if (src(i+1,j-1) > spv) then
              indx=indx+5  ;   vlu=vlu + corn*( src(i+1,j-1)
     &                       +0.5*(FY(i+1,j-1)-FX(i+2,j-1)) )
            endif
            if (src(i+1,j+1) > spv) then
              indx=indx+5  ;   vlu=vlu + corn*( src(i+1,j+1)
     &                       -0.5*(FX(i+2,j+1)+FY(i+1,j+2)) )
            endif
#endif
#ifdef CIRC
            if (weight(indx) > threshold .and. ncrs < 2) then
#else
            if (weight(indx) > threshold) then
#endif
              inc_count=inc_count+1
              targ(i,j)=vlu*invwgt(indx)
            else
              targ(i,j)=spv_set
            endif
          endif
        enddo
      enddo

      if (WESTERN_EDGE) then
        do j=jstr,jend
          targ(istr-1,j)=targ(istr,j)
        enddo
      endif
      if (EASTERN_EDGE) then
        do j=jstr,jend
          targ(iend+1,j)=targ(iend,j)
        enddo
      endif
      if (SOUTHERN_EDGE) then
        do i=istr,iend
          targ(i,jstr-1)=targ(i,jstr)
        enddo
      endif
      if (NORTHERN_EDGE) then
        do i=istr,iend
          targ(i,jend+1)=targ(i,jend)
        enddo
      endif
      if (WESTERN_EDGE .and. SOUTHERN_EDGE) then
        targ(istr-1,jstr-1)=targ(istr,jstr)
      endif
      if (WESTERN_EDGE .and. NORTHERN_EDGE) then
        targ(istr-1,jend+1)=targ(istr,jend)
      endif
      if (EASTERN_EDGE .and. SOUTHERN_EDGE) then
        targ(iend+1,jstr-1)=targ(iend,jstr)
      endif
      if (EASTERN_EDGE .and. NORTHERN_EDGE) then
        targ(iend+1,jend+1)=targ(iend,jend)
      endif
      end
