      program gebco 

! Purpose: Read "raw" topography from specified data file in GEBCO
!          netCDF format and interpolate or coarsen it onto ROMS model
!          grid. The outcome is written as netCDF variable "hraw".
!
! NOTE: (1) the content of this file is just a driver: the actual
!           interpolation/coarsening routine is in "compute_hraw.F".

!       (2) nothing is done here about setting mask, which needs to be
!           processed separately via "rmask.F". This is done to avoid
!           accidental overwriting of land mask field "rmask" which
!           may be subject to elaborate hand editing.

!       (3) The specific features of GEBCO netCDF format are:

!       (i) there are no x,y dimensions in the usual netCDF way;
!           the topography data is written as a large 1D array of
!           nx*ny lenght;

!      (ii) there are no coordinate variables, instead the gridding
!           information is stored as four pairs of numbers
!                          dimension = (nx,ny);
!                            spacing = (dx,dy);
!                            x_range = (min,max longitude);
!                            y_range = (min,max latitude);  

!     (iii) the first point of topography data is at the north-west
!           corner of the dataset, which means that logical x-axis
!           (index "i" in would-be 2D array) is east-to-west, while
!           logical y-axes (index "j") is north-to-south.

!      (iv) the data values are considered to be cell-averages, which
!           means that, for example, if the specified x_range is xmin
!           to xmax, then discrete coordinates of data values are
!
!                         x(i)=xmin+dx*(i-0.5) for i=1,nx; 
!           thus
!                        xmin+dx/2  <= x(i) <= xmax-dx/2
!
!           with dx=(xmax-xmin)/nx always holds.

! Created and maintained by Alexander Shchepetkin old_galaxy@yahoo.com

      use roms_grid_vars
      use comm_vars_hraw

      implicit none
      real(kind=8) radius,  dx,dy, ptest, epsil
      real(kind=8), allocatable :: xlon(:),ylat(:),xtmp(:),hraw(:,:)
      real(kind=4), allocatable :: htopo(:,:)
      integer(kind=2), allocatable :: bffr(:,:), tmp(:) 
      integer nargs, Lm,Mm, lon,lat, iwest,ieast,jsouth,jnorth, period,
     &     shft_west,shft_east, ishft,jshft, nx_lon,ny_lat, nxpts, i,j,
     &     ncid, varid, dim(2),  ierr, lsrc, lgrd, lenstr, indx_bound
      character(len=80) srcfile, grid
      character(len=16) str

#include "netcdf.inc"

      nargs=iargc()
      if (nargs==3) then
        call getarg(1,str)     ; read(str,*) radius
        call getarg(2,srcfile) ; lsrc=lenstr(srcfile)
        call getarg(3,grid)    ; lgrd=lenstr(grid)

        write(*,'(/1x,5A/1x,A,F8.5/)') 'Processing ''',srcfile(1:lsrc),
     &   ''' --> ''', grid(1:lgrd), '''', 'smoothing radius =', radius

      else
        write(*,'(/1x,A/ /8x,A/)')   'usage of "gebco" should be:',
     &                  'gebco smoothing_radius source.nc grid.nc'
        stop
      endif

      lon=0 ; shft_west=0  ! RESET EVERYTHING
      lat=0 ; shft_east=0

! 1. Open and read ROMS grid file. Find geographical limits of the grid.

      call read_roms_grid(grid, Lm,Mm)
      call roms_grid_geo_bounds(lon_r,lat_r, Lm,Mm, radius)

      write(*,'(1x,2A,2F16.8/22x,A,2F16.8/)')  'roms grid extremes: ',
     &   'longitude:', lon_min, lon_max, 'latitude:', lat_min,lat_max

! 2. Open and read topography data file....  

      ierr=nf_open (srcfile, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        ierr=nf_inq_varid (ncid, 'dimension', varid)
        if (ierr == nf_noerr) then
          ierr=nf_get_var_int (ncid, varid, dim)
          if (ierr == nf_noerr) then
            lon=dim(1) ; lat=dim(2)
          else
            write(*,'(/1x,4A/12x,A/)')       '### ERROR: Cannot read ',
     &              'variable ''dimension'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
          endif 
        else
          write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot get netCDF ',
     &       'variable ID for ''dimension'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
        endif

        write(*,'(1x,A,I6,2x,A,I6)') 'Found dataset dimensions',
     &                                              lon, 'x', lat

        if (lon>0 .and. lat>0) then
          allocate(xlon(lon)) ; allocate(ylat(lat))

          ierr=nf_inq_varid (ncid, 'spacing', varid)
          if (ierr == nf_noerr) then
            ierr=nf_get_var_double (ncid, varid, xlon)
            if (ierr == nf_noerr) then
              dx=xlon(1) ; dy=xlon(2)
              ierr=nf_inq_varid (ncid, 'x_range', varid)
              if (ierr == nf_noerr) then
                ierr=nf_get_var_double (ncid, varid, xlon)
                if (ierr == nf_noerr) then
                  if ( abs(dx - (xlon(2)-xlon(1))/dble(lon))
     &                                     < 0.001*dx ) then 
                    epsil=xlon(1)-0.5D0*dx
                    do i=1,lon
                      xlon(i)=epsil +dx*dble(i)
                    enddo
                  else
                    write(*,*) '### ERROR: longitude spacing missmatch'
                    ierr=nf_noerr-1
                  endif
                else
                  write(*,'(/1x,4A/12x,A/)')      '### ERROR: Cannot ',
     &                            'read variable ''x_range'' from  ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,4A/12x,A/)')    '### ERROR: Cannot get ',
     &                   'netCDF variable ID for ''x_range'' from  ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
              endif
              if (ierr == nf_noerr) then
                ierr=nf_inq_varid (ncid, 'y_range', varid)
                if (ierr == nf_noerr) then
                  ierr=nf_get_var_double (ncid, varid, ylat)
                  if (ierr == nf_noerr) then
                    if ( abs(dy - (ylat(2)-ylat(1))/dble(lat)) 
     &                                       < 0.001*dy ) then  
                      epsil=ylat(1)+0.5D0*dy
                      do i=1,lat
                        ylat(i)=epsil+dy*dble(i)
                      enddo
                    else
                      write(*,*) '### ERROR: latitude spacing missmatch'
                      ierr=nf_noerr-1
                    endif
                  else
                    write(*,'(/1x,4A/12x,A/)')    '### ERROR: Cannot ',
     &                            'read variable ''y_range'' from  ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot get ',
     &                   'netCDF variable ID for ''y_range'' from  ''',
     &                       srcfile(1:lsrc), '''.', nf_strerror(ierr)
                endif
              endif
            else
              write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot read ',
     &                'variable ''spacing'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot get netCDF ',
     &         'variable ID for ''spacing'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
          endif

          if (ierr == nf_noerr) then
            ierr=nf_inq_varid (ncid, 'z', varid)
            if (ierr == nf_noerr) then
              allocate(bffr(lon,lat))
              ierr=nf_get_var_int2 (ncid, varid, bffr)
              if (ierr == nf_noerr) then

                allocate(tmp(lon))             ! Because GEBCO dataset
                do j=1,lat/2                   ! storage order has its
                  jshft=lat-j+1                ! logical y-axis (index
                  do i=1,lon                   ! "j" of would be 2D
                    tmp(i)=bffr(i,j)           ! array of data) going
                    bffr(i,j)=bffr(i,jshft)    ! from north to south,
                    bffr(i,jshft)=tmp(i)       ! this segment reverses
                  enddo                        ! the "j" direction.
                enddo
                deallocate(tmp)

                write(*,*) 'retrieved topography data from ''',
     &                                 srcfile(1:lsrc), '''.'
              else
                write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot read ',
     &                      'variable ''z'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,4A/12x,A/)')      '### ERROR: Cannot get ',
     &        'netCDF variable ID for ''z'' from  ''', srcfile(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
            endif
          endif
          ierr=nf_close (ncid)

          if (allocated(bffr)) then

! The overall logic is similar to that in "etopo.F", except that here
! all the data from netCDF file is already read into "bffr" and instead
! reading relevant rectangular regions from the file the data is simply
! copied from the buffer array.  

! Determine bounding indices for the relevant portion of topographic
! data (this is the only part of data which will be actually extracted
! from the data set. Compute dimensions of array to hold this portion
! of data, and move the relevant parts of coordinate arrays to the
! beginning. Then allocate the array and read the data.

            jsouth=indx_bound(ylat,lat, lat_min)
            jnorth=indx_bound(ylat,lat, lat_max)

            if ( 0 < jsouth .and. jsouth < lat .and.
     &           0 < jnorth .and. jnorth < lat ) then
              if (jsouth > 1) jsouth=jsouth-1
              jnorth=jnorth+1
            else
              write(*,'(/1x,2A,2F12.6/12x,A/)')  '### ERROR: North-',
     &          'south extents of the dataset',  ylat(1),  ylat(lat),
     &          'are not sufficient to cover the entire model grid.'
              stop
            endif

            iwest=indx_bound(xlon,lon, lon_min)
            ieast=indx_bound(xlon,lon, lon_max)

            if ( 0 < iwest .and. iwest < lon .and. 
     &           0 < ieast .and. ieast < lon ) then
              if (iwest > 1) iwest=iwest-1
              ieast=ieast+1

              write(*,'(1x,2A,2I5,2x,A,2F12.6)')   'Single-region ',
     &                'dataset, iwest,ieast =',  iwest, ieast,
     &                'xlon(iwest:ieast) =', xlon(iwest),xlon(ieast)
            else

! Model grid does not fit directly into the east-west extents of the
! dataset. Check whether the data set longitude covers the entire 360
! degrees, so it will be possible to treat is as periodic.
! Two variants are of datasets are known to exist: either the end
! points may repeated,
!
!                      x(1)+360 = x(lon)
! or may be not,
!                      x(1)+360 = x(lon)+dx.
!
! Identify which is the case and find periodicity range accordingly.

              ptest=xlon(lon)-xlon(1)-360.D0
              dx=(xlon(lon)-xlon(1))/dble(lon-1)
              epsil=0.01D0*abs(dx)
              if (abs(ptest) < epsil) then
                period=lon-1
              elseif (abs(ptest+dx) < epsil) then
                period=lon
              else
                period=0
              endif

              if (period>0) then
                write(*,'(/1x,2A,I6,1x,A,I6/)')    'Identified ',
     &                  'periodicity domain in data of ', period,
     &                                      'points out of', lon
              else
                write(*,'(/1x,2A/)') '### ERROR: The data does ',
     &                               'not cover the entire grid.'
                stop
              endif

              shft_west=0
              if (iwest == 0) then
                shft_west=-1
                iwest=indx_bound(xlon,lon, lon_min+360.D0)
              elseif (iwest == lon) then
                shft_west=+1
                iwest=indx_bound(xlon,lon, lon_min-360.D0)
              endif

              shft_east=0
              if (ieast == 0) then
                shft_east=-1 
                ieast=indx_bound(xlon,lon, lon_max+360.D0)
              elseif (ieast == lon) then
                shft_east=+1
                ieast=indx_bound(xlon,lon, lon_max-360.D0)
              endif

              if ( 0 < iwest .and. iwest < lon .and.
     &             0 < ieast .and. ieast < lon ) then
                if (iwest > 1  ) iwest=iwest-1
                if (ieast < lon) ieast=ieast+1
              else
                write(*,'(/1x,2A/12x,A/)') '### ERROR: Longitude of ',
     &                'dataset covers 360 degrees, but still cannot ',
     &                           'find  starting and ending indices.'
                stop
              endif
            endif !<--  0 < iwest < lon .and. 0 < ieast < lon

            write(*,'(/1x,3A/2(1x,A,I5,I6,1x,A,I6))')   'Bounding ',
     &              'indices of the relevant part to be extracted ',
     &                                   'from the entire dataset:',
     &              'iwest,ieast =',  iwest,ieast,   'out of', lon,
     &              'jsouth,jnorth =',jsouth,jnorth, 'out of', lat

            ny_lat=jnorth-jsouth+1
            jshft=jsouth-1
            do j=1,ny_lat,+1
              ylat(j)=ylat(j+jshft)
            enddo

! Once we made it up to here, there are two possibilities: either the
! topographic data covering the entire model grid can be retrieved as
! a single rectangular subregion within the dataset, or it has to be
! pieced together from two parts using 360-degree periodicity.

            if (iwest < ieast) then 
              nx_lon=ieast-iwest+1
              allocate(htopo(nx_lon,ny_lat))

              ishft=iwest-1
              do j=1,ny_lat
                do i=1,nx_lon
                  htopo(i,j)=bffr(i+ishft,j+jsouth-1)
                enddo
              enddo
              if (shft_west > 0 .and. shft_east > 0) then
                do i=1,nx_lon,+1
                  xlon(i)=xlon(i+ishft) +360.D0
                enddo
              elseif (shft_west < 0 .and. shft_east < 0) then
                do i=1,nx_lon,+1
                  xlon(i)=xlon(i+ishft) -360.D0
                enddo
              elseif (shft_west == 0 .and. shft_east == 0) then
                do i=1,nx_lon,+1
                  xlon(i)=xlon(i+ishft)
                enddo
              else
                write(*,'(/1x,3A/12x,2A/)')    '### ERROR: Algoritm ',
     &               'failure of finding proper 360-degree shift in ',
     &               'longitude',  'when reading topography data as ',
     &                                'a single geographical region.'
                stop
              endif

            elseif (iwest>ieast) then
              write(*,'(/1x,3A)') 'Constructing topography from two ',
     &                      'separate pieces adjacent to each other ',
     &                            'through 360-degree periodicity...'

              nx_lon=ieast + period-iwest+1
              allocate(htopo(nx_lon,ny_lat))
              allocate(xtmp(nx_lon))

! The first sub-region...

              ishft=nx_lon-ieast
              do j=1,ny_lat
                do i=1,ieast
                  htopo(i+ishft,j)=bffr(i,j+jshft)
                enddo
              enddo
              if (shft_east > 0) then
                do i=1,ieast
                  xtmp(i+ishft)=xlon(i) +360.D0
                enddo
              elseif (shft_east < 0) then
                do i=1,ieast
                  xtmp(i+ishft)=xlon(i) -360.D0
                enddo
              else
                do i=1,ieast
                  xtmp(i+ishft)=xlon(i)
                enddo
              endif

! The second sub-region...

              ishft=iwest-1
              nxpts=period-ishft
              do j=1,ny_lat
                do i=1,nxpts
                  htopo(i,j)=bffr(i+ishft,j+jshft)
                enddo
              enddo

              if (shft_west > 0) then
                do i=1,nxpts
                  xtmp(i)=xlon(i+ishft) +360.D0
                enddo
              elseif (shft_west < 0) then
                do i=1,nxpts
                  xtmp(i)=xlon(i+ishft) -360.D0
                enddo
              else
                do i=1,nxpts
                  xtmp(i)=xlon(i+ishft)
                enddo
              endif
              do i=1,nx_lon
                xlon(i)=xtmp(i)
              enddo
              deallocate(xtmp)

! Check whether the patched together longitude is equidistant with some
! tollerance [merely to verify that there is no algorithmic failure].
                      
              dx=(xlon(nx_lon)-xlon(1))/dble(nx_lon-1)
              epsil=0.D0
              do i=2,nx_lon
                ptest=abs(xlon(i)-xlon(i-1)-dx)
                if (ptest > epsil) epsil=ptest
              enddo
              if (epsil > 0.01D0*dx) then
                write(*,'(/1x,2A/12x,A,2x,A/12x,A,F12.8/)')
     &                 '### ERROR: Algorithm failure in constructing ',
     &                              'monotone longitudinal coordinate',
     &                    'when reading topography as two subregions.',
     &                                        'Maximum relative error', 
     &                 'in uniformity of grid interval =', epsil/dx
                stop
              endif
            endif     !<-- iwest<ieast
          endif     !<-- allocated(bffr)
        else
          write(*,'(/1x,4A/)')    '### ERROR: Cannot determine sizes ',
     &             'of lon-lat dimensions for topography data file ''',
     &                                          srcfile(1:lsrc), '''.'
        endif
      else
        write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open netCDF ',
     &            'file ''', srcfile(1:lsrc), ''' in read-only mode.',
     &                                              nf_strerror(ierr)
      endif
      deallocate(bffr)


! 3. Once both the input topography data and the target roms grid
! file are successfully opened and the relevant data is read, do the
! actual interpolation: 

      if (ierr == nf_noerr .and. allocated(htopo)) then
        allocate (hraw(0:Lm+1,0:Mm+1)) ; ierr=0

C$OMP PARALLEL SHARED(nx_lon,ny_lat, xlon,ylat, htopo, Lm,Mm,
C$OMP&                        lon_r,lat_r, pm,pn, hraw, radius, ierr)
        call compute_hraw_thread(nx_lon,ny_lat,  xlon,ylat,  htopo,
     &                  Lm,Mm, lon_r,lat_r, pm,pn, hraw, radius, ierr)
C$OMP END PARALLEL
        if (ierr==0) call write_hraw(grid, srcfile,radius, Lm,Mm,hraw)
      endif
      end
