      program merge_coast_topo

! Read data for points of irregular measurements of bathymetry, then
! read coastline data, and, assuming that depth goes to zero at the
! coast, prepare combined dataset for interpolating topography onto
! regular grid. This is needed to address the following problem: if
! using points of measured topography only, griddata (or any other
! interpolation procedure) knows nothing about the fact that depth
! goes to zero at the coast, so at best it tries to extrapolate the
! measured data toward the coast resulting in erroneous (sometimes
! negative) values there. Adding points at the coast with zero
! depths resolves the issue.

! In addition to the above, this program checks whether some of
! the measured points are too close to each other and potentially
! contain contradictory data.  Average these points, if too close
! to each other.

      implicit none
      character(len=64) bathymetry, coastline, str, merged_bath_coast
      real(kind=8), allocatable :: x(:),y(:),z(:), bffr(:,:)
      real(kind=8) xmin,xmax,ymin,ymax, xc,yc,csA, cffX,cffY, rmin,r2
      integer nargs, max_pts, npts, ierr, i,j, count,
     &                                          lbth,lcst,lmrg,lenstr
      real(kind=8), parameter  ::  pi=3.14159265358979323846D0,
     &   Eradius=6371315.D0, deg2rad=pi/180.D0, rad2deg=180.D0/pi

      nargs=iargc(); lbth=0 ; lcst=0; rmin=0.; lmrg=0
      if (nargs == 3 .or. nargs == 4) then
        call getarg(1,bathymetry) ; lbth=lenstr(bathymetry)
        if (nargs == 4) then
          call getarg(2,coastline) ; lcst=lenstr(coastline)
        endif
        call getarg(nargs-1,str)  ; read(str,*,iostat=ierr) rmin
        call getarg(nargs,merged_bath_coast)
                             lmrg=lenstr(merged_bath_coast)
        write(*,*) 'arguments: ''', bathymetry(1:lbth), ''' ''',
     &                    coastline(1:lcst), ''' ', rmin, ' ''',
     &                          merged_bath_coast(1:lmrg), ''''
        if (ierr /= 0) then
          write(*,'(/8x,A,I3/1x,A/)')  '### ERROR: Argument #',
     &                            nargs-1, 'must be a number.'
          stop
        endif
      else
        write(*,'(/1x,2A/,/6x,2A/1x,A/6x,2A/)') 'This command-line ',
     &     'operator expects 3 or 4 arguments,', 'merge_coast_topo ',
     &     'bathymetry.dat coastline.dat rmin merged_bath_coast.dat',
     &           'or', 'merge_coast_topo   bathymetry.dat   rmin   ',
     &                                      'merged_bath_coast.dat'
        stop
      endif

      npts=0; max_pts=256; allocate(x(max_pts),y(max_pts),z(max_pts))
      open(unit=11, file=bathymetry(1:lbth),status='old',iostat=ierr)
      if (ierr == 0) then
        do while(ierr == 0)
          npts=npts+1
          if (npts > max_pts) then
            allocate(bffr(npts-1,3))
            bffr(1:npts-1,1)=x(1:npts-1)
            bffr(1:npts-1,2)=y(1:npts-1)
            bffr(1:npts-1,3)=z(1:npts-1)
            deallocate(x,y,z)
            max_pts=2*max_pts ; write(*,*) 'extended max_pts =',max_pts
            allocate(x(max_pts),y(max_pts),z(max_pts))
            x(1:npts-1)=bffr(1:npts-1,1)
            y(1:npts-1)=bffr(1:npts-1,2)
            z(1:npts-1)=bffr(1:npts-1,3)
            deallocate(bffr)
          endif
          read(11,*,iostat=ierr) y(npts),x(npts),z(npts)
        enddo
        npts=npts-1
        close(unit=11)
        write(*,*) 'Bathymetry data, npts =', npts
      else
        write(*,'(1x,3A)') '### ERROR: Cannot open file ''',
     &                bathymetry(1:lbth), ''' for reading.'
        stop
      endif

      if (lcst > 0) then
        open(unit=12, file=coastline(1:lcst),status='old',iostat=ierr)
        if (ierr == 0) then 
          do while(ierr == 0)
            npts=npts+1
            if (npts > max_pts) then
              allocate(bffr(npts-1,3))
              bffr(1:npts-1,1)=x(1:npts-1)
              bffr(1:npts-1,2)=y(1:npts-1)
              bffr(1:npts-1,3)=z(1:npts-1)
              deallocate(x,y,z)
              max_pts=2*max_pts; write(*,*) 'extended max_pts =',max_pts
              allocate(x(max_pts),y(max_pts),z(max_pts))
              x(1:npts-1)=bffr(1:npts-1,1)
              y(1:npts-1)=bffr(1:npts-1,2)
              z(1:npts-1)=bffr(1:npts-1,3)
              deallocate(bffr)
            endif
            read(12,*,iostat=ierr) y(npts),x(npts) ;  z(npts)=0.D0
          enddo
          npts=npts-1
          close(unit=12)
          write(*,*) 'Bathymetry + coastline, npts =', npts
        else
          write(*,'(1x,3A)') '### ERROR: Cannot open file ''',
     &                   coastline(1:lcst), ''' for reading.'
          stop
        endif
      else
        write(*,*) 'No user-specified coastline file --> skipping'
      endif

! Convert into quasi-Cartesian coordinates: this is needed to correctly
! judge distances between the points because displacement in meters by
! the same fraction of a degree in geographical coordinates in E-W
! direction is only cos(lat) of that in N-S.  To evaluate cos(lat) and
! also to minimize roundoff errors determine center (half-way between
! min and max in each direction) and use it as the origin. The resultant
! units of (x,y) are meters.

      xmin=x(1); xmax=x(1); ymin=y(1); ymax=y(1)
      do i=2,npts
        xmin=min(xmin,x(i)) ; xmax=max(xmax,x(i))
        ymin=min(ymin,y(i)) ; ymax=max(ymax,y(i))
      enddo
      xc=0.5D0*(xmin+xmax) ; yc=0.5D0*(ymin+ymax)
      csA=cos(pi*yc/180.D0)
      write(*,*) 'cos(A) =', csA
      cffY=Eradius*pi/180.D0 ; cffX=csA*cffY
      do i=1,npts
        x(i)=cffX*(x(i)-xc) ; y(i)=cffY*(y(i)-yc)
c**        write(*,*) y(i),x(i)
      enddo


! Determine points which are too close to each other and average them.
! This is done iteratively while gradually increasing threshold of
! minimum allowed distance. Overall this measure is needed to avoid
! sudden jumps in bottom topography due to contradictory measurements
! from different passes along the measuring trajectory of the boat.

      r2=(rmin/36.D0)**2
      do while (r2 < rmin**2)
        count=1
        do while(count > 0)
          count=0
          do j=1,npts
            if (z(j) > -1.) then
              do i=j+1,npts
                if (z(i) > -1.) then
                  if ((x(i)-x(j))**2 + (y(i)-y(j))**2 < r2) then
                    x(j)=0.5D0*(x(i)+x(j))
                    y(j)=0.5D0*(y(i)+y(j))
                    z(j)=0.5D0*(z(i)+z(j))
                    z(i)=-2.D0; count=count+1
                  endif
                endif
              enddo
            endif
          enddo
          if (count > 0) write(*,*) 'count =',count, 'rmin =',sqrt(r2)
        enddo
        r2=r2*1.0575897D0
      enddo

! Convert back to geographical coordinates (degrees and decimal
! fractions thereof) and save into a file.

      open(unit=13, file=merged_bath_coast(1:lmrg))
      do i=1,npts
        if (z(i) > -1.) then
          x(i)=x(i)/cffX+xc; y(i)=y(i)/cffY+yc
          write(13,'(2F14.10,F8.2)') y(i),x(i),z(i)
        endif
      enddo
      close(unit=13)
      end

