      program mtspline

! Computes time derivatives of time-dependent fields with the purpose
! subsequent of spline interpolation in time.  Essentially the same as
! "tspline" (read preamble there), except that

!     (i) can process sequence of multiple files;

!    (ii) time variable and all the variables to be interpolated must
!         be explicitly specified by the user; all other time-dependent
!         variables will be ignored (timeless variables will be copied
!         into target file, COPY_TIMELESS_VARS must be activated).

!   (iii) fully capable to identify netCDF4/hdf5 chunking patterns and
!         performs all reads and writes aligned with block boundaries
!         (the target file it creates is of the same format (netCDF-3
!         of -4, and is geometrically identical to the source file(s)
!         in terms of all dimensions and variable chunking).

! WARNING: parameter "max_memory_size_bytes" cannot be set to a number
! greater than 2047*1024*1024 because of integer(kind=4) overflow.


#define WRITE_TIME_VAR
#define COPY_TIMELESS_VARS
#define NC4CHUNKING
#define TIMING
c--#define VERBOSE

      implicit none
      integer, parameter :: undefined=-9999
      integer(kind=8) max_buff_size_bytes, max_space_pts
      integer(kind=8),parameter :: max_memory_size_bytes
     &                             = 64_8 * 8192_8 * 8192_8

      character(len=256), allocatable, dimension(:) :: src_fname,
     &                                                 trg_fname
      character(len=32) dimname, attname, varname, time_dim_name,
     &                                             time_var_name
      character(len=32), allocatable, dimension(:) :: vnames

      logical excl_vars, quintic, timedep, not_masked
      integer nargs, ncsrc, nctarg, ndims, unlimdimid, ngatts, nvars,
     &    vartype, vardims, varatts, id,vid,vidtrg, timdimid,timvarid,
     &    ivar, nrec_total, nspc, net_ext_count,   old_fill_mode,
     &    num_chunks, ichunk, chunk_size, size, tstr,tend, jstr,jend,
     &    ksplit, is,ie, i,j,m,n, ierr, ltdm, ltvr, lsrc, ltrg, lvar,
     &                                             latt, lstr, lenstr

      integer, allocatable, dimension(:) :: dimid, vardimids, dsize,
     &                                       vardsizes, start, count
#ifdef NC4CHUNKING
     &                                            , nblocks, blksize

      integer ncfmt, num_subblk_count, num_whole_blks, mwhblk,jsbblk
      logical :: nc4format=.false., nc4chunked=.false.
      integer :: shuffle=1, deflate=1, deflate_level=6 ! <-- default
#endif
      character(len=1), allocatable, dimension(:) :: cbfr
      integer(kind=1),  allocatable, dimension(:) :: ibfr1
      integer(kind=2),  allocatable, dimension(:) :: ibfr2
      integer(kind=4),  allocatable, dimension(:) :: ibffr,    nrecs
      real(kind=4), allocatable, dimension(:) :: bffr4, dbfr4, d2fr4
      real(kind=8), allocatable, dimension(:) :: bffr8, dbfr8, d2fr8,
     &                                                         time
      integer :: nvnames=0, nfiles=0, alloc_dims=0,  alloc_time=0,
     &                  alloc_cbfr=0, alloc_ibfr1=0, alloc_ibfr2=0,
     &                 alloc_ibffr=0, alloc_bffr4=0, alloc_bffr8=0

      character(len=9), parameter :: signame='nc_format'
      character(len=256) string
      real(kind=4) :: spv4=0.0_4    ! Special values,
      real(kind=8) :: spv8=0.0_8    ! if will be found.
#ifdef TIMING
      real(kind=4) tstart,run_time
      real(kind=8) ReadSize, WrtSize, ReadTime, CompTime, WrtTime,
     &                            TargAlloc, ClsSrcTime, ClsTrgTime
      integer(kind=4) iclk(2), nclk, clk_rate, clk_max
      integer(kind=8) :: inc_clk,  net_read_size=0,  net_wrt_size=0,
     &  net_clk=0,  net_read_clk=0,  net_wrt_clk=0,  net_comp_clk=0,
     &  net_trg_alloc=0, net_src_sync=0, net_trg_sync=0
#endif

#include "netcdf.inc"

#ifdef TIMING
      call cpu_time(tstart)
      nclk=1
      call system_clock(iclk(nclk), clk_rate, clk_max)
# ifdef VERBOSE
      write(*,*) 'clk_rate=', clk_rate, ' clk_max =', clk_max
# endif
#endif

! Sort out arguments.
!===== === ==========

      quintic=.false.
      excl_vars=.false.
      LTvr=0 ; m=0
      do i=1,len(time_var_name)
        time_var_name(i:i)=' '  !<-- blank out name
      enddo

      nargs=iargc()
      if (nargs > 0) then
        do n=1,nargs
          call getarg(n,string) ; lstr=lenstr(string)
          if (string(1:1) == '-') then
            if (string(1:10) == '--qspl') then
              quintic=.true.
            elseif (string(1:10) == '--novars') then
              excl_vars=.true.
            elseif (string(1:10) == '--timevar=') then
              if (lstr > 10) then
                time_var_name=string(11:lstr) ; ltvr=lstr-10
              endif
            elseif(string(1:7) == '--vars=') then
              nvnames=1   !<-- at least one; each comma adds one more.
              do i=8,lstr
                if (string(i:i) == ',') nvnames=nvnames+1
              enddo
              allocate(vnames(nvnames))
              j=0 ; is=8
              do while(is < lstr+1)
                ie=is
                do while(string(ie:ie) /= ',' .and. ie < lstr)
                  ie=ie+1
                enddo
                if (string(ie:ie) == ',') ie=ie-1
                j=j+1 ; vnames(j)=string(is:ie) ; is=ie+2
              enddo
            endif
          else
            if (nfiles==0) then
              nfiles=nargs-n+1
              allocate(nrecs(nfiles))
              allocate(src_fname(nfiles))
              allocate(trg_fname(nfiles))
            endif
            m=m+1 ; src_fname(m)=string(1:lstr)
          endif
        enddo

        if (quintic) then
          write(*,*)   'Building quintic spline.'
          max_buff_size_bytes=max_memory_size_bytes/3
        else
          max_buff_size_bytes=max_memory_size_bytes/2
        endif
        if (excl_vars) write(*,*)   'Saving derivatives only.'
        if (ltvr > 0) write(*,'(1x,4A)') 'user supplied time ',
     &             'variable: ''', time_var_name(1:ltvr), ''''

        if (nvnames > 0) then
          write(*,'(1x,A,I3,1x,3A)')      'user supplied variables,',
     &    nvnames, 'total: ''', vnames(1)(1:lenstr(vnames(1))), ''''
          do i=2,nvnames
            write(*,'(36x,3A)') '''', vnames(i)(1:lenstr(vnames(i))),
     &                                                          ''''
          enddo
        endif
      endif
      if (nfiles == 0 .or. ltvr == 0 .or. nvnames == 0) then
        write(*,'(/1x,A/8x,3A/)')   'Usage:', 'mtspline [--qspl] ',
     &         '[--novars] --timevar=name --vars=name1,name2,... ',
     &                                     'file1.nc file2.nc ...'
        stop
      endif

! Create target name: add prefix to source file name, however the
! actual file name may contain directory name(s), so find the last
! slash (if any) and insert prefix there.

      nrec_total=0
      do n=1,nfiles
        lsrc=lenstr(src_fname(n))
        ie=lsrc
        do while(ie > 1 .and. src_fname(n)(ie:ie) /= '/')
          ie=ie-1
        enddo
        if (src_fname(n)(ie:ie) == '/') then
          trg_fname(n)=src_fname(n)(1:ie)/ /
     &               'ts_'/ /src_fname(n)(ie+1:lsrc)
        else
          trg_fname(n)='ts_'/ /src_fname(n)(ie:lsrc)
        endif
        ltrg=lenstr(trg_fname(n))

! Open source netCDF file:  Try to identify time dimension: check
!===== ====== ====== =====  whether user-supplied time variable has
! only one dimension, and this is it...

        ierr=nf_open(src_fname(n)(1:lsrc), nf_nowrite, ncsrc)
        if (ierr == nf_noerr) then
          write(*,'(/1x,3A)') 'Processing ''', src_fname(n)(1:lsrc),
     &                                                       '''...'

#ifdef NC4CHUNKING
          nc4format=.false.
          ierr=nf_inq_format(ncsrc, ncfmt)
          if (ncfmt == nf_format_classic) then
            write(*,'(1x,A)') 'Classic netCDF format'
          elseif (ncfmt == nf_format_64bit) then
            write(*,'(1x,A)') '64bit netCDF-3 format'
          elseif (ncfmt == nf_format_netcdf4) then
            write(*,'(1x,A)') 'netCDF-4 format'
            nc4format=.true.
          elseif (ncfmt == nf_format_netcdf4_classic) then
            write(*,'(1x,A)') 'netCDF-4 format'
            nc4format=.true.
          else
            write(*,'(1x,A)') '### ERROR: Unknown format.'
          endif
#endif

          ierr=nf_inq(ncsrc, ndims, nvars, ngatts, unlimdimid)
          if (ierr == nf_noerr) then
#ifdef VERBOSE
            write(*,'(4(1x,A,I4))') 'ncsrc: ndims =',ndims,'nvars =',
     &            nvars,'ngatts =',ngatts,'unlimdimid =',unlimdimid
#endif
            if (ndims > alloc_dims) then
              alloc_dims=ndims
              if (allocated(dimid)) then
                deallocate(count)      ;  deallocate(start)
#ifdef NC4CHUNKING
                deallocate(nblocks)    ;  deallocate(blksize)
#endif
                deallocate(vardsizes)  ;  deallocate(dsize)
                deallocate(vardimids)  ;  deallocate(dimid)
              endif
              allocate(dimid(ndims))   ;  allocate(vardimids(ndims))
              allocate(dsize(ndims))   ;  allocate(vardsizes(ndims))
#ifdef NC4CHUNKING
              allocate(nblocks(ndims)) ;  allocate(blksize(ndims))
                             nblocks=0 ;  blksize=0
#endif
              allocate(start(ndims))   ;  allocate(count(ndims))
                             start=0   ;  count=0
            endif

! Check whether the user-specified time variable exists, verify that
! it has only one dimension: this is the time dimension and its length
! is the numbers of records. Use ltvr>0 as flag that the name of time
! variable is already known from above.

            timvarid=undefined ; ltdm=0
            timdimid=undefined ; nrecs(n)=0
            do i=1,len(time_dim_name)
              time_dim_name(i:i)=' '  !<-- blank out name
            enddo

            ierr=nf_inq_varid(ncsrc, time_var_name(1:ltvr), vid)
            if (ierr == nf_noerr) then
              ierr=nf_inq_var(ncsrc, vid,  varname,  vartype,
     &                            vardims, vardimids, varatts)
              if (ierr == nf_noerr) then
                if (vardims == 1) then
                  timvarid=vid
                  timdimid=vardimids(vardims)
                  ierr=nf_inq_dim(ncsrc, timdimid, dimname, size)
                  if (ierr == nf_noerr) then
                    time_dim_name=dimname ; ltdm=lenstr(time_dim_name)
                    nrecs(n)=size ; nrec_total=nrec_total+nrecs(n)
                    write(*,'(1x,2A,I3,1x,3A,I5,1x,3A)') 'identified ',
     &                   'dimension #', timdimid, 'as time dimension ',
     &                    time_dim_name(1:ltdm),' =',nrecs(n), 'of ''',
     &                                    time_var_name(1:ltvr), '''.'
                  else
                    write(*,'(/1x,2A,I3/12x,3A/12x,A/)') '### ERROR: ',
     &                 'Cannot determine name and size of dimension #',
     &               vardims, 'in netCDF file ''',src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,4A,I4,1x,A/)')      '### ERROR: User ',
     &             'specified time variable ''', time_var_name(1:ltvr),
     &                   ''' has', vardims, 'dimensions instead of 1.'
                endif
              else
                write(*,'(/1x,2A,I3/12x,4A/12x,A/)')     '### ERROR: ',
     &               'Cannot make general inquiry for variable #', vid,
     &                   'in ', 'netCDF file ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,6A/)')  '### ERROR: User specified time ',
     &                           'variable ''', time_var_name(1:ltvr),
     &              ''' not found in ''', src_fname(n)(1:lsrc), '''.'
            endif
            if (timvarid == undefined .or. timdimid == undefined) stop


! Create target file:
!======= ====== =====

#ifdef NC4CHUNKING
            if (nc4format) then
              ierr=nf_create(trg_fname(n)(1:ltrg),  nf_netcdf4,
     &                                                  nctarg)
            else
#endif
              ierr=nf_create(trg_fname(n)(1:ltrg), nf_clobber +
     &                                 nf_64bit_offset, nctarg)
#ifdef NC4CHUNKING
            endif
#endif
            if (ierr == nf_noerr) then
              ierr=nf_set_fill(nctarg, nf_nofill, old_fill_mode)
              write(*,'(4x,3A)')   'Created netCDF file ''',
     &                           trg_fname(n)(1:ltrg), '''.'

! Sort out dimensions:   If time dimension is not identified yet, make
!----- --- -----------   another attempt: assume that it is  (i) either
! unlimited dimension; or (ii) a dimension with name denting as "time".

              do id=1,ndims
                ierr=nf_inq_dim(ncsrc, id, dimname, dsize(id))
                if (ierr == nf_noerr) then
                  lvar=lenstr(dimname)

                  if (id == unlimdimid) then
                    j=nf_unlimited
                  else
                    j=dsize(id)
                  endif
                  ierr=nf_def_dim(nctarg,dimname(1:lvar), j, dimid(id))
                  if (ierr == nf_noerr) then
                    write(*,'(6x,3A,I5)') 'Created dimension ''',
     &                            dimname(1:lvar), '''=', dsize(id)
                  else
                    write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                        'define dimension ''', dimname(1:lvar),
     &                  ''' in netCDF file ''', trg_fname(n)(1:ltrg),
     &                                      '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,2A,I3/12x,3A/12x,A/)') '### ERROR: ',
     &               'Cannot determine name and size of dimension #',
     &                 id, 'in netCDF file ''', src_fname(n)(1:lsrc),
     &                                      '''.', nf_strerror(ierr)
                endif
              enddo

! Copy global attributes

              do id=1,ngatts
                ierr=nf_inq_attname(ncsrc, nf_global, id, attname)
                if (ierr == nf_noerr) then
                  latt=lenstr(attname) ; lvar=lenstr(signame)
                  if (attname(1:latt) /= signame(1:lvar)) then
                    ierr=nf_copy_att(ncsrc, nf_global, attname(1:latt),
     &                                              nctarg, nf_global)
                  if (ierr == nf_noerr) then
                    write(*,'(8x,3A)')    'Copied global attribute ''',
     &                                           attname(1:latt), '''.'
                  else
                    write(*,'(/1x,4A/12x,3A/12x,A/)')    '### ERROR: ',
     &              'Cannot copy global attribute ''', attname(1:latt),
     &              '''',  'into netCDF file ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  endif
                else
                  write(*,'(/1x,2A,I3,1x,3A/12x,A/)')    '### ERROR: ',
     &                   'Cannot determine name of global attribute #',
     &                 id, 'from netCDF file ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                endif
              enddo

! Create variables: scan all the variables, but mask out those which
! are not needed.

              do vid=1,nvars
                ierr=nf_inq_var(ncsrc, vid, varname,  vartype,
     &                            vardims, vardimids, varatts)
                if (ierr == nf_noerr) then
                  lvar=lenstr(varname)               ! remap sizes of
                  size=1 ; timedep=.false.           ! dimensions from
                  do j=1,vardims                     ! file to variable
                    vardsizes(j)=dsize(vardimids(j))
                    if (vardimids(j) == timdimid) then
                      timedep=.true.
                    else
                      size=size*vardsizes(j)
                    endif
                  enddo
#ifdef COPY_TIMELESS_VARS
                  if (vid == timvarid .or. .not.timedep) then
#else
                  if (vid == timvarid) then
#endif
                    not_masked=.true.
                  else
                    not_masked=.false.
                    do i=1,nvnames
                      lstr=lenstr(vnames(i))
                      if (lvar == lstr .and. varname(1:lvar) ==
     &                      vnames(i)(1:lstr)) not_masked=.true.
                    enddo
                  endif
                  if (not_masked) then
                    if (.not.excl_vars .or. .not.timedep .or. vid ==
     &                                                  timvarid) then

! Note that array "vardimids"  goes straight from "nf_inq_var" above
! (where it is output) into "nf_def_var" (as input) which implies that
! the dimension IDs are exactly the same in both files.

                      ierr=nf_def_var(nctarg, varname(1:lvar), vartype,
     &                                      vardims, vardimids, vidtrg)
                      if (ierr == nf_noerr) then
                        write(*,'(6x,3A,I10)') 'Created variable ''',
     &                           varname(1:lvar), ''', size =', size
#ifdef NC4CHUNKING
! Warning: call to "nf_inq_var_chunking" should be restricted to files
! known of having netCDF-4 format -- if the file is netCDF-3 attempt to
! call it causes segmentation fault rather than returns error status.

                        if (nc4format) then
                          nc4chunked=.false.
                          ierr=nf_inq_var_chunking(ncsrc,vid,j,blksize)

                          if (j == nf_chunked) nc4chunked=.true.
                          if (nc4chunked) then
                            do j=1,vardims
                              nblocks(j)=(vardsizes(j)+blksize(j)-1)
     &                                                /blksize(j)
# ifdef VERBOSE
                              write(*,'(1x,A,I3,4(1x,A,I7))') 'dim', j,
     &                           'size=', vardsizes(j), 'chunk size =',
     &                             blksize(j), 'nblocks =', nblocks(j),
     &                           'padd =',
     &                              blksize(j)*nblocks(j)-vardsizes(j)
# endif
                            enddo
                            ierr=nf_def_var_chunking(nctarg, vidtrg,
     &                                          nf_chunked, blksize)
                            if (ierr == nf_noerr) then
                              ierr=nf_def_var_deflate(nctarg, vidtrg,
     &                              shuffle, deflate, deflate_level)
                              if (ierr == nf_noerr) then
                                 write(*,'(8x,2A)')    'set chunking ',
     &                                                 'and deflation'
                              else
                                write(*,'(/1x,4A/12x,A/)')
     &                           '### ERROR: Cannot define deflation ',
     &                              'for variable ''', varname(1:lvar),
     &                                        '''.', nf_strerror(ierr)
                              endif
                            else
                              write(*,'(/1x,4A/12x,A/)') '### ERROR: ',
     &                        'Cannot define chunking for variable ''',
     &                       varname(1:lvar), '''.', nf_strerror(ierr)
                            endif
# ifdef VERBOSE
                          else
                            write(*,'(1x,3A)') 'No nf4chinking for ''',
     &                                            varname(1:lvar),'''.'
# endif
                          endif
                        endif !<-- nc4format
#endif

                        do i=1,varatts
                          ierr=nf_inq_attname(ncsrc, vid, i, attname)
                          latt=lenstr(attname)
                          ierr=nf_copy_att(ncsrc, vid, attname(1:latt),
     &                                                 nctarg, vidtrg)
                          if (ierr == nf_noerr) then
                            write(*,'(8x,3A)')   'Copied attribute ''',
     &                                          attname(1:latt), '''.'
                          else
                            write(*,'(/1x,5A/12x,3A/12x,A/)')
     &                           '### ERROR: Cannot copy attribute ''',
     &                           attname(1:latt), ''' for variable ''',
     &                                     varname(1:lvar), ''' into ',
     &                          'netCDF file ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                          endif
                        enddo
                      else
                        write(*,'(/1x,4A/12x,3A/12x,A/)')'### ERROR: ',
     &                    'Cannot create variable ''', varname(1:lvar),
     &                 '''', 'in netCDF file ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                      endif
                    endif

! Create variables for time derivatives.

                    if (timedep .and.  vid /= timvarid .and.
     &                                ierr == nf_noerr) then
                      ierr=nf_def_var(nctarg,'dt_'/ /varname(1:lvar),
     &                          vartype, vardims, vardimids, vidtrg)
                      if (ierr == nf_noerr) then
                        ierr=nf_put_att_text(nctarg,vidtrg,'long_name',
     &                              lvar+17, 'time derivative for '/ /
     &                                               varname(1:lvar))
                        if (ierr == nf_noerr) then
                          write(*,'(6x,4A,I10)')   'Created variable ',
     &                     '''dt_', varname(1:lvar),''', size =', size
                          ierr=nf_inq_att(ncsrc,vid,'_FillValue',is,ie)
                          if (ierr == nf_noerr) then
                            ierr=nf_copy_att(ncsrc, vid, '_FillValue',
     &                                                nctarg, vidtrg)
                            if (ierr /= nf_noerr) then
                              write(*,'(/1x,2A/12x,A)')  '### ERROR: ',
     &                           'Cannot copy attibute ''_FillValue''',
     &                                               nf_strerror(ierr)
                            endif
                            if (is == nf_real) then
                              if (spv4 == 0.0_4) then
                                ierr=nf_get_att_real( ncsrc, vid,
     &                                         '_FillValue', spv4)
                              endif
                            elseif (is == nf_double) then
                              if (spv8 == 0.0_8) then
                                ierr=nf_get_att_double(ncsrc, vid,
     &                                         '_FillValue', spv8)
                              endif
                            endif
                          else
                            ierr=nf_noerr
                          endif
                        else
                          write(*,'(/1x,4A/12x,A)')      '### ERROR: ',
     &                                'Cannot put attribute for ''dt_',
     &                       varname(1:lvar), '''.', nf_strerror(ierr)
                        endif
                      else
                        write(*,'(/1x,4A/12x,3A/12x,A/)')'### ERROR: ',
     &                 'Cannot create variable ''dt_', varname(1:lvar),
     &                 '''', 'in netCDF file ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                      endif

                      if (quintic .and. ierr == nf_noerr) then
                        ierr=nf_def_var(nctarg,'d2_'/ /varname(1:lvar),
     &                            vartype, vardims, vardimids, vidtrg)
                        if (ierr == nf_noerr) then
                          ierr=nf_put_att_text(nctarg, vidtrg,
     &                                       'long_name', lvar+19,
     &                                    'second derivative for '/ /
     &                                              varname(1:lvar))
                          if (ierr == nf_noerr) then
                            write(*,'(6x,4A,I10)')          'Created ',
     &                               'variable ''d2_', varname(1:lvar),
     &                                             ''', size =', size
                            ierr=nf_inq_att(ncsrc,vid,'_FillValue',
     &                                                      is,ie)
                            if (ierr == nf_noerr) then
                              ierr=nf_copy_att(ncsrc,vid, '_FillValue',
     &                                                 nctarg, vidtrg)
                              if (ierr /= nf_noerr) then
                                write(*,'(/1x,2A/12x,A)')'### ERROR: ',
     &                           'Cannot copy attibute ''_FillValue''',
     &                                               nf_strerror(ierr)
                              endif
                            else
                              ierr=nf_noerr
                            endif
                          else
                            write(*,'(/1x,4A/12x,A)')    '### ERROR: ',
     &                                'Cannot put attribute for ''d2_',
     &                       varname(1:lvar), '''.', nf_strerror(ierr)
                          endif
                        else
                          write(*,'(/1x,3A/12x,3A/12x,A/)')
     &                       '### ERROR: Cannot create variable ''d2_',
     &                      varname(1:lvar), '''', 'in netCDF file ''',
     &                                     trg_fname(n)(1:ltrg), '''.',
     &                                               nf_strerror(ierr)
                        endif
                      endif !<-- quintic
                    endif
                  endif !<-- not_masked
                else
                  write(*,'(/1x,2A,I3/12x,4A/12x,A/)')   '### ERROR: ',
     &                    'Cannot make general inquiry for variable #',
     &              vid, 'in ', 'netCDF file ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                endif
              enddo  !<--  vid=1,nvars

              ierr=nf_enddef(nctarg)
              write(*,'(6x,3A)')  'Leaving definition mode for ''',
     &                                  trg_fname(n)(1:ltrg), '''.'

#ifdef COPY_TIMELESS_VARS

! Copy single-record variables: basically read as a whole and write
! into the target file. This operation is completely passive in sense
! that no computation or reshaping takes place.

              do vid=1,nvars
                ierr=nf_inq_var(ncsrc, vid, varname,  vartype,
     &                            vardims, vardimids, varatts)
                if (ierr == nf_noerr) then
                  lvar=lenstr(varname)               ! remap sizes of
                  nspc=1 ; timedep=.false.           ! dimensions from
                  do j=1,vardims                     ! file to variable
                    vardsizes(j)=dsize(vardimids(j))
                    if (vardimids(j) == timdimid) then
                      timedep=.true.
                    else
                      nspc=nspc*vardsizes(j)
                    endif
                  enddo
                  if (.not.timedep) then

          if (vartype == nf_char .and. nspc > alloc_cbfr) then
            if (allocated(cbfr)) deallocate(cbfr)
            allocate(cbfr(nspc)) ;   alloc_cbfr=nspc
            write(*,*) 'allocated nf_char workspace', nspc
          elseif (vartype == nf_byte .and. nspc > alloc_ibfr1) then
            if (allocated(ibfr1)) deallocate(ibfr1)
            allocate(ibfr1(nspc)) ; alloc_ibfr1=nspc
            write(*,*) 'allocated nf_byte workspace', nspc
          elseif (vartype == nf_short .and. nspc > alloc_ibfr2) then
            if (allocated(ibfr2)) deallocate(ibfr2)
            allocate(ibfr2(nspc)) ; alloc_ibfr2=nspc
            write(*,*) 'allocated nf_int2 workspace', nspc
          elseif (vartype == nf_int .and. nspc > alloc_ibffr) then
            if (allocated(ibffr)) deallocate(ibffr)
            allocate(ibffr(nspc)) ; alloc_ibffr=nspc
            write(*,*) 'allocated nf_int workspace', nspc
          elseif (vartype == nf_float .and. nspc > alloc_bffr4) then
            if (allocated(bffr4)) deallocate(bffr4)
            allocate(bffr4(nspc)) ; alloc_bffr4=nspc
            write(*,*) 'allocated nf_real workspace', nspc
          elseif (vartype==nf_double .and. nspc > alloc_bffr8) then
            if (allocated(bffr8)) deallocate(bffr8)
            allocate(bffr8(nspc)) ; alloc_bffr8=nspc
            write(*,*) 'allocated nf_double workspace', nspc
          endif

                    ierr=nf_inq_varid(nctarg, varname(1:lvar), vidtrg)
                    if (ierr == nf_noerr) then

                      if (vartype == nf_char) then
                        ierr=nf_get_var_text  (ncsrc, vid, cbfr)
                      elseif (vartype == nf_byte) then
                        ierr=nf_get_var_int1  (ncsrc, vid, ibfr1)
                      elseif (vartype == nf_short) then
                        ierr=nf_get_var_int2  (ncsrc, vid, ibfr2)
                      elseif (vartype == nf_int) then
                        ierr=nf_get_var_int   (ncsrc, vid, ibffr)
                      elseif (vartype == nf_float) then
                        ierr=nf_get_var_real  (ncsrc, vid, bffr4)
                      elseif (vartype == nf_double) then
                        ierr=nf_get_var_double(ncsrc, vid, bffr8)
                      else
                        write(*,'(/8x,4A/)')  '### ERROR: Unknown ',
     &                 'type of variable ''', varname(1:lvar), '''.'
                        ierr=nf_noerr+1
                      endif
                      if (ierr == nf_noerr) then
# ifdef TIMING
              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_read_clk=net_read_clk+inc_clk
              if (vartype == nf_char .or. vartype == nf_byte) then
                net_read_size=net_read_size+nspc
              elseif (vartype == nf_short) then
                net_read_size=net_read_size+2*nspc
              elseif (vartype == nf_int .or. vartype == nf_float) then
                net_read_size=net_read_size+4*nspc
              elseif (vartype == nf_double) then
                net_read_size=net_read_size+8*nspc
              endif
# endif
                        if (vartype == nf_char) then
                          ierr=nf_put_var_text  (nctarg, vidtrg, cbfr)
                        elseif (vartype == nf_byte) then
                          ierr=nf_put_var_int1  (nctarg, vidtrg, ibfr1)
                        elseif (vartype == nf_short) then
                          ierr=nf_put_var_int2  (nctarg, vidtrg, ibfr2)
                        elseif (vartype == nf_int) then
                          ierr=nf_put_var_int   (nctarg, vidtrg, ibffr)
                        elseif (vartype == nf_float) then
                          ierr=nf_put_var_real  (nctarg, vidtrg, bffr4)
                        elseif (vartype == nf_double) then
                          ierr=nf_put_var_double(nctarg, vidtrg, bffr8)
                        endif
                        if (ierr == nf_noerr) then
                          write(*,'(1x,6A)')         'wrote timeless ',
     &                    'variable ''', varname(1:lvar), ''' into ''',
     &                                     trg_fname(n)(1:ltrg), '''.'
# ifdef TIMING
              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_wrt_clk=net_wrt_clk+inc_clk
              if (vartype == nf_char .or. vartype == nf_byte) then
                net_wrt_size=net_wrt_size+nspc
              elseif (vartype == nf_short) then
                net_wrt_size=net_wrt_size+2*nspc
              elseif (vartype==nf_int .or. vartype==nf_float) then
                net_wrt_size=net_wrt_size+4*nspc
              elseif (vartype == nf_double) then
                net_wrt_size=net_wrt_size+8*nspc
              endif
# endif
                        else
                          write(*,'(/1x,6A/12x,A/)')     '### ERROR: ',
     &                     'Cannot write variable ''', varname(1:lvar),
     &                              ''' into ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                        endif
                      else
                        write(*,'(/1x,6A/12x,A/)')       '### ERROR: ',
     &                      'Cannot read variable ''', varname(1:lvar),
     &                                 ''' from ''', src_fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                      'find variable ID for ''', varname(1:lvar),
     &                                ''' in ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  endif ! <-- .not.timedep
                else
                  write(*,'(/1x,2A,I3/12x,4A/12x,A/)')   '### ERROR: ',
     &                    'Cannot make general inquiry for variable #',
     &              vid, 'in ', 'netCDF file ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                endif
              enddo  !<--  vid=1,nvars
#endif /* COPY_TIMELESS_VARS */
              i=nf_close(nctarg)
#ifdef TIMING
              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_trg_sync=net_trg_sync+inc_clk
#endif
            else
              write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot create ',
     &                          'netCDF file ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,4A/12x,A/)')       '### ERROR: Cannot make ',
     &                           'general inquiry into netCDF file ''',
     &                  src_fname(n)(1:lsrc), '''.', nf_strerror(ierr)
          endif
          i=nf_close(ncsrc)
#ifdef TIMING
          nclk=3-nclk
          call system_clock(iclk(nclk), clk_rate,clk_max)
          inc_clk=iclk(nclk)-iclk(3-nclk)
          if (inc_clk < 0) inc_clk=inc_clk+clk_max
          net_clk=net_clk+inc_clk
          net_src_sync=net_src_sync+inc_clk
#endif
        else
          write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &               'file ''', src_fname(n)(1:lsrc),''' for reading.',
     &                                               nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) goto 99                      !--> BREXIT
      enddo  !<--n, loop over files

#ifdef WRITE_TIME_VAR

!    **********   ****       ******   ******    **********
!    *  ****  *   ****       ** ***   ** ***    ****     **
!       ****      ****      **  **** **  ****   ****
!       ****      ****     **   *****    ****   ********
!       ****      ****    **    ****     ****   ****
!       ****      ****   **     ***      ****   ****     **
!       ****      ****   **     **       ****   **********


      write(*,*) 'Total number of records in all files:', nrec_total
      if (nrec_total > alloc_time) then
        if (allocated(time)) deallocate(time)
        allocate(time(nrec_total)) ; alloc_time=nrec_total
        write(*,*) 'allocated time variable array, size =', alloc_time
      endif

! Process time variable, basically copy from one file to the other,
! however, CAUTION: Although it may seem that the entire variable can
! be read and written as a whole, using function nf_put_var_ instead of
! more complicated nf_put_vara_ (...start,count,...), an attempt to do
! so causes it not be written into the target file if its dimension is
! unlimited and is zero-size (meaning that nothing is written yet) at
! the time when attempt is made [?? it appears that nf_put_var_ infers
! the dimensions of the variable to be written from the file before
! writing it, and, as the result, it may be unsafe if the variable has
! unlimited dimension which is about to change (increase) as the result
! of writing].

      tend=0
      do n=1,nfiles
        lsrc=lenstr(src_fname(n)) ; ltrg=lenstr(trg_fname(n))
        ierr=nf_open(src_fname(n)(1:lsrc), nf_nowrite, ncsrc)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varid(ncsrc, time_var_name(1:ltvr), timvarid)
          if (ierr == nf_noerr) then
            start(1)=1  ; count(1)=nrecs(n)
            tstr=tend+1 ; tend=tstr+nrecs(n)-1
            ierr=nf_get_vara_double(ncsrc, timvarid, start,count,
     &                                                time(tstr))
            if (ierr == nf_noerr) then
# ifdef TIMING
              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_read_clk=net_read_clk+inc_clk
              net_read_size=net_read_size+8*nrecs(n)


              write(*,'(F10.1,2x,5A)') dble(net_clk)/dble(clk_rate),
# else
              write(*,'(12x,5A)')
# endif
     &               'read ''', time_var_name(1:ltvr), ''' from ''',
     &                                   src_fname(n)(1:lsrc), ''''

              ierr=nf_open(trg_fname(n)(1:ltrg), nf_write, nctarg)
              if (ierr == nf_noerr) then
                ierr=nf_set_fill(nctarg, nf_nofill, old_fill_mode)
                ierr=nf_inq_varid(nctarg,time_var_name(1:ltvr),vidtrg)
                if (ierr == nf_noerr) then
                  ierr=nf_put_vara_double(nctarg, vidtrg, start,count,
     &                                                     time(tstr))
                  if (ierr == nf_noerr) then
# ifdef TIMING
! Note: despite the relatively small data written, writing out time
! variable causes inflation of the unlimited dimension of the target
! file and may take very significant amount of time. For this reason
! this time is counted as _sync_ rather than _wrt_ and is reported
! accordingly; data associated with time variable is usually ignored,
! hence not counted.

              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_trg_alloc=net_trg_alloc+inc_clk

              write(*,'(F10.1,1x,5A)') dble(net_clk)/dble(clk_rate),
# else
              write(*,'(11x,5A)')
# endif
     &                'wrote ''', time_var_name(1:ltvr), ''' into ''',
     &                                     trg_fname(n)(1:ltrg), ''''
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &                 'write ''', time_var_name(1:ltvr), ''' into ''',
     &                  trg_fname(n)(1:ltrg), '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,6A/12x,A/)') '### ERROR: Cannot find ',
     &                     'variable ID for ''', time_var_name(1:ltvr),
     &                                ''' in ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                endif
                i=nf_close(nctarg)
# ifdef TIMING
                nclk=3-nclk
                call system_clock(iclk(nclk), clk_rate,clk_max)
                inc_clk=iclk(nclk)-iclk(3-nclk)
                if (inc_clk < 0) inc_clk=inc_clk+clk_max
                net_clk=net_clk+inc_clk
                net_trg_sync=net_trg_sync+inc_clk
# endif
              else
                write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open ',
     &                          'netCDF file ''', trg_fname(n)(1:lsrc),
     &                            ''' for writing.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot read ',
     &              'variable ''', time_var_name(1:ltvr), ''' from ''',
     &                  src_fname(n)(1:lsrc), '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,6A/12x,A/)')       '### ERROR: Cannot find ',
     &                     'variable ID for ''', time_var_name(1:ltvr),
     &                                ''' in ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
          endif
          i=nf_close(ncsrc)
# ifdef TIMING
          nclk=3-nclk
          call system_clock(iclk(nclk), clk_rate,clk_max)
          inc_clk=iclk(nclk)-iclk(3-nclk)
          if (inc_clk < 0) inc_clk=inc_clk+clk_max
          net_clk=net_clk+inc_clk
          net_src_sync=net_src_sync+inc_clk
# endif
        else
          write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &               'file ''', src_fname(n)(1:lsrc),''' for reading.',
     &                                               nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) goto 99                      !--> BREXIT
      enddo  !<--  n=1,nfiles
#endif



!      *****    *********    ******   *******    *********
!    ***   ***  *  ***  *   *   ***   ***   ***  *  ***  *
!    ***           ***     **   ***   ***   ***     ***
!      *****       ***    ***   ***   ***   **      ***
!          ***     ***    *********   ******        ***
!    ***   ***     ***    ***   ***   ***  **       ***
!      *****       ***    ***   ***   ***   ***     ***


      do ivar=1,nvnames
        n=1 ; lsrc=lenstr(src_fname(n))
        ierr=nf_open(src_fname(n)(1:lsrc), nf_nowrite, ncsrc)
        if (ierr == nf_noerr) then
          lvar=lenstr(vnames(ivar))
          ierr=nf_inq_varid(ncsrc, vnames(ivar)(1:lvar), vid)
          if (ierr == nf_noerr) then
            ierr=nf_inq_var(ncsrc, vid, varname,  vartype,
     &                        vardims, vardimids, varatts)
            if (ierr == nf_noerr) then
              lvar=lenstr(varname)
              do j=1,vardims                         ! remap sizes of
                vardsizes(j)=dsize(vardimids(j))     ! dimensions from
              enddo                                  ! file to variable

              if (vartype == nf_float) then
                max_space_pts=max_buff_size_bytes/(4*nrec_total)
              elseif (vartype==nf_double) then
                max_space_pts=max_buff_size_bytes/(8*nrec_total)
              else
                write(*,'(/8x,4A/)') '### ERROR: Unrecognized type ',
     &                       'of variable ''', varname(1:lvar), '''.'
                ierr=nf_noerr+1
              endif

#ifdef NC4CHUNKING
! WARNING: call to "nf_inq_var_chunking" should be restricted to files
! known of having netCDF-4 format -- if the file is netCDF-3 an attempt
! to call it causes SEGMENTATION FAULT internally inside netCDF library
! rather than returns error status.

! Note: In terminology adopted here netCDF-4/hdf5-chunks are called
! "blocks" (hence arrays "blksize"and "nblocks" which hold hdf5 chunk
! sizes and numbers of chunks in each dimension respectively; as well
! as "num_whole_blks" the total number of block within one time record
! of the array). Conversely, "num_chunks" and "chunk_size" have nothing
! with hdf5 and refer to the number of chunks and chunk size for the
! purpose of splitting one of the dimensions (dim. number "ksplit").

! The algorithm works as follows: depending on the hdf5-block size, the
! total number of records, and the amount of memory available, decide
! which dimension of the block needs to  be split (in the best case
! scenario, the outcome of this search is ksplit=vardims-1, and
! num_chunks=1, which means that the entire block can be read at once,
! however, generally speaking, the algorithm is prepared to handle any
! situation, where the block is larger (say the last dimension should
! be divided into chunks), or much larger (not just the last, but even
! one slice of the last dimension cannot be lifted at once).  So the
! first dimension for which the total memory would exceed the specified
! limit is identified as "ksplit".  Find an appropriate partitioning
! of that dimension, "num_chunks" and "chunk_size".  All dimensions
! below will be read at once; all above -- one slice at a time.

! The maximum granularity of reading in this algorithm is one block
! (no attempt to joint several blocks into one read operation is made).

! The netCDF-3 version of this algorithm (also applies to non-chunked
! netCDF-4 variables, if any) is exactly the same, except that instead
! of block sizes "blksize" it utilizes array dimensions.


              nc4chunked=.false.
              if (nc4format) then
                ierr=nf_inq_var_chunking(ncsrc, vid, j, blksize)
                if (ierr == nf_noerr) then
                  if (j == nf_chunked) nc4chunked=.true.
                else
                  write(*,'(1x,5A/12x,A)')'### ERROR: Cannot inquire ',
     &                     'chunking for variable ''', varname(1:lvar),
     &                                        '''.', nf_strerror(ierr)

                  goto 99                                  !--> BREXIT
                endif
              endif !<-- nc4format

              if (nc4chunked) then
                do j=1,vardims
                  nblocks(j)=(vardsizes(j)+blksize(j)-1)/blksize(j)
# ifdef VERBOSE
                      write(*,'(1x,A,I3,4(1x,A,I7))') 'dim',j, 'size=',
     &                        vardsizes(j), 'chunk size =', blksize(j),
     &                               'nblocks =', nblocks(j), 'padd =',
     &                              blksize(j)*nblocks(j)-vardsizes(j)
# endif
                enddo

                j=0 ;  ksplit=1 ;  size=1 ;  m=1
                do while(m < max_space_pts .and. j < vardims-1)
                  j=j+1 ; ksplit=j ;   size=m ;  m=size*blksize(j)
# ifdef VERBOSE
                  write(*,'(2(1x,A,I3),1x,A,I5,3(1x,A,I9))') 'j =', j,
     &             'ksplit =',ksplit, 'blksize =',blksize(j), 'm =',m,
     &             'size =', size,   'max_space_pts =', max_space_pts
# endif
                enddo

                chunk_size=max_space_pts/size
                num_chunks=(blksize(ksplit)+chunk_size-1)/chunk_size
                chunk_size=(blksize(ksplit)+num_chunks-1)/num_chunks

                num_subblk_count=1
                do j=ksplit+1,vardims-1
                  num_subblk_count=num_subblk_count*blksize(j)
                enddo

                num_whole_blks=1
                do j=1,vardims-1
                  num_whole_blks=num_whole_blks*nblocks(j)
                enddo
                net_ext_count=num_whole_blks*num_subblk_count
# ifdef VERBOSE
                write(*,'(/1x,A,I3,4(1x,A,I5),2/10x,2(4x,A,I4)/)')
     &                  'ksplit =', ksplit, 'chunk_size =',chunk_size,
     &        'num_chunks =', num_chunks, 'blksize =', blksize(ksplit),
     &                 'padd =', num_chunks*chunk_size-blksize(ksplit),
     &                          'num_subblk_count =', num_subblk_count,
     &                            'num_whole_blks =', num_whole_blks

# endif
              else !--> .not.nc4chunked -->
# ifdef VERBOSE
                write(*,'(1x,3A)') 'No nf4chinking for ''',
     &                               varname(1:lvar),'''.'
# endif
#endif

! Decide which dimension to split into chunks (dimension "ksplit"), how
! many chunks, and find appropriate chunk size.    All dimensions below
! "ksplit" will be read as a whole; all dimensions above will be read
! one index at a time.

                j=0 ;  ksplit=1 ;  size=1 ;  m=1
                do while(m < max_space_pts .and. j < vardims-1)
                  j=j+1 ; ksplit=j ;  size=m ;  m=size*vardsizes(j)
#ifdef VERBOSE
                  write(*,'(2(1x,A,I3),1x,A,I5,3(1x,A,I9))') 'j =', j,
     &                  'ksplit =', ksplit, 'dimsize =', vardsizes(j),
     &                        'm =',m,'size =',size,'max_space_pts =',
     &                                               max_space_pts
#endif
                enddo

                chunk_size=max_space_pts/size
                num_chunks=(vardsizes(ksplit)+chunk_size-1)/chunk_size
                chunk_size=(vardsizes(ksplit)+num_chunks-1)/num_chunks

                                        ! fold all dimensions above
                net_ext_count=1         ! "ksplit" into single counter
                do j=ksplit+1,vardims-1
                  net_ext_count=net_ext_count*vardsizes(j)
                enddo

! Report what is decided about chunking: dimension number, its size,
! number of chunks, chunk size, and combined size of all dimensions
! above ksplit.

                write(*,'(/1x,A,I2,1x,A,I5,1x,A,I3,1x,A,I4,1x,A,I5/)')
     &            'ksplit =', ksplit,   'dimsize =', vardsizes(ksplit),
     &            'num_chunks =',num_chunks, 'chunk_size =',chunk_size,
     &                                'net_ext_count =', net_ext_count

#ifdef NC4CHUNKING
              endif !<-- .not.nc4chunked
#endif
            else
              write(*,'(/1x,2A,I3/12x,4A/12x,A/)')       '### ERROR: ',
     &              'Cannot make general inquiry for variable # ', vid,
     &                   'in ', 'netCDF file ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,6A/12x,A/)')       '### ERROR: Cannot find ',
     &                      'variable ID for ''', vnames(ivar)(1:lvar),
     &                                ''' in ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
          endif
          i=nf_close(ncsrc)
#ifdef TIMING
          nclk=3-nclk
          call system_clock(iclk(nclk), clk_rate,clk_max)
          inc_clk=iclk(nclk)-iclk(3-nclk)
          if (inc_clk < 0) inc_clk=inc_clk+clk_max
          net_clk=net_clk+inc_clk
          net_src_sync=net_src_sync+inc_clk
#endif
        else
          write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open netCDF ',
     &               'file ''', src_fname(n)(1:lsrc),''' for reading.',
     &                                               nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) goto 99                      !--> BREXIT

! Note: the fundamental assumption here is that the spatial dimensions
! of the variable are the same in all the files (number of records may
! be different) so the code below retains and relies on the content of
! arrays "dsize" and "vardimids" determined from the first file.


        do m=0,net_ext_count-1
          do ichunk=0,num_chunks-1
#ifdef NC4CHUNKING
            if (nc4chunked) then
              mwhblk=m/num_subblk_count        ! whole-block index
              jsbblk=m-mwhblk*num_subblk_count ! sub-blocking index

              is=mwhblk
              do j=1,vardims-1,+1            ! this procedure makes
                ie=is/nblocks(j)             ! start(j) be whole-block
                start(j)=is-ie*nblocks(j)    ! index in j-direction
                is=ie  !<-- recursive        ! defined within the range
              enddo                          !    0,...,nblocks(j)-1

! At first, set "start" and "count" for ALL spatial dimensions as to
! cover the entire block corresponding to whole-block indices defined
! above.  No further action is needed for the non-split dimensions
! (the ones below "ksplit").

              do j=1,vardims-1
                start(j)=1+start(j)*blksize(j)
                jend=min(start(j)+blksize(j)-1, vardsizes(j))
                count(j)=jend-start(j)+1
              enddo

! Split dimension: modify "start" by incrementing it by (inner, meaning
! sub-block) chunk_size every time when "ichunk" increments by 1.

              jstr=start(ksplit) + ichunk*chunk_size
              jend=min(jstr+chunk_size-1, vardsizes(ksplit))
              start(ksplit)=jstr ; count(ksplit)=jend-jstr+1

! Dimensions above "ksplit": note that ONLY ONE start(j) among j from
! [ksplit+1 <= j <= vardims-1] increments by 1 for each increment of
! counter "jsbblk". Again, take start" from above as input and modify.

              is=jsbblk
              do j=ksplit+1,vardims-1,+1
                ie=is/blksize(j)
                start(j)=start(j) +is-ie*blksize(j)
                count(j)=1
                is=ie     !<-- recursive
              enddo
            else    !--> .not.nc4chunked
#endif
              do j=1,ksplit-1
                start(j)=1 ; count(j)=vardsizes(j)
              enddo

              jstr=1+ichunk*chunk_size
              jend=min(jstr+chunk_size-1, vardsizes(ksplit))
              start(ksplit)=jstr ; count(ksplit)=jend-jstr+1

              is=m                        ! dimensions above "ksplit":
              do j=ksplit+1,vardims-1,+1            ! unfold "m" into
                ie=is/vardsizes(j)                  ! multidimensional
                start(j)=1 +is-ie*vardsizes(j)      ! indices.
                count(j)=1
                is=ie     !<-- recursive
              enddo

#ifdef NC4CHUNKING
            endif !<-- .not.nc4chunked
#endif

            nspc=1                 ! determine the length of would
            do j=1,vardims-1       ! be "spatial" dimension an array
              nspc=nspc*count(j)   ! (nspc,nrec_total)
            enddo

            size=nspc*nrec_total !<-- needed to hold all time records

            if (size < 0) then
              write(*,'(/1x,A/)') '### ERROR: integer overflow at 866.'
              stop
            endif

            if (vartype == nf_float) then
              if (size > alloc_bffr4 .or. .not.allocated(dbfr4)) then
                if (allocated(dbfr4)) deallocate(dbfr4)
                if (allocated(bffr4)) deallocate(bffr4)
                if (allocated(d2fr4)) deallocate(d2fr4)
                allocate(bffr4(size)) ; allocate(dbfr4(size))
                if (quintic) allocate(d2fr4(size))
                alloc_bffr4=size
                write(*,*) 'allocated nf_real workspace', size
              endif
            elseif (vartype == nf_double) then
              if (size > alloc_bffr8 .or. .not.allocated(dbfr8)) then
                if (allocated(dbfr8)) deallocate(dbfr8)
                if (allocated(bffr8)) deallocate(bffr8)
                if (allocated(d2fr8)) deallocate(d2fr8)
                allocate(bffr8(size)) ; allocate(dbfr8(size))
                if (quintic) allocate(d2fr8(size))
                alloc_bffr8=size
                write(*,*) 'allocated nf_double workspace', size
              endif
            endif

#define size illegal

! Note that source files may contain different numbers of records, so
! the range of "absolute" record index corresponding to each "n"th-file
! is tstr:tend computed recursively throughout the sequence.

            tend=0
            do n=1,nfiles
              start(vardims)=1 ; count(vardims)=nrecs(n)
              tstr=tend+1 ; tend=tstr+nrecs(n)-1 ; is=1+(tstr-1)*nspc
#ifdef VERBOSE
              write(*,'(1x,3A,I2,1x,A,I5,T44,A,8I5)')    'reading ''',
     &              varname(1:lvar), ''' vid =', vid, 'file n =', n,
     &                               'start =',(start(i),i=1,vardims)
              write(*,'(T44,A,8I5)') 'count =',(count(i),i=1,vardims)
#endif

              lsrc=lenstr(src_fname(n)) ; lvar=lenstr(vnames(ivar))
              ierr=nf_open(src_fname(n)(1:lsrc), nf_nowrite, ncsrc)
              if (ierr == nf_noerr) then
                ierr=nf_inq_varid(ncsrc, vnames(ivar)(1:lvar), vid)
                if (ierr == nf_noerr) then
                  if (vartype == nf_float) then
                    ierr=nf_get_vara_real (ncsrc, vid, start,count,
     &                                                   bffr4(is))
                  elseif (vartype == nf_double) then
                    ierr=nf_get_vara_double(ncsrc, vid, start,count,
     &                                                   bffr8(is))
                  else
                    write(*,'(/8x,4A/)')  '### ERROR: Unrecognized ',
     &             'type of variable ''', vnames(ivar)(1:lvar), '''.'
                    ierr=nf_noerr+1
                  endif
                  if (ierr == nf_noerr) then
#ifdef TIMING
                    nclk=3-nclk
                    call system_clock(iclk(nclk), clk_rate,clk_max)
                    inc_clk=iclk(nclk)-iclk(3-nclk)
                    if (inc_clk < 0) inc_clk=inc_clk+clk_max
                    net_clk=net_clk+inc_clk
                    net_read_clk=net_read_clk+inc_clk
                    if (vartype == nf_float) then
                      net_read_size=net_read_size+4*nspc*nrecs(n)
                    elseif (vartype == nf_double) then
                      net_read_size=net_read_size+8*nspc*nrecs(n)
                    endif
#endif
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &                       'read variable ''',  vnames(ivar)(1:lvar),
     &                              ''' from ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,6A/12x,A/)') '### ERROR: Cannot find ',
     &                      'variable ID for ''', vnames(ivar)(1:lvar),
     &                                ''' in ''', src_fname(n)(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                endif
                i=nf_close(ncsrc)
#ifdef TIMING
                nclk=3-nclk
                call system_clock(iclk(nclk), clk_rate,clk_max)
                inc_clk=iclk(nclk)-iclk(3-nclk)
                if (inc_clk < 0) inc_clk=inc_clk+clk_max
                net_clk=net_clk+inc_clk
                net_src_sync=net_src_sync+inc_clk
#endif
              else
                write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open ',
     &                          'netCDF file ''', src_fname(n)(1:lsrc),
     &                            ''' for reading.', nf_strerror(ierr)
              endif
              if (ierr /= nf_noerr) goto 99                !--> BREXIT
            enddo !<-- n=1,nfiles


! Computational part goes here.

            if (quintic) then
              if (vartype == nf_float) then
C$OMP PARALLEL SHARED(nspc, nrec_total, time, bffr4,dbfr4,d2fr4, spv4)
                call comp_quintic_real( nspc, nrec_total,
     &                                  time, bffr4,dbfr4,d2fr4, spv4)
C$OMP END PARALLEL
              elseif (vartype == nf_double) then
C$OMP PARALLEL SHARED(nspc, nrec_total, time, bffr8,dbfr8,d2fr8, spv8)
                call comp_quintic_double(nspc,nrec_total,
     &                                  time, bffr8,dbfr8,d2fr8, spv8)
C$OMP END PARALLEL
              endif
            else
              if (vartype == nf_float) then
C$OMP PARALLEL SHARED(nspc, nrec_total, time, bffr4, dbfr4, spv4)
                call comp_spline_real(  nspc, nrec_total,
     &                                  time, bffr4, dbfr4, spv4)
C$OMP END PARALLEL
              elseif (vartype == nf_double) then
C$OMP PARALLEL SHARED(nspc, nrec_total, time, bffr8, dbfr8, spv8)
                call comp_spline_double(nspc, nrec_total,
     &                                  time, bffr8, dbfr8, spv8)
C$OMP END PARALLEL
              endif
            endif

#ifdef TIMING
            nclk=3-nclk
            call system_clock(iclk(nclk), clk_rate,clk_max)
            inc_clk=iclk(nclk)-iclk(3-nclk)
            if (inc_clk < 0) inc_clk=inc_clk+clk_max
            net_clk=net_clk+inc_clk
            net_comp_clk=net_comp_clk+inc_clk
#endif

            tend=0
            do n=1,nfiles
              start(vardims)=1 ; count(vardims)=nrecs(n)
              tstr=tend+1 ; tend=tstr+nrecs(n)-1 ; is=1+(tstr-1)*nspc

              ltrg=lenstr(trg_fname(n)) ; lvar=lenstr(vnames(ivar))
              ierr=nf_open(trg_fname(n)(1:ltrg), nf_write, nctarg)
              if (ierr == nf_noerr) then
                ierr=nf_set_fill(nctarg, nf_nofill, old_fill_mode)
                if (.not.excl_vars) then
                  ierr=nf_inq_varid(nctarg,vnames(ivar)(1:lvar),vidtrg)
                  if (ierr == nf_noerr) then
                    if (vartype == nf_float) then
                      ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                             count, bffr4(is))
                    elseif (vartype == nf_double) then
                      ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                             count, bffr8(is))
                    endif
                    if (ierr == nf_noerr) then
                      write(*,'(20x,3A)')   'wrote ''',
     &                      vnames(ivar)(1:lvar), ''''
#ifdef TIMING
                      nclk=3-nclk
                      call system_clock(iclk(nclk), clk_rate,clk_max)
                      inc_clk=iclk(nclk)-iclk(3-nclk)
                      if (inc_clk < 0) inc_clk=inc_clk+clk_max
                      net_clk=net_clk+inc_clk
                      net_wrt_clk=net_wrt_clk+inc_clk
                      if (vartype == nf_float) then
                        net_wrt_size=net_wrt_size + 4*nspc*nrecs(n)
                      elseif (vartype == nf_double) then
                        net_wrt_size=net_wrt_size + 8*nspc*nrecs(n)
                      endif
#endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                       'write variable ''', vnames(ivar)(1:lvar),
     &                       ''' into ''', trg_fname(n)(1:ltrg), '''.',
     &                                               nf_strerror(ierr)
                    endif
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &                 'find variable ID for ''', vnames(ivar)(1:lvar),
     &                                ''' in ''', trg_fname(n)(1:ltrg),
     &                                         '''.',nf_strerror(ierr)
                  endif
                endif  !<-- .not.excl_vars

                if (ierr == nf_noerr) then
                  ierr=nf_inq_varid(nctarg,
     &                          'dt_'/ /vnames(ivar)(1:lvar), vidtrg)
                  if (ierr == nf_noerr) then
                    if (vartype == nf_float) then
                      ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                              count, dbfr4(is))
                    elseif (vartype == nf_double) then
                      ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                              count, dbfr8(is))
                    endif
                    if (ierr == nf_noerr) then
#ifdef NC4CHUNKING
                       if (nc4chunked) then
                         write(*,'(4I6,2x,3A)')   mwhblk+1, jsbblk+1,
     &                                    ichunk+1, n, 'wrote ''dt_',
     &                                     vnames(ivar)(1:lvar), ''''
                       else
#endif
                         write(*,'(3I6,2x,3A)')    m+1, ichunk+1, n,
     &                     'wrote ''dt_', vnames(ivar)(1:lvar), ''''
#ifdef NC4CHUNKING
                       endif
#endif
#ifdef TIMING
                      nclk=3-nclk
                      call system_clock(iclk(nclk), clk_rate, clk_max)
                      inc_clk=iclk(nclk)-iclk(3-nclk)
                      if (inc_clk < 0) inc_clk=inc_clk+clk_max
                      net_clk=net_clk+inc_clk
                      net_wrt_clk=net_wrt_clk+inc_clk
                      if (vartype == nf_float) then
                        net_wrt_size=net_wrt_size + 4*nspc*nrecs(n)
                      elseif (vartype == nf_double) then
                        net_wrt_size=net_wrt_size + 8*nspc*nrecs(n)
                      endif

                      if (ichunk==num_chunks-1 .and. n==nfiles) then
                        write(*,'(F10.1,1x,A,F12.2,1x,A,F8.1,3A)')
     &                  dble(net_clk)/dble(clk_rate),   'sec elapsed,',
     &                  dble(net_wrt_size)/1048576D0,'MBytes written,',
     &                             dble(100*(m+1))/dble(net_ext_count),
     &                  '% complete for ''', vnames(ivar)(1:lvar),''''
                      endif
#endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                             'write ''dt_', vnames(ivar)(1:lvar),
     &                              ''' into ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &              'find variable ID for ''dt_', vnames(ivar)(1:lvar),
     &                                ''' in ''', trg_fname(n)(1:ltrg),
     &                                         '''.',nf_strerror(ierr)
                  endif
                endif

                if (quintic .and. ierr == nf_noerr) then
                  ierr=nf_inq_varid(nctarg, 'd2_'/ /
     &                                   vnames(ivar)(1:lvar), vidtrg)
                  if (ierr == nf_noerr) then
                    if (vartype == nf_float) then
                        ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                               count, d2fr4(is))
                    elseif (vartype == nf_double) then
                      ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                               count, d2fr8(is))
                    endif
                    if (ierr == nf_noerr) then
                      write(*,'(20x,4A)')      'wrote ''d2_',
     &                             vnames(ivar)(1:lvar),''''
#ifdef TIMING
                      nclk=3-nclk
                      call system_clock(iclk(nclk), clk_rate,clk_max)
                      inc_clk=iclk(nclk)-iclk(3-nclk)
                      if (inc_clk < 0) inc_clk=inc_clk+clk_max
                      net_clk=net_clk+inc_clk
                      net_wrt_clk=net_wrt_clk+inc_clk
                      if (vartype == nf_float) then
                        net_wrt_size=net_wrt_size + 4*nspc*nrecs(n)
                      elseif (vartype == nf_double) then
                        net_wrt_size=net_wrt_size + 8*nspc*nrecs(n)
                      endif
#endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                             'write ''d2_', vnames(ivar)(1:lvar),
     &                              ''' into ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &              'find variable ID for ''d2_', vnames(ivar)(1:lvar),
     &                                ''' in ''', trg_fname(n)(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                  endif
                endif  !<-- quintic

                i=nf_close(nctarg)
#ifdef TIMING
                nclk=3-nclk
                call system_clock(iclk(nclk), clk_rate,clk_max)
                inc_clk=iclk(nclk)-iclk(3-nclk)
                if (inc_clk < 0) inc_clk=inc_clk+clk_max
                net_clk=net_clk+inc_clk
                net_trg_sync=net_trg_sync+inc_clk
#endif
              else
                write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open ',
     &                          'netCDF file ''', trg_fname(n)(1:ltrg),
     &                            ''' for writing.', nf_strerror(ierr)
              endif
              if (ierr /= nf_noerr) goto 99                !--> BREXIT
            enddo !<-- n=1,nfiles
          enddo  !<-- ichunk=0,num_chunks-1
        enddo  !<-- m=0,net_ext_count-1
      enddo  !<-- ivar

  99  continue
#ifdef TIMING
      call cpu_time(run_time) ; run_time=run_time-tstart
      write(*,'(/2x,A,F11.2,1x,A/)') 'cpu time =', run_time, 'sec'

      if (clk_rate > 0) then
        ReadSize=1.0D-6*net_read_size
        WrtSize=1.0D-6*net_wrt_size
        ReadTime=net_read_clk/dble(clk_rate)
        CompTime=net_comp_clk/dble(clk_rate)
        WrtTime = net_wrt_clk/dble(clk_rate)
        TargAlloc=net_trg_alloc/dble(clk_rate)
        ClsSrcTime=net_src_sync/dble(clk_rate)
        ClsTrgTime=net_trg_sync/dble(clk_rate)

        write(*,'(6x,A,F12.2,1x,A,F12.2,1x,A,F8.2,1x,A)')
     &           'Total data read :', ReadSize, 'MBytes in', ReadTime,
     &                           'sec (', ReadSize/ReadTime, 'MB/sec)'
        write(*,'(14x,A,F12.2,1x,A,F12.2,1x,A,F8.2,1x,A)') 'written :',
     &          WrtSize, 'MBytes in',  WrtTime, 'sec (',
     &                                      WrtSize/WrtTime, 'MB/sec)'
        write(*,'(12x,A,22x,F12.2,1x,A)') 'Computing :',CompTime,'sec'

        write(*,'(2x,A,22x,F12.2,1x,A)')  'Alloc. target files :',
     &                                                TargAlloc, 'sec'

        write(*,'(4x,A,22x,F12.2,1x,A)')  'Close input files :',
     &                                               ClsSrcTime, 'sec'
        write(*,'(3x,A,22x,F12.2,1x,A)') 'Close output files :',
     &                                               ClsTrgTime, 'sec'
        write(*,'(45x,A/10x,A,12x,F12.2,1x,A/)')  '------------------',
     &  'Elapsed wall-clock time', dble(net_clk)/dble(clk_rate), 'sec'
      endif
#endif
      stop
      end
