c--#define VERBOSE

! This package contains a complete set of basic operators for reading
! and writing netCDF data files coming from standard datasets, such as
! NCEP, ECMWF, TRMM, COADS, Levitus, etc.  For the "standard" implies
! that data file format has two, three, or four dimensions: lon, lat,
! [vertical], and time  dimension (usually, but not necessarily time
! is  unlimited).  The file format is presumed to be "CF-compliant",
! which implies the existence of "coordinate variables" having exactly
! the same names as the corresponding dimensions.   The names of the
! coordinates are not known in advance (except that time variable is
! identified by having prefix or suffix "time" within its name) -- they
! are identified by matching names of dimensions and one-dimensional
! variables (as expected for coordinate variables), which are then
! further examined for their role in dimensioning fields stored in the
! file.   Module "nc_file_coords" is used by this package to keep track
! of the identified coordinate dimension-variable pairs in order to
! verify consistency of between different files.

! All operations for reading and writing are instrumented to ensure
! access by name and provides error safety by matching dimensions of
! requested variable and performing all necessary error checking on
! the way. They are:

!       nc_read_grid       (fname, lon,lat, xlon,ylat)
!       nc_read_vert_coord (fname, N, zlev)

!       nc_find_dims   (fname, lon,lat,N)
!       nc_check_dims  (fname, lon,lat,N)

!       nc_read_sclr_real    (fname, vname,          rec, var)
!       nc_read_sclr_double  (fname, vname,           rec, var)
!       nc_write_sclr_real   (fname, vname,           rec, var)
!       nc_write_sclr_double (fname, vname,           rec, var)

!       nc_read_var_real    (fname, vname, lon,lat,N, rec, var)
!       nc_read_var_double  (fname, vname, lon,lat,N, rec, var)
!       nc_write_var_real   (fname, vname, lon,lat,N, rec, var)
!       nc_write_var_double (fname, vname, lon,lat,N, rec, var)

! where "fname" is name of netCDF file; "lon","lat" are dimensions
! ("nc_find_dims" reads them from file and outputs; all others expect
! them as input arguments); "nc_check_dims" just checks whether input
! values of "lon,lat,N" match corresponding dimensions inside netCDF
! file and terminates if mismatch occurs; one-dimensional arrays
! "xlon,ylat" are grid coordinates of a separable spherical Mercator
! grid provided by NCEP, NCAR, Levitus, or other alike CF-compliant
! datasets;  Reading vertical coordinate (typically z-like) is
! separated from the horizontal because most likely use of this is to
! prepare surface forcing data where there is not vertical dimension.
! All the other operations "nc_[read/write]_var_type" are generic read
! or write for two- or three-dimensional field, which is also assumed
! to have record-dimension (time); "rec" is index of record within
! time-dimension; only one record is read at a time; "var" is a
! user-provided array to hold data, which is assumed to be single
! precision.

! NOTES: (1) For all operations within this package the access model
!            is by file name -- variable name, that is all netCDF file
!            are expected to pre-exist along with the matching
!            variables, and be in not-open state -- they these routines
!            open them internally and close upon completion);

!          [ Alternatively, access by netCDF file ID -- variable name
!            with netCDF files open externally (and staying open) is
!            provided by get/put_var/sclr/rec_by_name_type routines
!            from roms_read_write.F. ]

!        (2) Error handling: all these operations are designed to
!            write a diagnostic error message and stop immediately
!            -- thus terminating the execution of calling program.
!            This way (as opposite to returning non-zero error
!            status) is motivated to minimize programming effort in
!            calling program by completely eliminating the need for
!            error handling there.



#if !defined WRITER && !defined DOUBLE

      module nc_file_coords
       integer x_lon_size, y_lat_size, vert_size, time_size
       character(len=16) x_lon_coord, y_lat_coord, vert_coord,
     &                                             time_coord
      end module nc_file_coords


      subroutine nc_read_grid(fname, lon,lat, xlon,ylat)

! Read coordinates arrays xlon,ylat. Their sizes and the corresponding
! names of netCDF variable names are assumed to be known at this moment
! (stored as x_lon_coord and y_lat_coord in module nc_file_coords) from
! the preceding call to nc_find_dims.

      use nc_file_coords
      implicit none
      character(len=*) fname
      integer lon,lat
      real(kind=8) xlon(lon),ylat(lat)
!>
      integer ncid,varid, ierr, lfnm, lvrX,lvrY, lenstr
# include "netcdf.inc"

      lfnm=lenstr(fname)
      call nc_check_dims(fname, lon,lat,0)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        lvrX=lenstr(x_lon_coord)
        ierr=nf_inq_varid(ncid, x_lon_coord(1:lvrX), varid)
        if (ierr == nf_noerr) then
          ierr=nf_get_var_double(ncid, varid, xlon)
          if (ierr == nf_noerr) then
            lvrY=lenstr(y_lat_coord)
            ierr=nf_inq_varid(ncid, y_lat_coord(1:lvrY), varid)
            if (ierr == nf_noerr) then
              ierr=nf_get_var_double (ncid, varid, ylat)
              if (ierr == nf_noerr) then
                ierr=nf_close (ncid)
                if (ierr == nf_noerr) then
                  write(*,'(1x,7A)')   'retrieved grid coordinates ''',
     &           x_lon_coord(1:lvrX), ''' and ''', y_lat_coord(1:lvrY),
     &                              ''' from ''', fname(1:lfnm), '''.'
                  return !--> successful return
                else
                  write(*,'(/1x,4A,1x,A/)') '### ERROR: Cannot close ',
     &                         'netCDF file ''',  fname(1:lfnm), ''':',
     &                                               nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot read ',
     &                   'coordinate variable ''', y_lat_coord(1:lvrY),
     &           ''' from ''', fname(1:lfnm),''': ', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,4A/12x,A/)')      '### ERROR: Cannot get ',
     &                'netCDF ID for variable ''', y_lat_coord(1:lvrY),
     &           ''' from ''', fname(1:lfnm),''': ', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,6A/12x,A/)')       '### ERROR: Cannot read ',
     &                   'coordinate variable ''', x_lon_coord(1:lvrX),
     &          ''' from ''', fname(1:lfnm), ''': ', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot get netCDF ',
     &                       'ID for variable ''', x_lon_coord(1:lvrX),
     &           ''' from ''', fname(1:lfnm),''': ', nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)')   '### ERROR: Cannot open netCDF ',
     & 'file ''', fname(1:lfnm), ''' for reading.', nf_strerror(ierr)
      endif
      stop
      end

      subroutine nc_read_vert_coord(fname, N, zlev)
      use nc_file_coords
      implicit none
      character(len=*) fname
      integer N,  ierr, ncid, varid, ndims, dimid(2),
     &                  size, lnm, lfnm,lvar, lenstr
      real(kind=8) zlev(N)
      character(len=32) name
# include "netcdf.inc"

      lfnm=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        lvar=lenstr(vert_coord)
        ierr=nf_inq_varid(ncid, vert_coord(1:lvar), varid)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varndims(ncid, varid, ndims)
          if (ierr == nf_noerr .and. ndims == 1) then
            ierr=nf_inq_vardimid(ncid, varid, dimid)
            if (ierr == nf_noerr) then
              ierr=nf_inq_dim(ncid, dimid(1), name, size)
              if (ierr == nf_noerr) then
                lnm=lenstr(name)
                if (name(1:lnm) == vert_coord(1:lvar)) then
                  ierr=nf_get_var_double(ncid, varid, zlev)
                  if (ierr == nf_noerr) then
                    ierr=nf_close(ncid)
                    if (ierr == nf_noerr) then
                      write(*,*)   'retrieved vertical coordinate ''',
     &                               vert_coord(1:lvar), ''' from ''',
     &                                           fname(1:lfnm), '''.'

                      return !--> successful return
                    else
                      write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot ',
     &                           'close netCDF file ''', fname(1:lfnm),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  else
                    write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &               'read coordinate variable ''', vert_coord(1:lvar),
     &           ''' from ''', fname(1:lfnm), '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(1x,6A)')  '### ERROR: nc_read_vert_coord ',
     &          ':: coordinate variable -- dimension name mismatch ''',
     &              vert_coord(1:lvar), ''' vs ''', name(1:lnm), '''.'
                endif
              else
                write(*,'(/1x,2A,I3,1x,4A/12x,4A/)')   '### ERROR: ',
     &           'Cannot get name and size of dimension #', dimid(1),
     &                 'for variable ''', vert_coord(1:lvar),  ''' ',
     &                    'from netCDF',  'file ''',  fname(1:lfnm),
     &                    ''': ',          nf_strerror(ierr)
              endif

            else
              write(*,'(/1x,4A/12x,4A/)')     '### ERROR: Cannot get ',
     &             'dimension IDs for variable ''', vert_coord(1:lvar),
     &                 ''' from netCDF file',    '''',   fname(1:lfnm),
     &                                       ''': ', nf_strerror(ierr)
            endif
          elseif (ierr == nf_noerr) then
            write(*,'(/1x,5A/)')   '### ERROR: Coordinate variable ''',
     &                   vert_coord(1:lvar), ''' in ''', fname(1:lfnm),
     &                              ''' has other than one dimension.'
          else
            write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot determine ',
     &      'number of dimensions for variable ''', vert_coord(1:lvar),
     &             ''' in ''', fname(1:lfnm), '''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot get netCDF ',
     &          'ID for variable ''', vert_coord(1:lvar), ''' from ''',
     &                         fname(1:lfnm), '''.', nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)')    '### ERROR: Cannot open netCDF ',
     &  'file ''', fname(1:lfnm), ''' for reading.', nf_strerror(ierr)
      endif
      stop
      end
#endif /* !WRITER && !DOUBLE */


#ifndef DOUBLE
# ifndef CHECK_ONLY
      subroutine nc_find_dims(fname, lon,lat,N)
# else
      subroutine nc_check_dims(fname, lon,lat,N)
# endif
      use nc_file_coords
      implicit none
      character(len=*) fname
      integer  ncid, lon,lat,N
!>
      integer x_lon,y_lat,n_z, ndims,nvars,ngatts, unlimdimid,
     &        varid,vtype,vatts, ierr, i,j, lfnm,ldim,lvar,lenstr
      integer, allocatable, dimension(:,:) :: dimids
      integer, allocatable, dimension(:) :: dsize, vdims, mcoord
      character(len=16), allocatable, dimension(:) :: dname,vname

# include "netcdf.inc"

! At first, initialize dimensions to zeros: if not found their
! remaining zero status can be interpreted as error flag by the
! caller; the last three are the trial dimensions for checking;

# ifndef CHECK_ONLY
      lon=0   ; lat=0   ; N=0
# endif
      x_lon=0 ; y_lat=0 ; n_z=0

      lfnm=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        ierr=nf_inq(ncid, ndims, nvars, ngatts, unlimdimid)
        if (ierr == nf_noerr) then
          allocate(dname(ndims), vname(nvars))
          allocate(dsize(ndims), mcoord(ndims), vdims(nvars),
     &                                   dimids(ndims,nvars))
          do i=1,ndims
            mcoord(i)=0 ; dname(i)='       ' !<-- initialize to zero
            ierr=nf_inq_dim(ncid, i, dname(i), dsize(i))
            if (ierr /= nf_noerr) then
              write(*,'(/1x,3A,I3,1x,A/29x,3A/29x,A)') '### ERROR: ',
# ifndef CHECK_ONLY
     &                'nc_check_dims :: ',
# else
     &                 'nc_find_dims :: ',
# endif
     &                'Cannot find name and size of dimension #',
     &                 i, 'in',  'netCDF file ''', fname(1:lfnm),
     &                                  ''':', nf_strerror(ierr)
              stop
# ifdef VERBOSE
            else
              ldim=lenstr(dname(i))
              write(*,*) i, dname(i)(1:ldim), ' =', dsize(i)
# endif
            endif
          enddo

! Mapping physical dimensions onto netCDF file dimensions:
!-------- -------- ---------- ---- ------ ---- ------------
! Fields stored as arrays in netCDF files are expected to be f(X,Y),
! or f(X,Y,Z), or f(X,Y,T), or f(X,Y,Z,T), where the names of specific
! dimensions are not known in advance.   All what is expected is that
! the file is CF-compliant, that is for each coordinate there is exist
! a netCDF variable with the same name, X(X), Y(Y), Z(Z), and T(T).
! However dimensions in netCDF file can be declared in arbitrary order
! (not necessarily X,Y,Z,T), while the outcome of this routine should
! identify X as corresponding to the first spatial coordinate (the
! innermost index in Fortran netCDF convention of array "f") Y is the
! second; Z (if exists) is the third, and T is the last (the fourth or
! the third). So which dimension is which is decided based on how the
! dimensions appear in a two- or three-dimensional field. To do so:

! 1. Scan all one-dimensional variables in order to identify coordinate
! variables among them (as this term defined by CF standard - variables
! having the same name as their only dimension; the corresponding
! dimension is then labelled by mcoord(i)=-1 where "i" is dimension id.

          do varid=1,nvars
            ierr=nf_inq_var(ncid, varid, vname(varid),  vtype,
     &                   vdims(varid), dimids(1,varid), vatts)
            if (ierr == nf_noerr) then
              lvar=lenstr(vname(varid))
              if (vdims(varid) == 1) then
                i=dimids(1,varid) ; ldim=lenstr(dname(i))
                if ( lvar == ldim .and. vname(varid)(1:lvar)
     &                    == dname(i)(1:ldim) ) mcoord(i)=-1
# ifdef VERBOSE
                write(*,'(2I3,2(1x,A),I5,I3)') varid, i,
     &          vname(varid), dname(i), dsize(i), mcoord(i)
# endif
              endif
            else
              write(*,'(/1x,3A,I3,1x,A/29x,3A/29x,A)') '### ERROR: ',
# ifndef CHECK_ONLY
     &                'nc_check_dims :: ',
# else
     &                 'nc_find_dims :: ',
# endif
     &                'Cannot make general inquiry for variable #',
     &                 varid, 'in', 'netCDF file ''',fname(1:lfnm),
     &                                   ''':',  nf_strerror(ierr)
              stop
            endif
          enddo

! 2. Once the coordinate variables and dimensions are identified,
! examine variables which have more than one dimension to establish
! the order of dimensions in the arrays. As the result each coordinate
! dimension will be assigned rank equal to the maximum index number,
! say, if there are two arrays a(X,Y,T) and b(X,Y,Z,T) then  X -> 1,
! Y -> 2, Z -> 3, T -> 4 because T appears a index number 3 and 4, so
! the maximum is 4;   However note that coordinate dimensions can be
! preceeded by non-coordinate dimensions, like in d(n,X) or e(X,n,Y)
! where "n" is a non-coordinate, so existence of such variables should
! not increase the ranks of dimensions X and Y beyond 1 and 2
! respectively in this example.  To achieve this the code segment
! below uses "j" to count only coordinate dimensions and ignore
! non-coordinates.

          do varid=1,nvars
            if (vdims(varid) > 1) then
              j=0
              do i=1,vdims(varid)
                if (mcoord(dimids(i,varid)) /= 0) then
                  j=j+1
                  if (mcoord(dimids(i,varid)) < j) then
                    mcoord(dimids(i,varid))=j
                  endif
                endif
              enddo
            endif
          enddo

# ifdef VERBOSE
          write(*,*) 'summary of dimensions in ''',fname(1:lfnm),''''
          write(*,'(10x,A,8x,A,8x,A,1x,A)') 'id','rank','size','name'
          write(*,'(A)') ' -----------------------------------------'
          do i=1,ndims
            write(*,*) i, mcoord(i), dsize(i), dname(i)
          enddo
# endif

! Check that there is no confusion in determining which dimension plays
! the role of X and which of Y: strictly speaking the logical condition
! below is over-restrictive because a situation where there are two
! variables f(X,T) and g(Y,Z) will be flagged as error, when, in fact
! it is theoretically possible (although does not occur in practical
! datasets).

          ierr=0
          do j=2,ndims
            do i=1,j-1
              if (mcoord(i) == mcoord(j) .and. mcoord(i) < 3) then
                write(*,'(/1x,3A,I3,1x,A,I3/8x,4A,I3/)') '### ERROR: ',
# ifndef CHECK_ONLY
     &                'nc_check_dims :: ',
# else
     &                 'nc_find_dims :: ',
# endif
     &                'Ambiguous role of dimensions #', i, 'and #', j,
     &                'in netCDF file ''', fname(1:lfnm), ''': both ',
     &                            'appear as coordinate #', mcoord(i)
                ierr=ierr+mcoord(i)
              endif
            enddo
          enddo
          if (ierr /= 0) stop

! The above, however may still lead to an ambiguity because, if, for
! example, there are two arrays, a(X,Y,T) and b(X,Y,Z) then both Z and
! T are assigned rank 3. This is resolved by identifying time variable
! by the name first, and then identifying Z are the other (remaining)
! variable having rank 3 (if it exists, otherwise return Z=0).

          do i=1,ndims
            if (mcoord(i) > 0) then
              ldim=lenstr(dname(i))
              if ( dname(i)(max(1,ldim-3):ldim) == 'time' .or.
     &                             dname(i)(1:4) == 'time' .or.
     &          (ldim == 1 .and. dname(i)(1:1) == 'T')  ) then
# ifdef VERBOSE
                write(*,'(I4,1x,4A,I4)') i, 'Found time coordinate ',
     &                'variable ', dname(i)(1:ldim), ' =', dsize(i)
# endif
# ifndef CHECK_ONLY
                time_size=dsize(i)  ; time_coord=dname(i)(1:ldim)
# endif
              else
# ifdef VERBOSE
                write(*,'(I4,1x,2A,I3,2(1x,A),I4)') i, 'Found ',
     &                       'coordinate variable #', mcoord(i),
     &                        dname(i)(1:ldim),  '=',  dsize(i)
# endif
                if (mcoord(i) == 3) then
                  n_z=dsize(i)
# ifndef CHECK_ONLY
                  vert_size=dsize(i)  ; vert_coord=dname(i)(1:ldim)
# endif
                elseif (mcoord(i) == 1) then
                  x_lon=dsize(i)
# ifndef CHECK_ONLY
                  x_lon_size=dsize(i) ; x_lon_coord=dname(i)(1:ldim)
# endif
                elseif (mcoord(i) == 2) then
                  y_lat=dsize(i)
# ifndef CHECK_ONLY
                  y_lat_size=dsize(i) ; y_lat_coord=dname(i)(1:ldim)
# endif
                endif
              endif   !<-- time dimension or spatial dimensions
            endif   !<-- mcoord > 0
          enddo

# ifndef CHECK_ONLY
          ierr=nf_close(ncid)
          if (ierr == nf_noerr) then
            lon=x_lon ; lat=y_lat ; N=n_z

            return !--> successful return
# else
          if (x_lon /= lon .or. y_lat /= lat) then
            write(*,'(/1x,2A,2I4,1x,3A,2I4)')          '### ERROR: ',
     &       'nc_check_dims :: previously found lon,lat =', lon, lat,
     &       'contradict with same from netCDF file ''',
     &                      fname(1:lfnm), ''', X,Y =', x_lon, y_lat
            ierr=1
          endif
          if (N > 0  .and.  n_z > 0  .and.  N /= n_z) then
            write(*,'(1x,3A,I4/12x,3A,I4/)')           '### ERROR: ',
     &       'nc_check_dims :: previously found vertical dimension ',
     &       'N =', N,  'contradicts with same from netCDF file  ''',
     &                                fname(1:lfnm), ''', Z =', n_z
            ierr=ierr+1
          endif
          if (ierr == nf_noerr) return !--> successful return
# endif
# ifndef CHECK_ONLY
          else
            write(*,'(/1x,4A,1x,A/)')    '### ERROR: Cannot close ',
     &                   'netCDF file ''', fname(1:lfnm), ''':',
     &                                            nf_strerror(ierr)
          endif
# endif
        else
          write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot determine ',
     &                     'number of dimensions in netCDF file ''',
     &                      fname(1:lfnm), ''':', nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A,1x,A/)')  '### ERROR: Cannot open netCDF ',
     &           'file ''', fname(1:lfnm), ''':', nf_strerror(ierr)
      endif
      stop
      end
#endif /* DOUBLE */


#undef KIND_TYPE
#undef nf_get_var_TYPE
#undef nf_put_var_TYPE
#undef nf_get_vara_TYPE
#undef nf_put_vara_TYPE
#undef nc_read_sclr_TYPE
#undef nc_write_sclr_TYPE
#undef nc_read_var_TYPE
#undef nc_write_var_TYPE


#ifdef DOUBLE
# define KIND_TYPE 8
# define nf_get_var_TYPE nf_get_var_double
# define nf_put_var_TYPE nf_put_var_double
# define nf_get_vara_TYPE nf_get_vara_double
# define nf_put_vara_TYPE nf_put_vara_double
# define nc_read_sclr_TYPE nc_read_sclr_double
# define nc_write_sclr_TYPE nc_write_sclr_double
# define nc_read_var_TYPE nc_read_var_double
# define nc_write_var_TYPE nc_write_var_double
#else
# define KIND_TYPE 4
# define nf_get_var_TYPE nf_get_var_real
# define nf_put_var_TYPE nf_put_var_real
# define nf_get_vara_TYPE nf_get_vara_real
# define nf_put_vara_TYPE nf_put_vara_real
# define nc_read_sclr_TYPE nc_read_sclr_real
# define nc_write_sclr_TYPE nc_write_sclr_real
# define nc_read_var_TYPE nc_read_var_real
# define nc_write_var_TYPE nc_write_var_real
#endif


! Retrieve/write a record of scalar variable

! error #8284: If the actual argument is scalar, the dummy argument
! shall be scalar unless the actual argument is of type character or
! is an element of an array that is not assumed shape, pointer, or
! polymorphic.



#ifndef WRITER
      subroutine  nc_read_sclr_TYPE(fname, vname, rec, var)
#else
      subroutine nc_write_sclr_TYPE(fname, vname, rec, var)
#endif
      implicit none
      character(len=*) fname, vname
      integer rec
      real(kind=KIND_TYPE) var

      integer ncid, varid, ndims, start(4),count(4), ierr,
#ifndef WRITER
     &                                     dimid(4), size,
#endif
     &                                 lfnm, lvar, lenstr

#include "netcdf.inc"

      lfnm=lenstr(fname) ; lvar=lenstr(vname)

#ifndef WRITER
      ierr=nf_open(fname, nf_nowrite, ncid)
#else
      ierr=nf_open(fname, nf_write,   ncid)
#endif
      if (ierr == nf_noerr) then
        ierr=nf_inq_varid(ncid, vname, varid)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varndims(ncid, varid, ndims)
          if (ierr == nf_noerr) then
            if (ndims == 1) then
              start(1)=rec ; count(1)=1
#ifndef WRITER
              ierr=nf_inq_vardimid(ncid, varid, dimid)
              if (ierr == nf_noerr) then
                ierr=nf_inq_dimlen(ncid, dimid(1), size)
                if (ierr == nf_noerr) then
                  if (0 < rec .and. rec <= size) then
                    ierr=nf_get_vara_TYPE(ncid,varid,start,count,var)
#else
                    ierr=nf_put_vara_TYPE(ncid,varid,start,count,var)
#endif
                    if (ierr == nf_noerr) then
                      ierr=nf_close (ncid)
                      if (ierr == nf_noerr) then
#ifndef WRITER
                        write(*,'(6x,A,I4,1x,5A)')    'read rec', rec,
     &                     'of scalar ''', vname(1:lvar), ''' from ''',
#else
                        write(*,'(5x,A,I4,1x,5A)')    'wrote rec', rec,
     &                     'of scalar ''', vname(1:lvar), ''' into ''',
#endif
     &                                             fname(1:lfnm), ''''
                        return !--> successful return
                      else
                        write(*,'(/1x,5A/)')      '### ERROR: Cannot ',
     &                  'close netCDF file ''',  fname(1:lfnm), ''': ',
     &                                               nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,7A/12x,A)')   '### ERROR: Cannot ',
#ifndef WRITER
     &                  'read variable ''',  vname(1:lvar), ''' from ',
#else
     &                  'write variable ''', vname(1:lvar), ''' into ',
#endif
     &                  'netCDF file ''',    fname(1:lfnm),      ''':',
     &                                               nf_strerror(ierr)
                    endif
#ifndef WRITER
                  else
                    write(*,'(/1x,2A,I4,1x,6A,I4/)')     '### ERROR: ',
     &               'Requested record number ',   rec,  'for scalar ',
     &               'variable ''',   vname(1:lvar),  ''' in file ''',
     &               fname(1:lfnm), ''' exceeds dimension bound', size
                  endif
                else
                  write(*,'(/1x,2A,I3,1x,3A/12x,4A/)')  '### ERROR: ',
     &               'Cannot determine size of dimension #', dimid(1),
     &               'from netCDF file ''',    fname(1:lfnm),  ''': ',
     &                                              nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,4A/12x,4A/)')   '### ERROR: Cannot get ',
     &                'dimension IDs for variable ''',   vname(1:lvar),
     &                 ''' from netCDF file',    '''',   fname(1:lfnm),
     &                                       ''': ', nf_strerror(ierr)
              endif
#endif
            elseif (ndims == 0) then
#ifndef WRITER
              ierr=nf_get_var_TYPE(ncid, varid, var)
#else
              ierr=nf_put_var_TYPE(ncid, varid, var)
#endif
              if (ierr == nf_noerr) then
                ierr=nf_close (ncid)
                if (ierr == nf_noerr) then
#ifndef WRITER
                  write(*,'(6x,5A)')       'read scalar variable ''',
     &              vname(1:lvar), ''' from ''', fname(1:lfnm), ''''
#else
                  write(*,'(5x,5A)')      'wrote scalar variable ''',
     &              vname(1:lvar), ''' into ''', fname(1:lfnm), ''''
#endif
                  return !--> successful return
                else
                  write(*,'(/1x,5A/)')      '### ERROR: Cannot close',
     &                       'netCDF file ''',  fname(1:lfnm), ''': ',
     &                                              nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,7A/12x,A)')        '### ERROR: Cannot ',
#ifndef WRITER
     &          ' read scalar variable ''', vname(1:lvar), ''' from ',
#else
     &          'write scalar variable ''', vname(1:lvar), ''' into ',
#endif
     &                'netCDF file ''',    fname(1:lfnm),      ''':',
     &                                            nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,5A,I4/)')  '### ERROR: Expected scalar ''',
     &                 vname(1:lvar), ''' from file ''', fname(1:lfnm),
     &                ''' has more then one dimension, ndims =', ndims
            endif
          else
            write(*,'(/1x,4A/12x,4A/)') '### ERROR: Cannot determine ',
     &           'number of dimensions for variable ''', vname(1:lvar),
     &             ''' from netCDF', 'file ''', fname(1:lfnm),  ''': ',
     &                                              nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,6A/12x,A)')   '### ERROR: Cannot get netCDF ',
     &         'ID for variable ''', vname(1:lvar), ''' from file ''',
     &                       fname(1:lfnm), ''':',  nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot open netCDF ',
     &          'file ''', fname(1:lfnm), ''':', nf_strerror(ierr)
      endif
      stop
      end


#ifndef WRITER
      subroutine  nc_read_var_TYPE(fname, vname, lon,lat,N, rec, var)
#else
      subroutine nc_write_var_TYPE(fname, vname, lon,lat,N, rec, var)
#endif
      use mod_io_size_acct
      use nc_file_coords

      implicit none
      character(len=*) fname, vname
      integer lon,lat,N, rec
      real(kind=KIND_TYPE) var(*)
!>
      integer ncid, varid, ndims, dimid(4), start(4),count(4),
     &        i, size, rec_size, ierr, lnm,lfnm,lvar, lenstr
      logical matched_dims
      character(len=8) name

#include "netcdf.inc"

      lfnm=lenstr(fname) ; lvar=lenstr(vname)

#ifndef WRITER
      ierr=nf_open(fname, nf_nowrite, ncid)
#else
      ierr=nf_open(fname, nf_write,   ncid)
#endif
      if (ierr == nf_noerr) then
        ierr=nf_inq_varid(ncid, vname, varid)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varndims(ncid, varid, ndims)
          if (ierr == nf_noerr) then
            ierr=nf_inq_vardimid(ncid, varid, dimid)
            if (ierr == nf_noerr) then
              matched_dims=.true.
              do i=1,ndims
                start(i)=0 ; count(i)=0
                ierr=nf_inq_dim(ncid, dimid(i), name, size)
                if (ierr == nf_noerr) then
                  lnm=lenstr(name)
                  if (name(1:lnm) == x_lon_coord) then
                    if (size == lon) then
                       start(i)=1 ; count(i)=lon
                    else
                      write(*,'(/1x,5A/27x,3A/27x,A,2(I5,1x,A)/)')
#ifndef WRITER
     &           '### ERROR: nc_read_var :: Mismatch in dimension ''',
#else
     &           '### ERROR: nc_write_var :: Mismatch in dimension ''',
#endif
     &                name(1:lnm), ''' of variable ''', vname(1:lvar),
     &                '''', 'in netCDF file ''', fname(1:lfnm), ''':',
     &                           'expected',  lon, 'instead of', size,
     &                                           'found in the file.'
                      matched_dims=.false.
                    endif
                  elseif (name(1:lnm) == y_lat_coord) then
                    if (size == lat) then
                       start(i)=1 ; count(i)=lat
                    else
                      write(*,'(/1x,5A/27x,3A/27x,A,2(I5,1x,A)/)')
#ifndef WRITER
     &           '### ERROR: nc_read_var :: Mismatch in dimension ''',
#else
     &           '### ERROR: nc_write_var :: Mismatch in dimension ''',
#endif
     &                name(1:lnm), ''' of variable ''', vname(1:lvar),
     &                '''', 'in netCDF file ''', fname(1:lfnm), ''':',
     &                           'expected',  lat, 'instead of', size,
     &                                           'found in the file.'
                      matched_dims=.false.
                    endif
                  elseif (name(1:lnm) == vert_coord) then
                    if (size == N) then
                      start(i)=1 ; count(i)=N
                    else
                      write(*,'(/1x,5A/27x,3A/27x,A,2(I5,1x,A)/)')
#ifndef WRITER
     &           '### ERROR: nc_read_var :: Mismatch in dimension ''',
#else
     &           '### ERROR: nc_write_var :: Mismatch in dimension ''',
#endif
     &                name(1:lnm), ''' of variable ''', vname(1:lvar),
     &                '''', 'in netCDF file ''', fname(1:lfnm), ''':',
     &                            'expected',  N, 'instead of', size,
     &                                          'found in the file.'
                      matched_dims=.false.
                    endif
                  elseif (name(1:lnm) == time_coord) then
                    if (i == ndims) then
                      start(i)=rec         ! here it is expected that
                      count(i)=1           ! record dimension is always
                    endif                  ! the last dimension.
                  endif
                else
                  write(*,'(/1x,2A,I3,1x,4A/12x,4A/)') '### ERROR: ',
     &                    'Cannot get name and size of dimension', i,
     &                    'for variable ''', vname(1:lvar),  ''' ',
     &                    'from netCDF',  'file ''',  fname(1:lfnm),
     &                    ''': ',          nf_strerror(ierr)
                  matched_dims=.false.
                endif
              enddo
              if (matched_dims) then
                do i=1,ndims
                  if (start(i) == 0 .or. count(i) == 0) then
                    write(*,'(/1x,2A,I3,1x,5A/)')
#ifndef WRITER
     &        '### ERROR: nc_read_var :: Cannot identify dimension #',
#else
     &        '### ERROR: nc_write_var :: Cannot identify dimension #',
#endif
     &                 i, 'for variable ''', vname(1:lvar), ''' in ''',
     &                                           fname(1:lfnm),  '''.'
                     matched_dims=.false.
                  endif
                enddo
              endif
              if (matched_dims) then
                if (start(ndims) == 0) then
                  start(ndims)=rec ; count(ndims)=1
                endif
#ifndef WRITER
                ierr=nf_get_vara_TYPE(ncid,varid, start,count, var)
#else
                ierr=nf_put_vara_TYPE(ncid,varid, start,count, var)
#endif
                if (ierr == nf_noerr) then
                  ierr=nf_close (ncid)
                  if (ierr == nf_noerr) then
                    rec_size=1
                    do i=1,ndims
                      rec_size=rec_size*count(i)
                    enddo
#ifndef WRITER
                    sz_write_acc = sz_write_acc + rec_size * KIND_TYPE
                    write(*,'(6x,A,I4,1x,5A)')  'read rec', rec,
     &                        'of ''', vname(1:lvar), ''' from ''',
#else
                    sz_read_acc = sz_read_acc + rec_size * KIND_TYPE
                    write(*,'(5x,A,I4,1x,5A)')      'wrote rec', rec,
     &                          'of ''', vname(1:lvar), ''' into ''',
#endif
     &                                          fname(1:lfnm), ''''
                    return !--> successful return
                  else
                    write(*,'(/1x,5A/)')  '### ERROR: Cannot close',
     &                            'netCDF file ''',  fname(1:lfnm),
     &                            ''': ',  nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,7A/12x,A)')     '### ERROR: Cannot ',
#ifndef WRITER
     &                'read variable ''',  vname(1:lvar), ''' from ',
#else
     &                'write variable ''', vname(1:lvar), ''' into ',
#endif
     &                'netCDF file ''',    fname(1:lfnm),      ''':',
     &                                            nf_strerror(ierr)
                endif
              else                  ! there is no need to write an
                ierr=+1000          ! error message here because it
              endif                 ! is already done before.
            else
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot get ',
     &                 'dimension IDs for variable ''', vname(1:lvar),
     &          ''' from ''', fname(1:lfnm), '''.', nf_strerror(ierr)
            endif
          else
            write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot determine ',
     &           'number of dimensions for variable ''', vname(1:lvar),
     &          ''' from ''', fname(1:lfnm), '''. ', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,6A/12x,A)')   '### ERROR: Cannot get netCDF ',
     &         'ID for variable ''', vname(1:lvar), ''' from file ''',
     &                       fname(1:lfnm), '''.',  nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot open netCDF ',
     &          'file ''', fname(1:lfnm), '''.', nf_strerror(ierr)
      endif
      stop
      end

#if !defined WRITER && !defined DOUBLE
# define WRITER
# define CHECK_ONLY
# include "nc_read_write.F"
# undef WRITER
# define DOUBLE
# include "nc_read_write.F"
# define WRITER
# include "nc_read_write.F"
#endif
