      program ncchname

! Shorthand for "netCDF--change--name", a generic tool to rename an
! object within a netCDF file. This is previously known as "ncrename",
! however because of name conflict with a standard operator from NCO
! package, the name has been changed.
!
! Should be used as:
!
!           ncchname name newname file.nc
!           ncchname name newname file1.nc file2.nc ...
!           ncchname name newname file*.nc
!
! i.e., needs 3 or more arguments, which are
!
!          'name'     is name of object to be renamed
!          'newname'  is the new name of that object
!          'file.nc', file1.nc, etc, are the names of netCDF files
!
! The object may be either a variable, dimension, global attribute or
! an attribute to a variable.  In the last case, if more than one
! variable has attribute named "name", the matching attributes to
! a variables involved will be renamed.
! NOTE: in the case where more than one type of object exists under
! the same name, for example a variable and a dimension share the same
! name, this algorithm renames only one type of object, according to
! the precedence of types: variable --> dimension --> global attribute
! --> attribute of a variable.

      implicit none
      character(len=128) fname, vname, newname, name
      integer nargs, iarg, ierr, ncid, varid,  nvars,
     &        lstr,  lfmn, lvar, lnew, lenstr, iargc
      logical fnd_var, fnd_dim, fnd_att
#include "netcdf.inc"
                                    ! Check that the correct number
      nargs=iargc()                 ! of arguments is supplied, if not 
      if (nargs >= 3) then          ! print instruction how to use and 
        call getarg(1,vname)        ! quit. Otherwise open netCDF file
        lvar=lenstr(vname)          ! and switch it to redefine mode.
        call getarg(2,newname)      ! Diagnose and complain about
        lnew=lenstr(newname)        ! errors, if something goes wrong.
      else
       write(*,'(/1x,A/3(/15x,A))') 'Usage of this command should be:',
     &                   'ncchname name newname file.nc',
     &                   'ncchname name newname file1.nc file2.nc ...',
     &                   'ncchname name newname file*.nc'
       write(*,'(1x,A/8x,A,4(/15x,A),2(/ /5x,A)/15x,A)')     'where',
     &'''name'' is the name of the object to be renamed, which may be',
     &      'either a variable, a dimension, or an attribute (global,',
     &      'or an attribute to a variable, or both - if more than',
     &      'one variable has attribute with the matching name, then',
     &      'all of them will be renamed);',
     &'''newname'' is the desired name of the object;',
     &'''file.nc'', ''file1.nc'', etc. are the names of netCDF files',
     &                          'to process (wildcards are accepted).'

       write(*,'(4(/1x,2A)/)')        'Overall behavior and safety: ',
     & 'a message is printed for every an object',       'found and ',
     & 'renamed; if an attempt to change a name would cause conflict',
     &    'with the name of another object, no action will be taken ',
     &           'and an error',  'message will be printed instead.'

       write(*,'(/1x,2A/7x,2A/7x,A/)')   'NOTE: This was previously ',
     &  'known as "ncrename", however, its name has', 'been changed ',
     &           'because of conflict with a standard operator from',
     &                                                'NCO package.'
        stop
      endif

      do iarg=3,nargs
        call getarg(iarg,fname)
        lfmn=lenstr(fname)
        ierr=nf_open (fname(1:lfmn), nf_write, ncid)
        if (ierr == nf_noerr) then
          ierr=nf_redef (ncid)
          if (ierr /= nf_noerr) then
            write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot switch ',
     &                                 'netCDF file ''', fname(1:lfmn),
     &                 ''' into redefinition mode.', nf_strerror(ierr)
            ierr=nf_close (ncid)
            stop
          endif
        else
          write(*,'(/1x,4A/12x,A/)')         '### ERROR: Cannot open ',
     &              'netCDF file ''', fname(1:lfmn), ''' for writing.',
     &                                               nf_strerror(ierr)
          stop
        endif

! Check, whether object to be renamed is a variable and if it is,
! rename it as requested.

        fnd_var=.false.
        ierr=nf_inq_varid (ncid, vname(1:lvar), varid)
        if (ierr == nf_noerr) then
          fnd_var=.true.
          ierr=nf_rename_var (ncid, varid, newname(1:lnew))
          if (ierr == nf_noerr) then
            write(*,'(/8x,7A/)')  'Renamed variable ''', vname(1:lvar),
     &       ''' into ''', newname(1:lnew), ''' inside netCDF file ''',
     &                                            fname(1:lfmn), '''.'
          elseif (ierr. eq. nf_enameinuse) then
            write(*,'(/1x,8A/12x,A/)')     '### ERROR: Cannot rename ',
     &     'variable ''', vname(1:lvar), ''' into ''', newname(1:lnew),
     &               ''' inside netCDF file ''',  fname(1:lfmn), ''',',
     &               'This name is already used by another variable. '
          else
             write(*,'(/1x,8A/12x,A/)')    '### ERROR: Cannot rename ',
     &     'variable ''', vname(1:lvar), ''' into ''', newname(1:lnew),
     &               ''' inside netCDF file ''',  fname(1:lfmn), ''',',
     &                                               nf_strerror(ierr)
          endif
        endif

! Check, whether object to be renamed is a dimension and rename
! it as requested.

        fnd_dim=.false.
        ierr=nf_inq_dimid (ncid, vname(1:lvar), varid)
        if (ierr == nf_noerr) then
          fnd_dim=.true.
          ierr=nf_rename_dim (ncid, varid, newname(1:lnew))

          if (ierr == nf_noerr) then
            write(*,'(/8x,7A/)')               'Renamed dimension ''',
     &                   vname(1:lvar), ''' into ''', newname(1:lnew),
     &               ''' inside netCDF file ''', fname(1:lfmn), '''.'
          elseif (ierr. eq. nf_enameinuse) then
            write(*,'(/1x,8A/12x,A/)')    '### ERROR: Cannot rename ',
     &                   'dimension ''',  vname(1:lvar), ''' into ''',
     &                    newname(1:lnew), ''' inside netCDF file ''',
     &                                           fname(1:lfmn), ''',',
     &              'This name is already used by another dimension.'
          else
            write(*,'(/1x,8A/12x,A/)')    '### ERROR: Cannot rename ',
     &                   'dimension ''',  vname(1:lvar), ''' into ''',
     &                    newname(1:lnew), ''' inside netCDF file ''',
     &                        fname(1:lfmn), ''',', nf_strerror(ierr)
          endif
        endif

! Check, whether object to be renamed is a global attribute, and
! if so, rename it as requested.

        fnd_att=.false.
        ierr=nf_rename_att (ncid, nf_global, vname(1:lvar),
     &                                       newname(1:lnew))
        if (ierr == nf_noerr) then
          write(*,'(/8x,7A/)')          'Renamed global attribute ''',
     &                   vname(1:lvar), ''' into ''', newname(1:lnew),
     &                ''' inside netCDF file ''', fname(1:lfmn), '''.'
          fnd_att=.true.
        endif

! Scan among all the variables and check whether the name matches
! name of an attribute to a variable: if found, rename it rename it
! as requested for each individual variable.

        ierr=nf_inq_nvars (ncid, nvars)
        if (ierr == nf_noerr) then
          do varid=1,nvars
            ierr=nf_rename_att (ncid, varid, vname(1:lvar),
     &                                     newname(1:lnew))
            if (ierr == nf_noerr) then
              ierr=nf_inq_varname (ncid, varid, name)
              lstr=lenstr(name)
              write(*,'(/8x,9A/)')             'Renamed attribute ''',
     &                   vname(1:lvar), ''' into ''', newname(1:lnew),
     &                             ''' for variable ''', name(1:lstr),
     &               ''' inside netCDF file ''', fname(1:lfmn), '''.'
              fnd_att=.true.
            endif
          enddo
        else
          write(*,'(/8x,5A/)') 'Cannot determine number of variables ',
     &                       'in netCDF file ''', fname(1:lfmn), '''.'
        endif

        if (.not.fnd_var .and. .not.fnd_dim .and. .not.fnd_att)
     &    write(*,'(/8x,6A/)')  '### WARNING: Cannot find an object ',
     &         'named ''', vname(1:lvar),  ''' within netCDF file ''',
     &                                           fname(1:lfmn), '''.'

        ierr=nf_enddef (ncid)
        if (ierr/=nf_noerr) then
          write(*,'(/8x,4A/13x,A)')     '### ERROR: Cannot switch to ',
     &            'data mode for netCDF file ''', fname(1:lfmn), '''.',
     &                                               nf_strerror(ierr)
        endif
        ierr=nf_close (ncid)

      enddo !<-- iarg
      stop
      end
