      subroutine read_gshhs_data

! Read GSHHS coastline data from a file in its original native format.
! [This is an alternative to C program read_gshhs.c, with the exception
! that in addition to justst reading data it determines the necessary
! sizes and allocates all the arrays in the module above.]  GSHHS data
! can be obtained from 
!
!      http://www.ngdc.noaa.gov/mgg/shorelines/data/gshhs/latest/
!
! where it is found inside  gshhg-bin-VERSION.zip archive. The files
! named as gshhs_X.b  where X={f,i,h,c} depending on the resolution are
! essentially sequential binary, big-endian files containing integer
! values only.  Each coastline polygon corresponds to two records in
! the file: a fixed-size header with the description followed by a
! variable size record containing the actual points. Since the files
! were written by a C-program, they have no 4-byte offsets at the
! beginning and the end of the records (unlike Fortran sequential
! files), so to read this kind of file by a Fortran program it is
! decided to read the whole file at once (as it would be a direct
! access file containing just a single record), perform byte swap if 
! necessary to convert it into little endian, then sort out its
! content. 

! The following switch must be defined on all litte-endian computers,
! such as Intel/AMD; undefined on big-endian (IBM, etc...)

#define CONVERT_BIG_TO_LITTLE_ENDIAN

c--#define TEST_DECODE

      use mod_gshhs
      implicit none
      character(len=64) fname
      integer, allocatable, dimension(:) :: ibffr

      integer fsize,ierr, id, i,j,k,n, this_segm, lfnm, lenstr
#ifdef CONVERT_BIG_TO_LITTLE_ENDIAN
      integer(kind=1),  dimension(4) :: bytes_in, bytes_out
#endif
#ifdef CRUDE
      fname='/opt/gshhs/gshhs_i.b'
#else
      fname='/opt/gshhs/gshhs_f.b'
#endif

      fsize=0 ; lfnm=lenstr(fname)
      inquire(file=fname(1:lfnm), size=fsize)
      if (fsize > 0) then 
        write(*,'(/1x,3A,I12,1x,A)') 'size of ''', fname(1:lfnm),
     &                                   ''' is', fsize, 'Bytes'
        fsize=fsize/4 ; allocate(ibffr(fsize))

! Different Fortran compilers use different units (Bytes, 4-Byte
! words, etc) to specify record length for direct access files;

#if defined __IFC 
# define RECL fsize
#elif defined GCC
# define RECL 4*fsize 
#else
# error Cannot specify RECL in "read_gshhs_data.F"   
#endif
        open(unit=11, file=fname(1:lfnm), form='unformatted',
     &              access='direct', recl=RECL, status='old',
     &                                          iostat=ierr)
        if (ierr == 0) then
          read(unit=11, rec=1, iostat=ierr) ibffr
          close(unit=11)
          if (ierr == 0) then
            write(*,*) 'read_gshhs_data :: buffering ''',
     &                      fname(1:lfnm), ''' complete'
          else
            write(*,'(/1x,3A/)')    '### ERROR: read_gshhs_data :: ',
     &                  'Cannot read file ''', fname(1:lfnm),  '''.'
          endif
        else
          write(*,'(/1x,3A/)')      '### ERROR: read_gshhs_data :: ',
     &                   'Cannot open file ''', fname(1:lfnm), '''.' 
        endif
      else
        write(*,'(/1x,4A/)') '### ERROR: read_gshhs_data :: Cannot ',
     &             'determine size of file ''', fname(1:lfnm), '''.'
      endif
      if (fsize <= 0 .or. ierr /= 0) stop

#ifdef CONVERT_BIG_TO_LITTLE_ENDIAN
      do i=1,fsize
        bytes_in=transfer(ibffr(i), bytes_in)
        bytes_out(1)=bytes_in(4) ; bytes_out(2)=bytes_in(3)
        bytes_out(4)=bytes_in(1) ; bytes_out(3)=bytes_in(2)
        ibffr(i)=transfer(bytes_out,ibffr(i))
      enddo
#endif

      nsegm=0 ; npts_total=0 ; i=1
      do while(i < fsize)
        nsegm=nsegm+1 ; id=ibffr(i)
        n=ibffr(i+1)  ; npts_total=npts_total+n
c**     write(*,*) nsegm, id, n
        if (nsegm /= id+1) then
          write(*,'(1x,3A)') '### ERROR: read_gshhs_data :: Logical ',
     &                    'error in reading ''', fname(1:lfnm), '''.'
        endif
        i=i + (header_size+2*n)
      enddo
      write(*,*) 'read_gshhs_data :: found', nsegm, ' segments ,',
     &                               npts_total,' points total'

      if (i == fsize+1) then
        write(*,*) 'matching last index of bffr to file size', fsize
      elseif (i > fsize+1) then
        write(*,'(/1x,4A/12x,2A)')   '### ERROR: read_gshhs_data :: ',
     &        'Possible truncation of file ''',  fname(1:lfnm), ''':',
     &        'according to its header the last segment goes beyond ',
     &                                         'the end of the file.' 
         write(*,*) 'last segment end at', i-1, ' EOF at', fsize 
        stop 
      elseif (i < fsize+1) then
        write(*,'(/1x,4A,12x,2A/)')  '### ERROR: read_gshhs_data :: ',
     &       'Unrequested data found at the end of ''', fname(1:lfnm),
     &             ''':', 'according to its header the last segment ',
     &                              'terminates before the file end.'
        write(*,*) 'last segment end at', i-1, ' EOF at', fsize
        stop
      endif

      allocate(isegm(nsegm), segmhdr(header_size,nsegm),
     &                             segmpt(2,npts_total))
      i=1 ; k=0 ; this_segm=1
      do while(i < fsize)
        k=k+1 ;
        do j=1,header_size
          segmhdr(j,k)=ibffr(i+j-1)
        enddo
        n=segmhdr(2,k) ; isegm(k)=this_segm
        do j=0,n-1
          segmpt(1,j+this_segm)=ibffr(i+2*j    +header_size)
          segmpt(2,j+this_segm)=ibffr(i+2*j +1 +header_size)
        enddo
        this_segm=this_segm+n
        i=i + (header_size+2*n)
      enddo
      deallocate(ibffr)
      write(*,*) 'read_gshhs_data :: decoding complete' 
      end

! The following is a functionally equivalent substitute for
! C-function decode_flag_ in read_gshhs.c 

!       flag = level + version << 8 + greenwich << 16
!              + source << 24 + river << 25 + p << 26

! so it contains 6 items, as follows:

!    level 1 land, 2 lake, 3 island_in_lake, 4 pond_in_island_in_lake
!  version should be 9 for GSHHG release 9
! greenwich 0 if neither Greenwich nor Dateline are crossed;
!           1 if Greenwich is crossed; 2 ifDateline is crossed; 
!           3 if both are crossed.
!    source 0 CIA WDBII, 1 WVS
!     river 0 not set; 1 river-lake and GSHHG level=2 (or WDBII class 0)
! area_scale area magnitude scale p (as in 10^p) We divide area by 10^p


      subroutine dec_segm_flag(flag, level, version, greenwich,
     &                               source, river, area_scale)
      implicit none
      integer(kind=4) flag, level, version, greenwich, source,
     &                               river, area_scale,  ich
      character(len=1) bytes(4)

      bytes=transfer(flag, bytes)   ! Note: character rather than
                                    ! integer(kind=1) is used here
      level=ichar(bytes(1))         ! to achieve behavior of
      version=ichar(bytes(2))       ! unsigned integer;
      ich=ichar(bytes(3))           ! ichar(char) yields values
      greenwich=mod(ich,4)          ! from 0 to 255, while             
                                    ! transfer(flag, int_kind1(4))
      ich=ichar(bytes(4))           ! yields both positive and 
      source=mod(ich,2)             ! negative numbers appearing 
      ich=ich/2                     ! in int_kind1.
      river=mod(ich,2)
      area_scale=ich/2 
      end

#ifdef TEST_DECODE
      program test_decode_flag
      implicit none
      integer(kind=4) flag, level, version, greenwich,
     &                      source, river, area_scale
 1    read(*,*) flag
        call ec_segm_flag(flag, level, version, greenwich,
     &                          source, river, area_scale)

        write(*,'(1x,A,I10,6(1x,A,I3))')  'flag =', flag,
     &            'level =', level, 'version =', version,
     &      'greenwich =', greenwich, 'source =', source,
     &      'river =', river, 'area_scale =', area_scale
      goto 1
      end
#endif

