      program rfact

! A basic command-line tool to assess smoothness of bottom topography
! and grid stiffness ratios (Haney, 1991 numbers) of pending vertical
! S-coordinate transformation.  The usage is as follows: rfact takes
! five arguments:
!
!           rfact grid_file.nc  N  theta_s  theta_b  hc
!
! which are: (1) grid netCDF file containing bottom topography, and,
! optionally, land mask; (2) N is pending vertical grid resolution;
! (3-5) the remaining three arguments, theta_s, theta_b, hc,  are
! vertical stretching parameters.  If successfully executed, this
! program prints two numbers, "rx0" and "rx1", which exactly match
! those printed by the actual ROMS code. This is a purely diagnostic
! tool and it makes no change in grid netCDF file.

      implicit none
      real(kind=8) hc, theta_s, theta_b
      real(kind=8), dimension(:), allocatable :: Cs_w,Cs_r,
     &                                      h,rmask,z_r,z_w
      integer nargs, Lm,Mm,N, ncgrd, i, max_horiz_pts, ierr,
     &           lgrd,  lenstr, read_str_int, read_str_dble
      character grid_file*64, string*24
#include "netcdf.inc"

      nargs=iargc()
      if (nargs==5) then
        call getarg(1,grid_file)
        lgrd=lenstr(grid_file)
        write(*,*) 'grid_file = ''', grid_file(1:lgrd), ''''
        ierr=nf_open(grid_file, nf_nowrite, ncgrd)
        if (ierr == nf_noerr) then
          call getarg (2,string)
          ierr=read_str_int(string, N)
          if (ierr == 0) then
            write(*,'(9x,A,I3)') 'N =', N
            call getarg (3,string)
            ierr=read_str_dble(string, theta_s)
            if (ierr == 0) then
              write(*,'(3x,A,F8.5)')  'theta_s =', theta_s
              call getarg (4,string)
              ierr=read_str_dble(string, theta_b)
              if (ierr == 0) then
                write(*,'(3x,A,F8.5)')  'theta_b =', theta_b
                call getarg (5,string)
                ierr=read_str_dble(string, hc)
                if (ierr == 0) then
                  write(*,'(8x,A,F8.3/)')  'hc =', hc
                  call roms_find_dims (ncgrd, grid_file, Lm,Mm, i)

                  allocate (Cs_w(0:N), Cs_r(N))

                  max_horiz_pts=(Lm+2)*(Mm+2)
                  allocate (h(max_horiz_pts))
                  allocate (rmask(max_horiz_pts))
                  allocate (z_r(max_horiz_pts*N))
                  allocate (z_w(max_horiz_pts*(N+1)))

                  call get_var_by_name_double(ncgrd, 'h', h)

                  ierr=nf_inq_varid(ncgrd, 'mask_rho', i)
                  if (ierr==nf_noerr) then
                    ierr=nf_get_var_double(ncgrd, i, rmask)
                  else
                    write(*,'(/1x,4A/14x,3A/)')    '### WARNING: ',
     &                   'Cannot find land mask from grid file ''',
     &                    grid_file(1:lgrd),   '''.',  'Assuming ',
     &                   'rmask=1 everywhere: ', nf_strerror(ierr)
                    ierr=nf_noerr
                    do i=1,max_horiz_pts
                      rmask(i)=1.D0
                    enddo
                  endif
                  ierr=nf_close(ncgrd)

                  call set_scoord (theta_s,theta_b, N, Cs_r,Cs_w)
                  call set_depth (Lm,Mm,N, Cs_r, Cs_w, hc,'SM09',
     &                                               h, z_r, z_w)
                  call grid_stiffness (Lm,Mm, N, rmask,z_w)

                else
                  write(*,'(/1x,A/)') '### ERROR in arg. #5: hc=?'
                endif
              else
                write(*,'(/1x,A/)') '### ERROR in arg. #4: theta_b=?'
              endif
            else
              write(*,'(/1x,A/)') '### ERROR in arg. #3: theta_s=?'
            endif
          else
            write(*,'(/1x,A/)') '### ERROR in arg. #2: N=?'
          endif
        else
          write(*,'(/1x,4A/12x,A/)')  '### ERROR: Cannot open ROMS ',
     &    'grid file ''', grid_file(1:lgrd), ''' in read-only mode.',
     &                                            nf_strerror(ierr)
        endif
      else
        write(*,'(1x,A/ /12x,A/)') 'Usage of  rfact should be:',
     &            'rfact grid_file.nc  N  theta_s  theta_b  hc'
      endif
      end



      subroutine grid_stiffness(Lm,Mm, N, rmask,z_w)

!                   z(i,k)                  Survey three-dimensional
!                     /|                    grid in order to determine
!                    / |                    the maximum grid stiffness
!                   /  |                    ratio,
!                  /   |                                  slope*dx
!                 /    +  ----------               r_x = ----------
!                /     |          !                         dZ
!  -------------/      |          !
!   !          /       |          !         This is done for purely
!   !         /        |          !         diagnostic purposes and
!   !        /     z(i,k-1)       !         does not affect the
!   !       /         /           !         outcome of model
!  dZ      /         /         slope*dx     computations.
!   !     /         /             !
!   !  z(i-1,k)    /              !
!   !    |        /               !
!   !    |       /                !
!  ----- | - - -/                 !
!        |     /                  !
!        +- - / - ------------------
!        |   /
!        |  /
!        | /                    z(i,k)-z(i-1,k)+z(i,k-1)-z(i-1,k-1)
!        |/              r_x = -------------------------------------
!    z(i-1,k-1)                 z(i,k)+z(i-1,k)-z(i,k-1)-z(i-1,k-1)

      implicit none
      integer Lm,Mm, N, istr,iend,jstr,jend, istrU,jstrV, i,j,k
      real(kind=8) rmask(0:Lm+1,0:Mm+1), z_w(0:Lm+1,0:Mm+1,0:N),
     &                                           my_rx0, my_rx1

      istr=1 ; iend=Lm ; istrU=istr+1 ; my_rx0=0.D0
      jstr=1 ; jend=Mm ; jstrV=jstr+1 ; my_rx1=0.D0

      do j=jstr,jend
        do i=istrU,iend
          if (rmask(i,j) > 0.5 .and. rmask(i-1,j) > 0.5) then
            my_rx0=max(my_rx0, abs( (z_w(i,j,0)-z_w(i-1,j,0))
     &                             /(z_w(i,j,0)+z_w(i-1,j,0))  ))
            do k=1,N
              my_rx1=max(my_rx1, abs(
     &         (z_w(i,j,k)-z_w(i-1,j,k)+z_w(i,j,k-1)-z_w(i-1,j,k-1))
     &        /(z_w(i,j,k)+z_w(i-1,j,k)-z_w(i,j,k-1)-z_w(i-1,j,k-1))
     &                                                            ))
            enddo

          endif
        enddo
        if (j >= jstrV) then
          do i=istr,iend
            if (rmask(i,j)>0.5 .and. rmask(i,j-1)>0.5) then
              my_rx0=max(my_rx0, abs( (z_w(i,j,0)-z_w(i,j-1,0))
     &                                /(z_w(i,j,0)+z_w(i,j-1,0)) ))
              do k=1,N
                my_rx1=max(my_rx1, abs(
     &         (z_w(i,j,k)-z_w(i,j-1,k)+z_w(i,j,k-1)-z_w(i,j-1,k-1))
     &        /(z_w(i,j,k)+z_w(i,j-1,k)-z_w(i,j,k-1)-z_w(i,j-1,k-1))
     &                                                            ))
              enddo
            endif
          enddo
        endif
      enddo
      write(*,'(/1x,2A,F16.10/38x,A,F16.10/)') 'Maximum topography ',
     &   'stiffness ratios: rx0 =',my_rx0, 'rx1 =',my_rx1
      end
