! Content:  Rotation of 2D vector components with simultaneous
!--------   application of land mask. Parallelization by tiling.
!
! rotate_in_place         ! both vector components are co-located
! rotate_in_place_thread  ! at RHO-points; the rotated vectors are 
! rotate_in_place_tile    ! placed into the same arrays as input;
!
! rotate_cgrd             ! input at RHO-points; output at 
! rotate_cgrd_thread      ! U,V-points of C-grid; output goes
! rotate_cgrd_tile        ! into new pair of arrays;



      module rotate_priv_vars
       integer, save :: alloc_rtvr_size=0
C$OMP THREADPRIVATE(alloc_rtvr_size)
       real(kind=8), allocatable, dimension(:) :: ur1,vr1
C$OMP THREADPRIVATE(ur1,vr1)
      end module rotate_priv_vars


! Rotate co-located vector components ur,v defined at RHO-points:


      subroutine rotate_in_place(Lm,Mm, csA,snA, msk, ur,vr)
      implicit none
      integer Lm,Mm
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4), dimension(0:Lm+1,0:Mm+1) :: ur,vr
C$OMP PARALLEL SHARED(Lm,Mm, csA,snA, msk, ur,vr)
      call rotate_in_place_thread(Lm,Mm, csA,snA, msk, ur,vr) 
C$OMP END PARALLEL
      end

      subroutine rotate_in_place_thread(Lm,Mm, csA,snA, msk, ur,vr)
      implicit none
      integer Lm,Mm
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4), dimension(0:Lm+1,0:Mm+1) :: ur,vr

      integer numthreads,trd, nsub_x,nsub_y, size,tile,
     &           my_first,my_last, istr,iend,jstr,jend
C$    integer omp_get_num_threads, omp_get_thread_num
      numthreads=1 ; trd=0
C$    numthreads=omp_get_num_threads() ; trd=omp_get_thread_num()
      call set_tiles(Lm,Mm, nsub_x,nsub_y)
      size=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*size
      my_last=min(my_first+size-1, nsub_x*nsub_y-1)

      do tile=my_last,my_first,-1
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)

        call rotate_in_place_tile(istr,iend,jstr,jend,
     &                        Lm,Mm, csA,snA, msk, ur,vr)
      enddo
      end

      subroutine rotate_in_place_tile(istr,iend,jstr,jend,
     &                        Lm,Mm, csA,snA, msk, ur,vr)
      implicit none
      integer istr,iend,jstr,jend, Lm,Mm, i,j
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4), dimension(0:Lm+1,0:Mm+1) :: ur,vr
      real(kind=4) u1,v1
#include "compute_extended_bounds.h"
      do j=jstrR,jendR
        do i=istrR,iendR
          if (msk(i,j) == 1) then
            u1=ur(i,j) ; v1=vr(i,j)
            ur(i,j)=u1*csA(i,j)+v1*snA(i,j)
            vr(i,j)=v1*csA(i,j)-u1*snA(i,j)
          else
            ur(i,j)=0.
            vr(i,j)=0.
          endif
        enddo
      enddo
      end


! C-grid component placement: ur,vr are non-rotated co-located at
! RHO-points components ofvector field; uc,vc are C-grid-staggered;
! intermediate ur1,vr1 (private arrays used internally) are rotated
! but still co-located.

      subroutine rotate_cgrd(Lm,Mm, csA,snA, msk, ur,vr,uc,vc)
      implicit none
      integer Lm,Mm
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4) ur(0:Lm+1,0:Mm+1), uc(1:Lm+1,0:Mm+1),
     &             vr(0:Lm+1,0:Mm+1), vc(0:Lm+1,1:Mm+1)
C$OMP PARALLEL SHARED(Lm,Mm, csA,snA, msk, ur,vr,uc,vc)
      call rotate_cgrd_thread(Lm,Mm, csA,snA, msk, ur,vr,uc,vc)
C$OMP END PARALLEL
      end

      subroutine rotate_cgrd_thread(Lm,Mm, csA,snA, msk, ur,vr,uc,vc)
      use rotate_priv_vars
      implicit none
      integer Lm,Mm
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4) ur(0:Lm+1,0:Mm+1), uc(1:Lm+1,0:Mm+1),
     &             vr(0:Lm+1,0:Mm+1), vc(0:Lm+1,1:Mm+1)

      integer numthreads,trd, nsub_x,nsub_y, size,tile,
     &           my_first,my_last, istr,iend,jstr,jend
C$    integer omp_get_num_threads, omp_get_thread_num
      numthreads=1 ; trd=0
C$    numthreads=omp_get_num_threads() ; trd=omp_get_thread_num()

      call set_tiles(Lm,Mm, nsub_x,nsub_y)

      size=((Lm+nsub_x-1)/nsub_x +2)*((Mm+nsub_y-1)/nsub_y +2)
      if (size > alloc_rtvr_size) then
        if (allocated(ur1)) deallocate(ur1,vr1)
        allocate(ur1(size), vr1(size)) ; alloc_rtvr_size=size
C$OMP CRITICAL(rt_cgrd_crgn) 
        write(*,'(8x,2A,I3)') 'rotate_cgrd_thread :: allocated ',
     &                           'private scratch, trd =', trd
C$OMP END CRITICAL(rt_cgrd_crgn)
      endif

      size=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*size
      my_last=min(my_first+size-1, nsub_x*nsub_y-1)

      do tile=my_last,my_first,-1
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)

        call rotate_cgrd_tile(istr,iend,jstr,jend, Lm,Mm,
     &               csA,snA, msk, ur,vr, uc,vc, ur1,vr1)
      enddo
      end


      subroutine rotate_cgrd_tile(istr,iend,jstr,jend, Lm,Mm,
     &                  csA,snA, msk, ur,vr, uc,vc, ur1,vr1)
      implicit none
      integer istr,iend,jstr,jend, Lm,Mm, i,j
      integer(kind=2), dimension(0:Lm+1,0:Mm+1) :: msk
      real(kind=8), dimension(0:Lm+1,0:Mm+1) :: csA,snA
      real(kind=4) ur(0:Lm+1,0:Mm+1), uc(1:Lm+1,0:Mm+1),
     &             vr(0:Lm+1,0:Mm+1), vc(0:Lm+1,1:Mm+1)
      real(kind=8) ur1(istr-1:iend+1,jstr-1:jend+1),
     &             vr1(istr-1:iend+1,jstr-1:jend+1)
#include "compute_extended_bounds.h"
      do j=jstr-1,jendR
        do i=istr-1,iendR
          ur1(i,j)=ur(i,j)*csA(i,j)+vr(i,j)*snA(i,j)
          vr1(i,j)=vr(i,j)*csA(i,j)-ur(i,j)*snA(i,j)
        enddo
      enddo
      do j=jstrR,jendR
        do i=istr,iendR
          if (msk(i,j) == 1 .and. msk(i-1,j) == 1) then
            uc(i,j)=0.5D0*(ur1(i,j)+ur1(i-1,j))
          else
            uc(i,j)=0.
          endif
        enddo
      enddo
      do j=jstr,jendR
        do i=istrR,iendR
          if (msk(i,j) == 1 .and. msk(i,j-1) == 1) then
            vc(i,j)=0.5D0*(vr1(i,j)+vr1(i,j-1))
          else
            vc(i,j)=0.
          endif
        enddo
      enddo
      end
