#ifndef MASK_NARROW_BAYS
      program single_connect
#else
      program mask_narrow_bays
#endif

! Purpose: to enforce single-domain connectivity of the unmasked area.
!--------- -- ------- ------------- ------------ -- --- -------- -----
! Connectivity means that is each unmasked RHO-point can be reached
! from any other unmasked RHO-point by going through unmasked U- and
! V-points.

! The difference between "check_connect" and "single_connect" is that
! the latter also closes narrow bays defined as unmasked points on the
! coastline which are connected to the main water body from one side
! only (i.e., the point is at the end a dead-ended fiord) and only by
! a single-point-wide passage.

! Either program takes a user-specified initial point, indices (i0,j0),
! which is assumed to belong to the main body of water and, starting
! from this point starts filling the adjacent non-masked points as
! "water" (so for each non-masked point on the grid it looks at all
! the immediate neighboring points and if at least one of them is
! already labelled as "water", then the point itself becomes "water").
! Note that at any time during the process there are three types of
! points: 1-masked ("land" according to the initial landmask mask),
! which remain unchanged; 2-unmasked, but non-labelled as "water",
! (initially all unmasked points); and 3-unmasked labelled as "water"
! (initially there is only one such point, i0,j0). The labelling
! continues until the number of "water" points no longer grows, which
! is used as termination signal.  At the end the new landmask is
! defined as the points which are not "water" (i.e., either initially
! masked points, or unmasked, but not labelled as "water").

#define TIMING

      implicit none
      character(len=80) grid_file
      real(kind=8), allocatable, dimension(:,:) :: rmask
      integer(kind=2), allocatable, dimension(:,:) :: imask, ms1,ms2
      integer nargs, iargc, ncgrd, Lm,Mm, error_count, i, ierr,
     &                                                  lgrd, lenstr
#ifndef MASK_NARROW_BAYS
      integer i0,j0, read_str_int
      character(len=16) str
#endif
#include "netcdf.inc"

      error_count=0
      nargs=iargc()

#ifndef MASK_NARROW_BAYS
      if (nargs == 3) then
        call getarg(1,str) ; ierr=read_str_int(str, i0)
        if (ierr /= 0) then
          error_count=error_count+1
          write(*,'(/2(1x,A)/)') '### ERROR: Wrong argument #1',
     &                                      str(1:lenstr(str))
        endif
        call getarg(2,str) ; ierr=read_str_int(str, j0)
        if (ierr /= 0) then
          error_count=error_count+1
          write(*,'(/2(1x,A)/)') '### ERROR: Wrong argument #2',
     &                                      str(1:lenstr(str))
        endif
#else
      if (nargs == 1) then
#endif
        call getarg(nargs,grid_file) ; lgrd=lenstr(grid_file)
      else
        write(*,'(/2(1x,A)/,/8x,A/1x,A/8x,A/,/2(1x,A)/1x,A)')
     &          'The usage of "single_connect" and related',
     &          '"mask_narrow_bays" should be:',
     &          'single_connect i0 j0 roms_grid_file.nc',
     &          'and', 'mask_narrow_bays roms_grid_file.nc',
     &          'where i0 j0 are user-specified indices of',
     &          'a water point within the selected',
     &          'body of water.'


       error_count=1
      endif
      if (error_count > 0) stop

      ierr=nf_open (grid_file, nf_write, ncgrd)
      if (ierr == nf_noerr) then
        call roms_find_dims(ncgrd, grid_file, Lm,Mm, i)

        allocate(rmask(0:Lm+1,0:Mm+1))
        call get_var_by_name_double(ncgrd, 'mask_rho', rmask)

#ifndef MASK_NARROW_BAYS
        if (rmask(i0,j0) > 0.5) then
#endif
          allocate(imask(0:Lm+1, 0:Mm+1), ms1(-1:Lm+2,-1:Mm+2),
     &                                    ms2(-1:Lm+2,-1:Mm+2))

#ifndef MASK_NARROW_BAYS
C$OMP PARALLEL SHARED(        Lm,Mm, i0,j0, rmask,imask, ms1,ms2)
          call sin_con_thread(Lm,Mm, i0,j0, rmask,imask, ms1,ms2)
#else
C$OMP PARALLEL SHARED(           Lm,Mm, rmask,imask, ms1,ms2)
          call narrow_bay_thread(Lm,Mm, rmask,imask, ms1,ms2)
#endif
C$OMP END PARALLEL

          call put_var_by_name_double(ncgrd, 'mask_rho', rmask)
          ierr=nf_close (ncgrd)

#ifndef MASK_NARROW_BAYS
        else
          write(*,'(/1x,2A,i4,2x,A,i4,1x,A/)')  '### ERROR: selected ',
     &    'point i =', i0, 'j =', j0, 'is on land. Try another point.'
        endif
#endif
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot open ROMS grid ',
     &   'netCDF file ''', grid_file(1:lgrd), ''' in read-write mode.',
     &                                               nf_strerror(ierr)
      endif
      end



#define WESTERN_EDGE istr==1
#define EASTERN_EDGE iend==Lm
#define SOUTHERN_EDGE jstr==1
#define NORTHERN_EDGE jend==Mm


#ifndef MASK_NARROW_BAYS
      subroutine sin_con_thread(Lm,Mm, i0,j0, rmask,imask, ms1,ms2)
#else
      subroutine narrow_bay_thread(Lm,Mm, rmask,imask, ms1,ms2)
#endif
      use single_conn_comm_vars
      implicit none
      integer Lm,Mm
      real(kind=8) rmask(0:Lm+1,0:Mm+1)
      integer(kind=2) imask(0:Lm+1,0:Mm+1), ms1(-1:Lm+2,-1:Mm+2),
     &                                      ms2(-1:Lm+2,-1:Mm+2)
      integer numthreads,trd, nsub_x,nsub_y, my_first,my_last,
     &        range, tile, istr,iend,jstr,jend, my_sum, iter, i,j
#ifndef MASK_NARROW_BAYS
     &                                                    , i0,j0
#else
     &                                                       , m
#endif
C$    integer omp_get_num_threads, omp_get_thread_num
#ifdef TIMING
      integer(kind=4) iclk_start, iclk_end, clk_rate, clk_max
      call system_clock(iclk_start, clk_rate, clk_max)
#endif
      numthreads=1                     ; trd=0
C$    numthreads=omp_get_num_threads() ; trd=omp_get_thread_num()

      call set_tiles(Lm,Mm, nsub_x,nsub_y)

c**   nsub_x=1 ; nsub_y=1 !<-- for testing parallel correctness only

      range=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*range
      my_last=min(my_first + range-1, nsub_x*nsub_y-1)

      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                 istr,iend,jstr,jend)

        if (WESTERN_EDGE) istr=istr-1 ; if (SOUTHERN_EDGE) jstr=jstr-1
        if (EASTERN_EDGE) iend=iend+1 ; if (NORTHERN_EDGE) jend=jend+1

        do j=jstr,jend                         ! Initialize integer
          do i=istr,iend                       ! mask: it is possibile
            if (rmask(i,j) > 0.5D0) then       ! that "rmask" read from
              imask(i,j)=1                     ! netCDF file contain
            elseif (rmask(i,j) < -1.5D0) then  ! negative values which
              imask(i,j)=-2                    ! label points masked
            elseif (rmask(i,j) < -0.5D0) then  ! out by single_connect
              imask(i,j)=-1                    ! or close_narrow_bays
            else                               ! applied previously:
              imask(i,j)=0                     ! these special values
            endif                              ! are inherited by
          enddo                                ! "imask" and passed
        enddo                                  ! back to "rmask".

        if (WESTERN_EDGE) istr=istr-1 ; if (SOUTHERN_EDGE) jstr=jstr-1
        if (EASTERN_EDGE) iend=iend+1 ; if (NORTHERN_EDGE) jend=jend+1

        do j=jstr,jend
          do i=istr,iend
#ifndef MASK_NARROW_BAYS
            ms1(i,j)=0 ; ms2(i,j)=0
#else
            ms1(i,j)=imask(i,j) ; ms2(i,j)=imask(i,j)
#endif
          enddo
        enddo
#ifndef MASK_NARROW_BAYS
        if (istr <= i0 .and. i0 <= iend .and.
     &      jstr <= j0 .and. j0 <= jend) then
          ms1(i0,j0)=1
        endif
#endif
      enddo   !<-- tile
C$OMP BARRIER


#ifndef MASK_NARROW_BAYS
C$OMP MASTER
      write(*,'(/1x,2A,2(I4,1x,A))') 'Enforce that all water points ',
     &            'can be reached from point  i =',i0, 'j =',j0,'...'
      trd_count=0 ; wtr_pts=0 ; incr_wtr_pts=+1  !<-- initialize
C$OMP END MASTER

      iter=0                             ! The following while() loop
      do while(incr_wtr_pts /= 0)        ! consists of two very similar
        iter=iter+1                      ! segments which differ only by
        my_sum=0                         ! switching "ms1" and "ms2".
        do tile=my_first,my_last,+1
          call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                istr,iend,jstr,jend)

          if (WESTERN_EDGE) istr=istr-1; if (SOUTHERN_EDGE) jstr=jstr-1
          if (EASTERN_EDGE) iend=iend+1; if (NORTHERN_EDGE) jend=jend+1

          do j=jstr,jend
            do i=istr,iend
              if ( imask(i,j) > 0 .and. ( ms1(i,j ) > 0  .or.
     &             ms1(i+1,j) > 0  .or.   ms1(i,j+1) > 0   .or.
     &             ms1(i-1,j) > 0  .or.   ms1(i,j-1) > 0 )) then
                my_sum=my_sum+1
                ms2(i,j)=1
              endif
            enddo
          enddo
        enddo  !<-- tile
C$OMP CRITICAL(cr_regn)
        if (trd_count == 0) then
          wtr_pts_bak=wtr_pts ; wtr_pts=0
        endif
        trd_count=trd_count+1
        wtr_pts=wtr_pts+my_sum
        if (trd_count == numthreads) then
          trd_count=0
          if (mod(iter,50) == 0 .or. wtr_pts == wtr_pts_bak) then
            write(*,'(8x,A,I7,2(2x,A,I10))') 'iter =', iter,
     &     'wtr_pts =',wtr_pts, 'changes =',wtr_pts-wtr_pts_bak
          endif
        endif
C$OMP END CRITICAL(cr_regn)
C$OMP BARRIER

        iter=iter+1
        my_sum=0
        do tile=my_last,my_first,-1
          call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                istr,iend,jstr,jend)
          if (WESTERN_EDGE) istr=istr-1; if (SOUTHERN_EDGE) jstr=jstr-1
          if (EASTERN_EDGE) iend=iend+1; if (NORTHERN_EDGE) jend=jend+1

          do j=jstr,jend
            do i=istr,iend
              if ( imask(i,j) > 0 .and. ( ms2(i,j ) > 0  .or.
     &             ms2(i+1,j) > 0  .or.   ms2(i,j+1) > 0    .or.
     &             ms2(i-1,j) > 0  .or.   ms2(i,j-1) > 0 )) then
                my_sum=my_sum+1
                ms1(i,j)=1
              endif
            enddo
          enddo
        enddo  !<-- tile
C$OMP CRITICAL(cr_regn)
        if (trd_count == 0) then
          wtr_pts_bak=wtr_pts ; wtr_pts=0
        endif
        trd_count=trd_count+1
        wtr_pts=wtr_pts+my_sum
        if (trd_count == numthreads) then
          trd_count=0 ; incr_wtr_pts=wtr_pts-wtr_pts_bak
          if (mod(iter,50) == 0 .or. wtr_pts == wtr_pts_bak) then
            write(*,'(8x,A,I7,2(2x,A,I10))') 'iter =', iter,
     &     'wtr_pts =',wtr_pts, 'changes =', incr_wtr_pts
          endif
        endif
C$OMP END CRITICAL(cr_regn)
C$OMP BARRIER
      enddo !<-- while()

      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                              istr,iend,jstr,jend)

        if (WESTERN_EDGE) istr=istr-1; if (SOUTHERN_EDGE) jstr=jstr-1
        if (EASTERN_EDGE) iend=iend+1; if (NORTHERN_EDGE) jend=jend+1

        do j=jstr,jend
          do i=istr,iend
            if (ms1(i,j) == 0 .and. imask(i,j) > 0) imask(i,j)=-1
          enddo
        enddo
      enddo
C$OMP BARRIER

#elif defined MASK_NARROW_BAYS

! The following part is to identify and close unresolved features
! of land mask, such as single-point bays.  For each water point the
! immediately adjacent surroundings is inspected and a weighting value
! is calculated by adding 8 for each water point adjacent on the east,
! west, north, and south sides, while adding 1 for each water point
! adjacent in diagonal direction.  If the resultant value falls below
! a pre-set threshold, the point is set to land.   The procedure is
! repeated iteratively until it is detected that further iterations
! do not result in any progress. In the three examples below "."
! means water, "x" land.
!
!       . . x x x x       . . x x x x x       . . x x x x x x x .
!       . . . . x x       . . . x x x x       . . . x x x x . . .
!       . . . . x x       . . . . o o x       . . . . . . . . . .
!       . . . x x x       . . x x x x x       . . x x x x . . . .
!       . . x x x x       . . x x x x x       . . x x x x x x . .
!
!   all water and         initially water       narrow passage: all
!   land points will      points "o" will       points will be kept
!   be kept as is         be turned to land     as they are

C$OMP MASTER
      write(*,'(/1x,A)') 'Closing unresolved narrow bays ...'
      trd_count=0 ; wtr_pts=0 ; incr_wtr_pts=+1 !<-- initialize
C$OMP END MASTER
C$OMP BARRIER

      iter=0
      do while(incr_wtr_pts /= 0)
        iter=iter+1

        do tile=my_first,my_last
          call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                 istr,iend,jstr,jend)
          do j=jstr,jend
            do i=istr,iend
              m=0
              if (imask(i+1,j  ) > 0) m=m+8
              if (imask(i+1,j+1) > 0) m=m+1
              if (imask(i  ,j+1) > 0) m=m+8
              if (imask(i-1,j+1) > 0) m=m+1
              if (imask(i-1,j  ) > 0) m=m+8
              if (imask(i-1,j-1) > 0) m=m+1
              if (imask(i  ,j-1) > 0) m=m+8
              if (imask(i+1,j-1) > 0) m=m+1
              ms1(i,j)=m
            enddo
          enddo

          if (WESTERN_EDGE) then
            do j=jstr,jend
              m=0
              if (imask(istr-1,j-1) > 0) m=m+8
              if (imask(istr  ,j-1) > 0) m=m+1
              if (imask(istr  ,j  ) > 0) m=m+8
              if (imask(istr  ,j+1) > 0) m=m+1
              if (imask(istr-1,j+1) > 0) m=m+8
              ms1(istr-1,j)=m
            enddo
          endif

          if (EASTERN_EDGE) then
            do j=jstr,jend
              m=0
              if (imask(iend+1,j+1) > 0) m=m+8
              if (imask(iend  ,j+1) > 0) m=m+1
              if (imask(iend  ,j  ) > 0) m=m+8
              if (imask(iend  ,j-1) > 0) m=m+1
              if (imask(iend+1,j-1) > 0) m=m+8
              ms1(iend+1,j)=m
            enddo
          endif

          if (SOUTHERN_EDGE) then
            do i=istr,iend
              m=0
              if (imask(i+1,jstr-1) > 0) m=m+8
              if (imask(i+1,jstr  ) > 0) m=m+1
              if (imask(i  ,jstr  ) > 0) m=m+8
              if (imask(i-1,jstr  ) > 0) m=m+1
              if (imask(i-1,jstr-1) > 0) m=m+8
              ms1(i,jstr-1)=m
            enddo
          endif

          if (NORTHERN_EDGE) then
            do i=istr,iend
              m=0
              if (imask(i-1,jend+1) > 0) m=m+8
              if (imask(i-1,jend  ) > 0) m=m+1
              if (imask(i  ,jend  ) > 0) m=m+8
              if (imask(i+1,jend  ) > 0) m=m+1
              if (imask(i+1,jend+1) > 0) m=m+8
              ms1(i,jend+1)=m
            enddo
          endif

          if (WESTERN_EDGE .and. SOUTHERN_EDGE) then
            m=0
            if (imask(istr  ,jstr-1) > 0) m=m+8
            if (imask(istr  ,jstr  ) > 0) m=m+1
            if (imask(istr-1,jstr  ) > 0) m=m+8
            ms1(istr-1,jstr-1)=m
          endif

          if (WESTERN_EDGE .and. NORTHERN_EDGE) then
            m=0
            if (imask(istr-1,jend  ) > 0) m=m+8
            if (imask(istr  ,jend  ) > 0) m=m+1
            if (imask(istr  ,jend+1) > 0) m=m+8
            ms1(istr-1,jend+1)=m
          endif

          if (EASTERN_EDGE.and. SOUTHERN_EDGE) then
            m=0
            if (imask(iend+1,jstr  ) > 0) m=m+8
            if (imask(iend  ,jstr  ) > 0) m=m+1
            if (imask(iend  ,jstr-1) > 0) m=m+8
            ms1(iend+1,jstr-1)=m
          endif

          if (EASTERN_EDGE .and. NORTHERN_EDGE) then
            m=0
            if (imask(iend  ,jend+1) > 0) m=m+8
            if (imask(iend  ,jend  ) > 0) m=m+1
            if (imask(iend+1,jend  ) > 0) m=m+8
            ms1(iend+1,jend+1)=m
          endif
        enddo  !<-- tile
C$OMP BARRIER

        my_sum=0
        do tile=my_last,my_first,-1
          call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                                istr,iend,jstr,jend)

          if (WESTERN_EDGE) istr=istr-1; if (SOUTHERN_EDGE) jstr=jstr-1
          if (EASTERN_EDGE) iend=iend+1; if (NORTHERN_EDGE) jend=jend+1

          do j=jstr,jend
            do i=istr,iend
              if (ms1(i,j) < 12 .and. imask(i,j) > 0) imask(i,j)=-2
              if (imask(i,j) > 0) my_sum=my_sum+1
            enddo
          enddo
        enddo
C$OMP CRITICAL(cr_regn)
        if (trd_count == 0) then
          wtr_pts_bak=wtr_pts ; wtr_pts=0
        endif
        trd_count=trd_count+1 ; wtr_pts=wtr_pts+my_sum
        if (trd_count == numthreads) then
          trd_count=0  ;  incr_wtr_pts=wtr_pts-wtr_pts_bak
          if (mod(iter,1) == 0 .or. incr_wtr_pts == 0) then
            write(*,*) 'iter =', iter, '  wtr_pts =', wtr_pts,
     &                            '  changes =', incr_wtr_pts
          endif
        endif
C$OMP END CRITICAL(cr_regn)
C$OMP BARRIER
      enddo !<-- while()
#endif  /* MASK_NARROW_BAYS */


      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)

        if (WESTERN_EDGE) istr=istr-1; if (SOUTHERN_EDGE) jstr=jstr-1
        if (EASTERN_EDGE) iend=iend+1; if (NORTHERN_EDGE) jend=jend+1

        do j=jstr,jend
          do i=istr,iend
            if (imask(i,j) > 0) then
              rmask(i,j)=1.D0              !<-- water stays water
            elseif (imask(i,j) == -1) then
              rmask(i,j)=-1.D0             !<-- water turned into land
            elseif (imask(i,j) == -2) then
              rmask(i,j)=-2.D0             !<-- narrow bays closed
            else
              rmask(i,j)=0.D0
            endif
          enddo
        enddo
      enddo
C$OMP BARRIER

#ifdef TIMING
      call system_clock(iclk_end, clk_rate, clk_max)
      if (clk_rate > 0) then
C$OMP CRITICAL(cr_regn)
        trd_count=trd_count+1
        if (trd_count == numthreads) then
          trd_count=0
          write(*,'(/ /1x,A,F8.2,1x,A,I4,1x,A/10x,A/)')
     &             'Wall clock time spent in computational part',
     &                      (iclk_end-iclk_start)/dble(clk_rate),
     &                    'sec  running', numthreads, 'threads.',
     &                    '[reading and writing files excluded]'
        endif
C$OMP END CRITICAL(cr_regn)
      endif
#endif
      end
