! Purpose: to smooth or "log-smooth" or rx-condition model topography
!--------- in order to achieve acceptable r-factors,
!
!                        | h(i+1) - h(i) |
!                  rx = -------------------  <  rx_max
!                          h(i+1) + h(i)
!
! It reads raw topography "hraw" from netCDF grid file calls the actual
! smoothing routine.     Upon completion it puts the resultant smoothed
! topography into netCDF variable "hsmth" (if exists) or into "h" of
! the same file leaving the original "hraw" unchanged.

! There are multiple versions of such smoothing/rx-limiting procedures
! which can be subdivided into two major groups: (i) the ones which are
! applied directly to "h" (so called rx-"smooth"ers); and (ii) the ones
! with first transform "h" into its logarithm,
!
!                    Lgh(i,j) = log[h(i,j)/hmin]
!
! and smooth it trying to achieve condition 
!
!      | Lgh(i+1,j) - Lgh(i+1,j) |  <  log[(1 + r_max)/(1 - r_max)]
!
! which is equivalent to rx < rx_max condition above, then transform
! it back, h(i,j)=hmin*exp[Lgh(i,j)]. These are called log-smoothers,
! or "lmsooth".  Besides this distinctions, there are others:
! handling boundary conditions at coast line; choice of discrete
! operator, etc. - resulting in a fair number of options.

! Usage: "smooth" or "lsmooth" take four arguments:
!
!              lsmooth hmin hmax r_max file.nc
! where
!        #1  hmin   desired minimum depth limit, meters, positive;
!        #2  hmax   --/ /-- maximum        --/ /--
!        #3  r_max  desired maximum r-factor, nondimensional;
!        #4- name of ROMS grid netCDF file;
!
! This file is merely a driver which decodes command-line arguments,
! reads model grid topography, calls relevant subroutine selected by
! CPP-switches (defined from compiler line to support multiple
! versions), and creates appropriate signature in the resultant grid
! file so all the parameters of execution these operators are
! documented to allow exact reproduction of the result relying
! solely on what is stored in netCDF file.

#if defined LOG_SMOOTHING
# define PROG_NAME 'lsmooth'
      program lsmooth
# elif defined LOG_SMOOTH_2
# define PROG_NAME 'lsmooth2'
      program lsmooth2
# elif defined LOG_SMOOTH_1
# define PROG_NAME 'lsmooth1'
      program lsmooth1
#elif defined LEGACY_LOG_SMOOTH
# define PROG_NAME 'lsmooth_legacy'
      program lsmooth_legacy
#elif defined NON_DECREASING_RX_COND
# define PROG_NAME 'cond_rx0_topo'
      program cond_rx0_topo
#else
# define PROG_NAME 'smooth'
      program smooth
#endif
      use smooth_topo_params
      implicit none
      character(len=80) grid
      integer(kind=2), allocatable, dimension(:,:) :: msk
      real(kind=8), dimension(:,:), allocatable ::  h, wrk,wrk1
      integer nargs, Lm,Mm, ncx,ncy, i,j, ierr, ncid, lgrd,
     &                                    lenstr, read_str_dble
      character(len=16) str

#include "netcdf.inc"

      nargs=iargc()
      if (nargs == 4) then
        call getarg(1,str) ; ierr=read_str_dble(str, hmin)
        if (ierr == 0) then
          call getarg(2,str) ; ierr=read_str_dble(str, hmax)
          if (ierr == 0) then
            call getarg(3,str) ; ierr=read_str_dble(str,r_max)
            if (ierr == 0) then
              call getarg(4,grid) ; lgrd=lenstr(grid)
              ierr=nf_open(grid(1:lgrd), nf_write, ncid)
              if (ierr == nf_noerr) then
                write(*,'(/1x,A,3(1x,A,F10.5),1x,3A/)')
     &             PROG_NAME, 'args: hmin =',hmin, 'hmax =',hmax,
     &            'r_max =',r_max, 'file = ''',grid(1:lgrd),'''.'
                goto 1
              else
                write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot ',
     &                          'open netCDF file ''', grid(1:lgrd),
     &                  ''' in read-write mode.', nf_strerror(ierr)
              endif
            else
              write(*,*) '### ERROR: Wrong argument #3, r_max ', str
            endif
          else
            write(*,*) '### ERROR: Wrong argument #2, hmax ', str
          endif
        else
          write(*,*) '### ERROR: Wrong argument #1, hmin ', str
        endif
      else
        write(*,'(/1x,A/,/16x,2A/)') 'This operator sould be used as',
     &                              PROG_NAME,
     &                            ' hmin hmax r_max roms_grid_file.nc'
      endif
      stop

  1   call roms_find_dims(ncid, grid, Lm,Mm,ierr)
      ncx=Lm+2 ; ncy=Mm+2
      allocate( h(ncx,ncy), wrk(ncx,ncy), wrk1(ncx,ncy),
     &                         msk(ncx,ncy), stat=ierr )
      if (ierr /= 0) then
         write(*,*) '### ERROR: Cannot allocate memory.'
        stop
      endif

      call get_var_by_name_double(ncid, 'hraw', h)
      call get_var_by_name_double(ncid, 'mask_rho', wrk)

      do j=1,ncy
        do i=1,ncx                      ! temporary use "wrk" just
          if (wrk(i,j) > 0.5D0) then    ! to read mask from the file;
            msk(i,j)=1                  ! from now on this code will
          else                          ! use integer(kind=2) mask
            msk(i,j)=0
          endif
          h(i,j)=max(hmin,min(hmax,h(i,j))) !<-- restrict
          wrk(i,j)=h(i,j)                   !<-- make another copy
        enddo
      enddo

#if defined LOG_SMOOTHING
      method='mask-dependent rx-conditioned log-smoothing'
C$OMP PARALLEL SHARED(Lm,Mm, msk, h, wrk,wrk1)
      call smooth_thread(Lm,Mm, msk, h, wrk,wrk1)
C$OMP END PARALLEL

# elif defined LOG_SMOOTH_2
      method='mask-dependent rx-conditioned log-smoothing v.2'
C$OMP PARALLEL SHARED(Lm,Mm, msk, h, wrk,wrk1)
      call smooth_thread(Lm,Mm, msk, h, wrk,wrk1)
C$OMP END PARALLEL

# elif defined LOG_SMOOTH_1
      method='mask-dependent rx-conditioned log-smoothing 90-deg only'
C$OMP PARALLEL SHARED(Lm,Mm, msk, h, wrk,wrk1)
      call smooth_thread(Lm,Mm, msk, h, wrk,wrk1)
C$OMP END PARALLEL

#elif defined LEGACY_LOG_SMOOTH
      method='mask-independent rx-conditioned log-smoothing'
C$OMP PARALLEL SHARED(Lm,Mm, msk, h, wrk,wrk1)
      call smooth_thread(Lm,Mm, msk, h, wrk,wrk1)
C$OMP END PARALLEL

#elif defined NON_DECREASING_RX_COND
      method='mask-dependent non-decreasing rx-conditioning'
C$OMP PARALLEL SHARED(ncx,ncy, msk, h,wrk)
      call cond_rx0_thread(ncx,ncy, msk, h,wrk)
C$OMP END PARALLEL

#else
      method='mask-dependent rx-conditioned z-space smoothing'
C$OMP PARALLEL SHARED(Lm,Mm, msk, h, wrk,wrk1)
      call smooth_thread(Lm,Mm, msk, h, wrk,wrk1)
C$OMP END PARALLEL
#endif

      call write_topo(ncid, grid, ncx,ncy, h)
      end


#ifdef NON_DECREASING_RX_COND
# include "cond_rx0_topo.F"
#else
# include "topo_smooth_subs.F"
#endif
