      program test_gshhs

! Reads GSHHS (Global Self-consistent Hierarchical High-resolution
! Shorelines) binary data and converts it into ASCII file.  De-facto
! this is merely a test program for C and Fortran interfaces provided
! by this package (read_gshhs.c and read_gshhs_data.F respectively),
! because reading and writing ASCII file of equivalent data is 50 to
! 100 times slower than for the original binary file, so there is no
! point in converting the original files. However ASCII file created
! by this program can be used for visual inspection, as well as to
! verify exact identity of C and Fortran readers.

! The GSHHS data is available from
!---- ----- ---- -- --------- ----
!      http://www.ngdc.noaa.gov/mgg/shorelines/gshhs.html
!      http://www.ngdc.noaa.gov/mgg/shorelines/data/gshhs/latest/

! as gshhs_X.b files, where X={f h i l c} for different resolutions,
! f = the finest, c = the lowest. This program is merely an interface
! to "read_gshhs" function written in C adopted from the original
! "gshhs.c" of Paul Wessel (wessel@soest.hawaii.edu) designed for the
! same purpose.

#define TIMING
c-#define USE_C_READER

      use mod_gshhs
      implicit none
      integer id, np, flag, container, ancestor, level, version,
     &                      greenwich, isrc, river, area_scale, i,j
      real(kind=8) lon,lat, area,area_full, west,east,south,north
      character(len=1) src
#ifdef USE_C_READER
      integer ierr, decode_flag, read_gshhs !<-- functions
#endif
#ifdef TIMING
      integer(kind=4) iclk(2), nclk, clk_rate, clk_max
      integer(kind=8) inc_clk, net_clk
      nclk=1 ; inc_clk=0 ; net_clk=0
      call system_clock(iclk(nclk), clk_rate, clk_max)
#endif

! Read GSHHS shoreline dataset...

#ifdef USE_C_READER
      nsegm=200000 ; npts_total=11000000
      allocate(isegm(nsegm), segmhdr(header_size,nsegm),
     &                            segmpt(2,npts_total))
      nsegm=0
      ierr=read_gshhs(nsegm, isegm, segmhdr,segmpt)
      if (ierr /= 0) stop
#else
      call read_gshhs_data
#endif

      write(*,'(1x,A,I8,1x,A,I10,1x,A)') 'retrieved', nsegm,
     &  'coastline segments,',  isegm(nsegm), 'points total'
#ifdef TIMING
      nclk=3-nclk
      call system_clock(iclk(nclk), clk_rate, clk_max)
      inc_clk=iclk(nclk)-iclk(3-nclk)
      if (inc_clk < 0) inc_clk=inc_clk+clk_max
      net_clk=net_clk+inc_clk
      write(*,'(1x,A,F5.2,1x,A)') 'reading complete in',
     &               dble(inc_clk)/dble(clk_rate), 'sec'
#endif

      write(*,'(/1x,A)') 'Converting into ASCII file, writting ...'
      open(unit=11, file='coast.dat', form='formatted')
      write(11,'(6x,A,1x,A,9x,A,6x,A,6x,A,5x,A)')  'id no_pts lev',
     &      'vr Gw s rv as area', 'west', 'east', 'south', 'north'
      do j=1,nsegm
        id=segmhdr(1,j) ; np=segmhdr(2,j) ; flag=segmhdr(3,j)
        if (j < nsegm) then
          if (isegm(j)+np /= isegm(j+1)) write(*,*)   'WARNING: ',
     &    'contradictory number of points for segment', j,isegm(j)
        endif

#ifdef USE_C_READER
            ierr=decode_flag(flag, level, version, greenwich,
#else
            call dec_segm_flag(flag, level, version, greenwich,
#endif
     &                              isrc,  river,  area_scale)
        if (isrc == 1) then
          src='W'
        else
          src='C'
        endif

         west=segmhdr(4,j) * 1.D-6 ; east=segmhdr(5,j) * 1.D-6
        south=segmhdr(6,j) * 1.D-6 ; north=segmhdr(7,j) * 1.D-6
         area=segmhdr(8,j) * 0.1D0 ; area_full=segmhdr(9,j)*0.1D0
        container=segmhdr(10,j)    ; ancestor=segmhdr(11,j)

        write(11,'(A,1x,I6,I8, 3I3,1x,A,2I3, F13.3, 4F10.5)')  'P',
     &               id, np, level, version, greenwich, src, river,
     &                 area_scale, area,  west, east, south, north
        do i=isegm(j),isegm(j)+np-1
          lon=segmpt(1,i) * 1.D-6  ; lat=segmpt(2,i) * 1.D-6
c**       write(11,'(2F10.5)') lon,lat
        enddo
      enddo  !<-- nsegm
      close(unit=11)

#ifdef TIMING
      nclk=3-nclk
      call system_clock(iclk(nclk), clk_rate, clk_max)
      inc_clk=iclk(nclk)-iclk(3-nclk)
      if (inc_clk < 0) inc_clk=inc_clk+clk_max
      net_clk=net_clk+inc_clk
      write(*,'(1x,A,F8.2,1x,A)') 'writing complete in',
     &              dble(inc_clk)/dble(clk_rate), 'sec'
#endif
      end
