      program tspline

! Computes time derivatives of time-dependent variables stored in
! netCDF file with the purpose of subsequent spline interpolation of
! that variable in time.

! Takes name of a netCDF file as an argument, attempts to find its
! time dimension and timing variable (trying unlimited dimension, if
! any, or dimension whose name ends with "...time", and then variable
! with that dimension as the sole dimension and with name ending as
! "...time").  Then it identifies all time-dependent variables, and
! builds time derivatives for each of using spline algorithm.  Due to
! global dependency inherent for splines, this means that all time
! records of the variable must participate in this computation,
! therefore, to avoid using too much memory (multi gigabyte-size files
! are generally expected) an automatic chunking algorithm is employed.
! Parameter "max_memory_size_bytes" below sets the maximum combined
! size of buffer arrays -- the program is designed to use two or three
! large arrays to hold one variable, its time derivative, and, in the
! case of quintic spline, second derivative as well).

! This program created an expanded version of the given file
! containing all the original variables + time derivatives where
! applicable.

#define TIMING
c--#define VERBOSE

      implicit none
      integer, parameter :: undefined=-9999
      integer, parameter :: max_memory_size_bytes=1024*1024*1024
      character(len=256) src_fname, trg_fname
      character(len=32) dimname, attname, varname, time_dim_name,
     &                                             time_var_name
      character(len=32), allocatable, dimension(:) :: vnames
      logical excl_vars, quintic, timedep, not_masked
      integer nargs, ncsrc, nctarg, ndims, unlimdimid, ngatts, nvars,
     &        vid,vidtrg, vartype, vardims, varatts, timdimid,timvarid,
     &        id, nrecs, size, nspc, net_ext_count, num_chunks, ichunk,
     &        chunk_size, jstr,jend, i,j,ksplit,m,n, is,ie, ierr, ltdm,
     &         old_fill_mode,  ltvr,
     &        max_buff_size_bytes, lsrc, ltrg, lvar, latt, lstr, lenstr

      integer, allocatable, dimension(:) :: dimid, dsize, vardimids,
     &                                                    start,count
      character(len=1), allocatable, dimension(:) :: cbfr
      integer(kind=1),  allocatable, dimension(:) :: ibfr1
      integer(kind=2),  allocatable, dimension(:) :: ibfr2
      integer(kind=4),  allocatable, dimension(:) :: ibffr
      real(kind=4), allocatable, dimension(:) :: bffr4, dbfr4, d2fr4
      real(kind=8), allocatable, dimension(:) :: bffr8, dbfr8, d2fr8,
     &                                                         time
      integer :: alloc_dims=0,  alloc_time=0,  alloc_cbfr=0,
     &           alloc_ibfr1=0, alloc_ibfr2=0, alloc_ibffr=0,
     &           nvnames=0,     alloc_bffr4=0, alloc_bffr8=0

      character(len=9), parameter :: signame='nc_format'
      character(len=256) string
      real(kind=4) :: spv4=0.0_4    ! Special values,
      real(kind=8) :: spv8=0.0_8    ! if will be found.
#ifdef TIMING
      real(kind=4) tstart, run_time
      real(kind=8) ReadSize, WrtSize, ReadTime, CompTime, WrtTime,
     &                                                    SyncTime
      integer(kind=4) iclk(2), nclk, clk_rate, clk_max
      integer(kind=8) :: inc_clk, net_read_size=0, net_wrt_size=0,
     &   net_clk=0, net_read_clk=0, net_wrt_clk=0, net_comp_clk=0,
     &                                             net_sync_clk=0
#endif

#include "netcdf.inc"

#ifdef TIMING
      call cpu_time(tstart)
      nclk=1
      call system_clock(iclk(nclk), clk_rate, clk_max)
# ifdef VERBOSE
      write(*,*) 'clk_rate=', clk_rate, ' clk_max =', clk_max
# endif
#endif

! Sort out arguments.
!===== === ==========

      quintic=.false.
      excl_vars=.false.
      ltvr=0; lsrc=0
      do i=1,len(time_var_name)
        time_var_name(i:i)=' '  !<-- blank out name
      enddo

      nargs=iargc()
      if (nargs > 0) then
        do n=1,nargs
          call getarg(n,string) ; lstr=lenstr(string)
          if (string(1:10) == '--qspl') then
            quintic=.true.
          elseif (string(1:10) == '--novars') then
            excl_vars=.true.
          elseif (string(1:10) == '--timevar=') then
            if (lstr > 10) then
              time_var_name=string(11:lstr) ; ltvr=lstr-10
            endif
          elseif(string(1:7) == '--vars=') then
            nvnames=1   !<-- at least one; each comma adds one more.
            do i=8,lstr
              if (string(i:i) == ',') nvnames=nvnames+1
            enddo
            allocate(vnames(nvnames))
            j=0 ; is=8
            do while(is < lstr+1)
              ie=is
              do while(string(ie:ie) /= ',' .and. ie < lstr)
                ie=ie+1
              enddo
              if (string(ie:ie) == ',') ie=ie-1
              j=j+1 ; vnames(j)=string(is:ie) ; is=ie+2
            enddo
          elseif(lsrc == 0) then
            src_fname=string(1:lstr) ; lsrc=lenstr(src_fname)
          else
            write(*,'(/1x,2A/1x,A/)')   '"tspline" expects a single ',
     &                      'netCDF file for the entire time series.',
     &        'Use "mtspline" to process sequence of multiple files.'
            stop
          endif
        enddo

        if (quintic) then
          write(*,*)   'Building quintic spline.'
          max_buff_size_bytes=max_memory_size_bytes/3
        else
          max_buff_size_bytes=max_memory_size_bytes/2
        endif
        if (excl_vars) write(*,*)   'Saving derivatives only.'
        if (ltvr > 0) write(*,'(1x,4A)') 'user supplied time ',
     &             'variable: ''', time_var_name(1:ltvr), ''''

        if (nvnames > 0) then
          write(*,'(1x,A,I3,1x,3A)')      'user supplied variables,',
     &    nvnames, 'total: ''', vnames(1)(1:lenstr(vnames(1))), ''''
          do i=2,nvnames
            write(*,'(36x,3A)') '''', vnames(i)(1:lenstr(vnames(i))),
     &                                                          ''''
          enddo
        endif
      endif
      if (lsrc == 0) then
        write(*,'(/1x,A/8x,2A/1x,A/8x,2A/)')   'Usage:', 'tspline ',
     &              'file.nc', 'or', 'tspline [--qspl] [--novars] ',
     &          '[--timevar=name] [--vars=name1,name2,...] file.nc'
        stop
      endif

! Create target name: add prefix to source file name, however the
! actual file name may contain directory nome(s), so find the last
! slash (if any) and insert prefix there.

      ie=lsrc
      do while(ie > 1 .and. src_fname(ie:ie) /= '/')
        ie=ie-1
      enddo
      if (src_fname(ie:ie) == '/') then
        trg_fname=src_fname(1:ie)/ /'ts_'/ /src_fname(ie+1:lsrc)
      else
        trg_fname=                  'ts_'/ /src_fname(ie:lsrc)
      endif
      ltrg=lenstr(trg_fname)


! Open source netCDF file:  Try to identify time dimension: at first,
!===== ====== ====== =====  check whether user-supplied time variable
! has only one dimension, and this is it; if no user-supplied time
! variable, then make another attempt to identify it below...


      ierr=nf_open(src_fname(1:lsrc), nf_nowrite, ncsrc)
      if (ierr == nf_noerr) then
        write(*,'(/1x,3A)') 'Processing ''',src_fname(1:lsrc),'''...'
        ierr=nf_inq(ncsrc, ndims, nvars, ngatts, unlimdimid)
        if (ierr == nf_noerr) then
#ifdef VERBOSE
          write(*,'(4(1x,A,I4))') 'ncsrc: ndims =',ndims,'nvars =',
     &          nvars,'ngatts =',ngatts,'unlimdimid =',unlimdimid
#endif
          if (ndims > alloc_dims) then
            alloc_dims=ndims
            if (allocated(dimid)) then
              deallocate(dimid) ; deallocate(dsize)
              deallocate(vardimids)
              deallocate(start) ; deallocate(count)
            endif
            allocate(dimid(ndims)) ; allocate(dsize(ndims))
            allocate(vardimids(ndims))
            allocate(start(ndims)) ; allocate(count(ndims))
          endif

! Check whether the user-specified time variable exists, verify that
! it has only one dimension: this is the time dimension and its length
! is the numbers of records. Use ltvr>0 as flag that the name of time
! variable is already known from above.

          timvarid=undefined ; ltdm=0
          timdimid=undefined ; nrecs=0
          do i=1,len(time_dim_name)
            time_dim_name(i:i)=' '  !<-- blank out name
          enddo

          if (ltvr > 0) then
            ierr=nf_inq_varid(ncsrc, time_var_name(1:ltvr), vid)
            if (ierr == nf_noerr) then
              ierr=nf_inq_var(ncsrc, vid,  varname,  vartype,
     &                             vardims, vardimids, varatts)
              if (ierr == nf_noerr) then
                if (vardims == 1) then
                  timvarid=vid
                  timdimid=vardimids(vardims)
                  ierr=nf_inq_dim(ncsrc, timdimid, dimname, size)
                  if (ierr == nf_noerr) then
                    time_dim_name=dimname ; ltdm=lenstr(time_dim_name)
                    nrecs=size
                    write(*,'(1x,2A,I3,1x,3A,I5,1x,3A)') 'identified ',
     &                   'dimension #', timdimid, 'as time dimension ',
     &                    time_dim_name(1:ltdm), ' =', nrecs,  'of ''',
     &                                    time_var_name(1:ltvr), '''.'
                  else
                    write(*,'(/1x,2A,I3/12x,3A/12x,A/)') '### ERROR: ',
     &                 'Cannot determine name and size of dimension #',
     &                  vardims, 'in netCDF file ''',src_fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
                  endif
                else
                  write(*,'(/1x,4A,I4,1x,A/)')      '### ERROR: User ',
     &             'specified time variable ''', time_var_name(1:ltvr),
     &                   ''' has', vardims, 'dimensions instead of 1.'
                endif
              else
                write(*,'(/1x,2A,I3/12x,4A/12x,A/)')     '### ERROR: ',
     &               'Cannot make general inquiry for variable #', vid,
     &                      'in ', 'netCDF file ''', src_fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,6A/)')  '### ERROR: User specified time ',
     &                           'variable ''', time_var_name(1:ltvr),
     &                 ''' not found in ''', src_fname(1:lsrc), '''.'
            endif
            if (timvarid == undefined .or. timdimid == undefined) stop
          endif  !<-- ltvr


! Create target file:
!======= ====== =====

          ierr=nf_create(trg_fname(1:ltrg), nf_clobber +
     &                          nf_64bit_offset, nctarg)
          if (ierr == nf_noerr) then
            ierr=nf_set_fill(nctarg, nf_nofill, old_fill_mode)
            write(*,'(4x,3A)')   'Created netCDF file ''',
     &                            trg_fname(1:ltrg), '''.'

! Sort out dimensions:   If time dimension is not identified yet, make
!----- --- -----------   another attempt: assume that it is  (i) either
! unlimited dimension; or (ii) a dimension with name denting as "time".

            do id=1,ndims
              ierr=nf_inq_dim(ncsrc, id, dimname, dsize(id))
              if (ierr == nf_noerr) then
                lvar=lenstr(dimname)

                if (timdimid == undefined) then
                  if (id == unlimdimid) then
                    timdimid=id ; nrecs=dsize(id)
                    time_dim_name=dimname(1:lvar) ; ltdm=lvar
                    write(*,'(1x,2A,I3,1x,3A,I5)')    'identified ',
     &                    'dimension #', timdimid, 'as unlimited, ',
     &                     time_dim_name(1:ltdm),  ' =',  nrecs
                  elseif (lvar > 3) then
                    if (dimname(lvar-3:lvar) == 'time') then
                      timdimid=id ; nrecs=dsize(id)
                      time_dim_name=dimname(1:lvar) ; ltdm=lvar
                      write(*,'(1x,2A,I3,1x,3A,I5)')  'identified ',
     &               'dimension #', timdimid, 'as time dimension, ',
     &                     time_dim_name(1:ltdm),  ' =',  nrecs
                    endif
                  endif
                endif

                if (id == unlimdimid) then
                  j=nf_unlimited
                else
                  j=dsize(id)
                endif
                ierr=nf_def_dim(nctarg, dimname(1:lvar), j, dimid(id))
                if (ierr == nf_noerr) then
                  write(*,'(6x,3A,I5)') 'Created dimension ''',
     &                            dimname(1:lvar), '''=', dsize(id)
                else
                  write(*,'(/1x,6A/12x,A/)')    '### ERROR: Cannot ',
     &                        'define dimension ''', dimname(1:lvar),
     &                     ''' in netCDF file ''', trg_fname(1:ltrg),
     &                                      '''.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,2A,I3/12x,3A/12x,A/)')   '### ERROR: ',
     &               'Cannot determine name and size of dimension #',
     &                id, 'in netCDF file ''',     src_fname(1:lsrc),
     &                                      '''.', nf_strerror(ierr)
              endif
            enddo

! Copy global attributes

            do id=1,ngatts
              ierr=nf_inq_attname(ncsrc, nf_global, id, attname)
              if (ierr == nf_noerr) then
                latt=lenstr(attname) ; lvar=lenstr(signame)
                if (attname(1:latt) /= signame(1:lvar)) then
                  ierr=nf_copy_att(ncsrc, nf_global, attname(1:latt),
     &                                            nctarg, nf_global)
                  if (ierr == nf_noerr) then
                    write(*,'(8x,3A)')  'Copied global attribute ''',
     &                                          attname(1:latt), '''.'
                  else
                    write(*,'(/1x,4A/12x,3A/12x,A/)')   '### ERROR: ',
     &             'Cannot copy global attribute ''', attname(1:latt),
     &             '''',  'into netCDF file ''',    trg_fname(1:ltrg),
     &                                       '''.', nf_strerror(ierr)
                  endif
                endif
              else
                write(*,'(/1x,2A,I3,1x,3A/12x,A/)')     '### ERROR: ',
     &                  'Cannot determine name of global attribute #',
     &                   id, 'from netCDF file ''', src_fname(1:lsrc),
     &                                       '''.', nf_strerror(ierr)
              endif
            enddo

! Define variables, their dimensions and attributes:  All variables
!------- ---------- ----- --- ------ --- -----------  fall into three
! categories: (1) time variable (there is only one); (2) variables
! which have time dimension; (3) variables which do not have it.
! If the name of time variable not supplied by the user, identify it
! just below as meeting two criteria:  (id) its name ends with "time";
! (ii) has only one dimension and it is the time dimension at it was
! already identified above.

            do vid=1,nvars
              ierr=nf_inq_var(ncsrc, vid, varname,  vartype,
     &                         vardims, vardimids, varatts)
              if (ierr == nf_noerr) then
                lvar=lenstr(varname)
                if (timvarid == undefined) then
                  if (vardims == 1 .and. lvar > 3) then
                    if (varname(lvar-3:lvar) == 'time' .and.
     &                  vardimids(vardims) == timdimid) then
                      timvarid=vid
                      time_var_name=varname(1:lvar) ; ltvr=lvar
                      write(*,*) 'identified time variable ''',
     &                             time_var_name(1:ltvr), ''''
                    endif
                  endif
                endif

                size=1 ; timedep=.false.
                do i=1,vardims
                  if (vardimids(i) == timdimid) then
                    timedep=.true.
                  else
                    size=size*dsize(vardimids(i))
                  endif
                enddo

! Logical switch "not_masked" is relevant only if user specified the
! names of desired variables explicitly (otherwise it os always true).

                if (vid == timvarid .or. .not.timedep) then
                  not_masked=.true.
                elseif (nvnames > 0) then
                  not_masked=.false.
                  do i=1,nvnames
                    lstr=lenstr(vnames(i))
                    if (lvar == lstr .and. varname(1:lvar) ==
     &                      vnames(i)(1:lstr)) not_masked=.true.
                  enddo
                else
                  not_masked=.true.
                endif
                if (not_masked) then

! Note that array "vardimids"  goes straight from "nf_inq_var" above
! (where it is output) into "nf_def_var" (as input) which implies that
! the dimension IDs are exactly the same in both files.

                  if (.not.excl_vars .or. .not.timedep .or. vid ==
     &                                                 timvarid) then
                    ierr=nf_def_var(nctarg, varname(1:lvar),  vartype,
     &                                     vardims, vardimids, vidtrg)
                    if (ierr == nf_noerr) then
                      write(*,'(6x,3A,I10)') 'Created variable ''',
     &                        varname(1:lvar), ''', size =', size

                      do i=1,varatts
                        ierr=nf_inq_attname(ncsrc, vid, i, attname)
                        latt=lenstr(attname)
                        ierr=nf_copy_att(ncsrc, vid, attname(1:latt),
     &                                               nctarg, vidtrg)
                        if (ierr == nf_noerr) then
                          write(*,'(8x,3A)') 'Copied attribute ''',
     &                                        attname(1:latt), '''.'
                        else
                          write(*,'(/1x,5A/12x,3A/12x,A/)')
     &                          '### ERROR: Cannot copy attribute ''',
     &                          attname(1:latt), ''' for variable ''',
     &                                    varname(1:lvar), ''' into ',
     &                            'netCDF file ''', trg_fname(1:ltrg),
     &                                       '''.', nf_strerror(ierr)
                        endif
                      enddo
                    else
                      write(*,'(/1x,4A/12x,3A/12x,A/)') '### ERROR: ',
     &                   'Cannot create variable ''', varname(1:lvar),
     &                   '''', 'in netCDF file ''', trg_fname(1:ltrg),
     &                                       '''.', nf_strerror(ierr)
                    endif
                  endif

! Create variables for time derivatives.

                  if (timedep .and. vid /= timvarid .and.
     &                             ierr == nf_noerr) then
                    ierr=nf_def_var(nctarg,'dt_'/ /varname(1:lvar),
     &                        vartype, vardims, vardimids, vidtrg)
                    if (ierr == nf_noerr) then
                      ierr=nf_put_att_text(nctarg,vidtrg,'long_name',
     &                             lvar+17, 'time derivative for '/ /
     &                                              varname(1:lvar))
                      if (ierr == nf_noerr) then
                        write(*,'(6x,4A,I10)')   'Created variable ',
     &                   '''dt_', varname(1:lvar),''', size =', size
                        ierr=nf_inq_att(ncsrc,vid,'_FillValue',is,ie)
                        if (ierr == nf_noerr) then
                          ierr=nf_copy_att(ncsrc, vid, '_FillValue',
     &                                                nctarg, vidtrg)
                          if (ierr /= nf_noerr) then
                            write(*,'(/1x,2A/12x,A)')   '### ERROR: ',
     &                          'Cannot copy attibute ''_FillValue''',
     &                                              nf_strerror(ierr)
                          endif
                          if (is == nf_real) then
                            if (spv4 == 0.0_4) then
                              ierr=nf_get_att_real( ncsrc, vid,
     &                                       '_FillValue', spv4)
                            endif
                          elseif (is == nf_double) then
                            if (spv8 == 0.0_8) then
                              ierr=nf_get_att_double(ncsrc, vid,
     &                                       '_FillValue', spv8)
                            endif
                          endif
                        else
                          ierr=nf_noerr
                        endif
                      else
                        write(*,'(/1x,4A/12x,A)')        '### ERROR: ',
     &                                'Cannot put attribute for ''dt_',
     &                       varname(1:lvar), '''.', nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,4A/12x,3A/12x,A/)')  '### ERROR: ',
     &                 'Cannot create variable ''dt_', varname(1:lvar),
     &                    '''', 'in netCDF file ''', trg_fname(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif

                    if (quintic .and. ierr == nf_noerr) then
                      ierr=nf_def_var(nctarg, 'd2_'/ /varname(1:lvar),
     &                           vartype, vardims, vardimids, vidtrg)
                      if (ierr == nf_noerr) then
                        ierr=nf_put_att_text(nctarg, vidtrg,
     &                                       'long_name', lvar+19,
     &                                    'second derivative for '/ /
     &                                              varname(1:lvar))
                        if (ierr == nf_noerr) then
                          write(*,'(6x,4A,I10)')           'Created ',
     &                              'variable ''d2_', varname(1:lvar),
     &                                            ''', size =', size
                          ierr=nf_inq_att(ncsrc,vid,'_FillValue',
     &                                                      is,ie)
                          if (ierr == nf_noerr) then
                            ierr=nf_copy_att(ncsrc, vid, '_FillValue',
     &                                                nctarg, vidtrg)
                            if (ierr /= nf_noerr) then
                              write(*,'(/1x,2A/12x,A)') '### ERROR: ',
     &                          'Cannot copy attibute ''_FillValue''',
     &                                              nf_strerror(ierr)
                            endif
                          else
                            ierr=nf_noerr
                          endif
                        else
                          write(*,'(/1x,4A/12x,A)')'### ERROR: ',
     &                                'Cannot put attribute for ''d2_',
     &                       varname(1:lvar), '''.', nf_strerror(ierr)
                        endif
                      else
                        write(*,'(/1x,3A/12x,3A/12x,A/)')
     &                       '### ERROR: Cannot create variable ''d2_',
     &                        varname(1:lvar),'''','in netCDF file ''',
     &                                        trg_fname(1:ltrg), '''.',
     &                                               nf_strerror(ierr)
                      endif
                    endif !<-- quintic
                  endif
#ifdef VERBOSE
                else
                  write(*,*) 'ignore variable ''',varname(1:lvar),''''
#endif
                endif !<-- not_masked
              else
                write(*,'(/1x,2A,I3/12x,4A/12x,A/)')   '### ERROR: ',
     &                  'Cannot make general inquiry for variable #',
     &                vid, 'in ', 'netCDF file ''', src_fname(1:lsrc),
     &                                       '''.', nf_strerror(ierr)
              endif
            enddo
            ierr=nf_enddef(nctarg)
            write(*,'(6x,3A)')  'Leaving definition mode for ''',
     &                                    trg_fname(1:ltrg), '''.'
          else
            write(*,'(/1x,4A/12x,A/)')     '### ERROR: Cannot create ',
     &     'netCDF file ''',trg_fname(1:ltrg),'''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,4A/12x,A/)')         '### ERROR: Cannot make ',
     &        'general inquiry into netCDF file ''', src_fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
        endif
      else
        write(*,'(/1x,4A/12x,A/)')    '### ERROR: Cannot open netCDF ',
     &                  'file ''', src_fname(1:lsrc),''' for reading.',
     &                                               nf_strerror(ierr)
        stop
      endif
      if (ltvr == 0) then
        write(*,'(/1x,4A/)') '### ERROR: No time variable is found ',
     &                             'in ''', src_fname(1:lsrc), '''.'

        goto 98                                            !--> BREXIT
      endif
      if (ierr /= nf_noerr) goto 98                        !--> BREXIT

#ifdef TIMING
      nclk=3-nclk
      call system_clock(iclk(nclk), clk_rate,clk_max)
      inc_clk=iclk(nclk)-iclk(3-nclk)
      if (inc_clk < 0) inc_clk=inc_clk+clk_max
      net_clk=net_clk+inc_clk
      net_sync_clk=net_sync_clk+inc_clk
#endif


!      *****    *********    ******   *******    *********
!    ***   ***  *  ***  *   *   ***   ***   ***  *  ***  *
!    ***           ***     **   ***   ***   ***     ***
!      *****       ***    ***   ***   ***   **      ***
!          ***     ***    *********   ******        ***
!    ***   ***     ***    ***   ***   ***  **       ***
!      *****       ***    ***   ***   ***   ***     ***


      do vid=1,nvars
        ierr=nf_inq_var(ncsrc, vid, varname,  vartype,
     &                    vardims, vardimids, varatts)
        if (ierr == nf_noerr) then
          lvar=lenstr(varname)
          timedep=.false.
          do i=1,vardims
            if (vardimids(i) == timdimid) timedep=.true.
          enddo

          if (nvnames > 0) then
            not_masked=.false.
            do i=1,nvnames
              lstr=lenstr(vnames(i))
              if (lvar == lstr .and. varname(1:lvar) ==
     &                vnames(i)(1:lstr)) not_masked=.true.
            enddo
          else
            not_masked=.true.
          endif
#ifdef VERBOSE
          write(*,*) 'vid =',vid, nvars,' name = ''',varname(1:lvar),
     &      ''', ', 'timedep =',timedep, ' not_masked =',not_masked
#endif

          if (vid == timvarid) then

! Process time variable, basically copy from one file to the other,
! however, CAUTION: Although it may seem that the entire variable can
! be read and written as a whole, using function nf_put_var_ instead of
! more complicated nf_put_vara_ (...start,count,...), an attempt to do
! so causes it not be written into the target file if its dimension is
! unlimited and is zero-size (meaning that nothing is written yet) at
! the time when attempt is made [?? it appears that nf_put_var_ infers
! the dimensions of the variable to be written from the file before
! writing it, and, as the result, it may be unsafe if the variable has
! unlimited dimension which is about to change (increase) as the result
! of writing].

            if (nrecs > alloc_time) then
              if (allocated(time)) deallocate(time)
              allocate(time(nrecs)) ; alloc_time=nrecs
              write(*,*) 'allocated time variable array',
     &                           ', size =', alloc_time
            endif
            start(1)=1 ; count(1)=nrecs
            ierr=nf_get_vara_double(ncsrc, vid, start,count, time)
            if (ierr == nf_noerr) then
              write(*,'(8x,3A)')  'read ''', varname(1:lvar), ''''
#ifdef TIMING
              nclk=3-nclk
              call system_clock(iclk(nclk), clk_rate,clk_max)
              inc_clk=iclk(nclk)-iclk(3-nclk)
              if (inc_clk < 0) inc_clk=inc_clk+clk_max
              net_clk=net_clk+inc_clk
              net_read_clk=net_read_clk+inc_clk
              net_read_size=net_read_size+8*nrecs
#endif

              ierr=nf_inq_varid(nctarg, time_var_name(1:ltvr),vidtrg)
              if (ierr == nf_noerr) then
                ierr=nf_put_vara_double(nctarg, vidtrg, start,count,
     &                                                          time)
                if (ierr == nf_noerr) then
                  write(*,'(8x,5A)') 'wrote ''',time_var_name(1:ltvr),
     &                          ''' into ''', trg_fname(1:ltrg), ''''

#ifdef TIMING
! Note: despite the relatively small data writen, writing out time
! variable causes inflation of the unlimited dimension of the target
! file and may take very significant amount of time. For this reason
! this time is counted as _sync_ rather than _wrt_ and is reported
! accordingly; data associated with time variable is usially ignored,
! hence not counted.

                  nclk=3-nclk
                  call system_clock(iclk(nclk), clk_rate,clk_max)
                  inc_clk=iclk(nclk)-iclk(3-nclk)
                  if (inc_clk < 0) inc_clk=inc_clk+clk_max
                  net_clk=net_clk+inc_clk
                  net_sync_clk=net_sync_clk+inc_clk
c**               net_wrt_clk=net_wrt_clk+inc_clk
c**               net_wrt_size=net_wrt_size+8*nrecs
#endif
                else
                  write(*,'(/1x,6A/12x,A/)')'### ERROR: Cannot write ',
     &              'variable ''', time_var_name(1:ltvr), ''' into ''',
     &                     trg_fname(1:ltrg), '''.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot find ',
     &                     'variable ID for ''', time_var_name(1:ltvr),
     &         ''' in ''', trg_fname(1:ltrg), '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot read ',
     &                    'variable ''', varname(1:lvar), ''' from ''',
     &                     src_fname(1:lsrc), '''.', nf_strerror(ierr)
            endif
          elseif(timedep .and. not_masked) then

! Process time-dependent variable.

            if (vartype == nf_float) then
              size=4*nrecs
            elseif (vartype==nf_double) then
              size=8*nrecs
            else
              write(*,'(/8x,4A/)') '### ERROR: Unrecognized type ',
     &                     'of variable ''', varname(1:lvar), '''.'
              ierr=nf_noerr+1
            endif

            m=size ; ksplit=0
            do j=1,vardims-1
              if (m < max_buff_size_bytes) then
                size=m ; ksplit=j ; m=size*dsize(vardimids(j))
              endif
#ifdef VERBOSE
              write(*,'(2(1x,A,I3),1x,A,I5,3(1x,A,I10))')   'j =', j,
     &            'ksplit =', ksplit, 'dsize =', dsize(vardimids(j)),
     &        'm =',m,'size =',size,'max_size =',max_buff_size_bytes
#endif
            enddo
            if (ksplit==0) then
              ksplit=2 ; chunk_size=1; num_chunks=dsize(vardimids(2))
            else
              chunk_size=max_buff_size_bytes/size
              num_chunks=( dsize(vardimids(ksplit))+chunk_size-1
     &                                            )/chunk_size
            endif
                                      ! fold all dimensions above
            net_ext_count=1           ! "ksplit" into single counter
            do j=ksplit+1,vardims-1
              net_ext_count=net_ext_count*dsize(vardimids(j))
            enddo

            write(*,'(1x,A,I2,2x,A,I5,2x,A,I4)') 'ksplit =',ksplit,
     &      'net_ext_count =',net_ext_count, 'num_chunks =',num_chunks

            do m=0,net_ext_count-1
              do ichunk=0,num_chunks-1
                do j=1,ksplit-1
                  start(j)=1 ; count(j)=dsize(vardimids(j))
                enddo

                jstr=1+ichunk*chunk_size
                jend=min(jstr+chunk_size-1, dsize(vardimids(ksplit)))
                start(ksplit)=jstr ; count(ksplit)=jend-jstr+1

                size=1           !--> unfold dimensions above "ksplit"
                do j=ksplit+1,vardims-1
                  start(j)=1+mod(m/size, dsize(vardimids(j)))
                  count(j)=1
                  size=size*dsize(vardimids(j)) !<-- recursive
                enddo
                start(vardims)=1 ; count(vardims)=nrecs

                nspc=1                ! determine the length of would
                do j=1,vardims-1      ! be "spatial" dimension an array
                  nspc=nspc*count(j)  ! (nspc,nrecs)
                enddo

                size=nspc*nrecs  !<-- needed to hold all time records

                if (vartype == nf_float) then
                  if (size>alloc_bffr4 .or. .not.allocated(dbfr4)) then
                    if (allocated(dbfr4)) deallocate(dbfr4)
                    if (allocated(bffr4)) deallocate(bffr4)
                    if (allocated(d2fr4)) deallocate(d2fr4)
                    allocate(bffr4(size)) ; allocate(dbfr4(size))
                    if (quintic) allocate(d2fr4(size))
                    alloc_bffr4=size
                    write(*,*) 'allocated nf_real workspace', size
                  endif
                elseif (vartype==nf_double) then
                  if (size>alloc_bffr8 .or. .not.allocated(dbfr8)) then
                    if (allocated(dbfr8)) deallocate(dbfr8)
                    if (allocated(bffr8)) deallocate(bffr8)
                    if (allocated(d2fr8)) deallocate(d2fr8)
                    allocate(bffr8(size)) ; allocate(dbfr8(size))
                    if (quintic) allocate(d2fr8(size))
                    alloc_bffr8=size
                    write(*,*) 'allocated nf_double workspace', size
                  endif
                endif

                if (vartype == nf_float) then
                  ierr=nf_get_vara_real (ncsrc, vid, start,count,bffr4)
                elseif (vartype == nf_double) then
                  ierr=nf_get_vara_double(ncsrc,vid, start,count,bffr8)
                else
                  write(*,'(/8x,4A/)') '### ERROR: Unrecognized ',
     &                'type of variable ''', varname(1:lvar), '''.'
                  ierr=nf_noerr+1
                endif
                if (ierr == nf_noerr) then
#ifdef TIMING
                  nclk=3-nclk
                  call system_clock(iclk(nclk), clk_rate,clk_max)
                  inc_clk=iclk(nclk)-iclk(3-nclk)
                  if (inc_clk < 0) inc_clk=inc_clk+clk_max
                  net_clk=net_clk+inc_clk
                  net_read_clk=net_read_clk+inc_clk
                  if (vartype == nf_float) then
                    net_read_size=net_read_size+4*size
                  elseif (vartype == nf_double) then
                    net_read_size=net_read_size+8*size
                  endif
#endif

! Computational part goes here.

                  if (quintic) then
                    if (vartype == nf_float) then
C$OMP PARALLEL SHARED(nspc, nrecs, time, bffr4, dbfr4, d2fr4, spv4)
                      call comp_quintic_real(   nspc,  nrecs,
     &                             time, bffr4, dbfr4, d2fr4, spv4)
C$OMP END PARALLEL
                    elseif (vartype == nf_double) then
C$OMP PARALLEL SHARED(nspc, nrecs, time, bffr8, dbfr8, d2fr8, spv8)
                      call comp_quintic_double( nspc,  nrecs,
     &                             time, bffr8, dbfr8, d2fr8, spv8)
C$OMP END PARALLEL
                    endif
                  else
                    if (vartype == nf_float) then
C$OMP PARALLEL SHARED(nspc, nrecs, time, bffr4, dbfr4, spv4)
                      call comp_spline_real(    nspc,  nrecs,
     &                             time, bffr4, dbfr4, spv4)
C$OMP END PARALLEL
                    elseif (vartype == nf_double) then
C$OMP PARALLEL SHARED(nspc, nrecs, time, bffr8, dbfr8, spv8)
                      call comp_spline_double(  nspc,  nrecs,
     &                             time, bffr8, dbfr8, spv8)
C$OMP END PARALLEL
                    endif
                  endif

#ifdef TIMING
                  nclk=3-nclk
                  call system_clock(iclk(nclk), clk_rate,clk_max)
                  inc_clk=iclk(nclk)-iclk(3-nclk)
                  if (inc_clk < 0) inc_clk=inc_clk+clk_max
                  net_clk=net_clk+inc_clk
                  net_comp_clk=net_comp_clk+inc_clk
#endif
                  if (.not.excl_vars) then
                    ierr=nf_inq_varid(nctarg, varname(1:lvar), vidtrg)
                    if (ierr == nf_noerr) then
                      if (vartype == nf_float) then
                        ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                                   count, bffr4)
                      elseif (vartype == nf_double) then
                        ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                                   count, bffr8)
                      endif
                      if (ierr == nf_noerr) then
                        write(*,'(15x,3A)') 'wrote ''',varname(1:lvar),
     &                                                            ''''
#ifdef TIMING
                        nclk=3-nclk
                        call system_clock(iclk(nclk), clk_rate,clk_max)
                        inc_clk=iclk(nclk)-iclk(3-nclk)
                        if (inc_clk < 0) inc_clk=inc_clk+clk_max
                        net_clk=net_clk+inc_clk
                        net_wrt_clk=net_wrt_clk+inc_clk
                        if (vartype == nf_float) then
                          net_wrt_size=net_wrt_size + 4*size
                        elseif (vartype == nf_double) then
                          net_wrt_size=net_wrt_size + 8*size
                        endif
#endif
                      else
                        write(*,'(/1x,6A/12x,A/)')       '### ERROR: ',
     &                     'Cannot write variable ''', varname(1:lvar),
     &                          ''' into ''', trg_fname(1:ltrg), '''.',
     &                                               nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                      'find variable ID for ''', varname(1:lvar),
     &                                   ''' in ''', trg_fname(1:ltrg),
     &                                         '''.',nf_strerror(ierr)
                    endif
                  endif  !<-- .not.excl_vars

                  if (ierr == nf_noerr) then
                    ierr=nf_inq_varid(nctarg, 'dt_'/ /varname(1:lvar),
     &                                                        vidtrg)
                    if (ierr == nf_noerr) then
                      if (vartype == nf_float) then
                        ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                                   count, dbfr4)
                      elseif (vartype == nf_double) then
                        ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                                   count, dbfr8)
                      endif
                      if (ierr == nf_noerr) then
                        write(*,'(2I7,1x,4A)') m,ichunk, 'wrote time ',
     &                       'derivative of ''', varname(1:lvar), ''''
#ifdef TIMING
                        nclk=3-nclk
                        call system_clock(iclk(nclk), clk_rate,clk_max)
                        inc_clk=iclk(nclk)-iclk(3-nclk)
                        if (inc_clk < 0) inc_clk=inc_clk+clk_max
                        net_clk=net_clk+inc_clk
                        net_wrt_clk=net_wrt_clk+inc_clk
                        if (vartype == nf_float) then
                          net_wrt_size=net_wrt_size + 4*size
                        elseif (vartype == nf_double) then
                          net_wrt_size=net_wrt_size + 8*size
                        endif
#endif
                      else
                        write(*,'(/1x,6A/12x,A/)')'### ERROR: Cannot ',
     &                  'write time derivative of ''', varname(1:lvar),
     &                                 ''' into ''', trg_fname(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                   'find variable ID for ''dt_', varname(1:lvar),
     &                                   ''' in ''', trg_fname(1:ltrg),
     &                                         '''.',nf_strerror(ierr)
                    endif
                  endif

                  if (quintic .and. ierr == nf_noerr) then
                    ierr=nf_inq_varid(nctarg, 'd2_'/ /varname(1:lvar),
     &                                                         vidtrg)
                    if (ierr == nf_noerr) then
                      if (vartype == nf_float) then
                        ierr=nf_put_vara_real  (nctarg, vidtrg, start,
     &                                                   count, d2fr4)
                      elseif (vartype == nf_double) then
                        ierr=nf_put_vara_double(nctarg, vidtrg, start,
     &                                                   count, d2fr8)
                      endif
                      if (ierr == nf_noerr) then
                        write(*,'(15x,4A)') 'wrote second derivative ',
     &                                   'of ''', varname(1:lvar),''''
#ifdef TIMING
                        nclk=3-nclk
                        call system_clock(iclk(nclk), clk_rate,clk_max)
                        inc_clk=iclk(nclk)-iclk(3-nclk)
                        if (inc_clk < 0) inc_clk=inc_clk+clk_max
                        net_clk=net_clk+inc_clk
                        net_wrt_clk=net_wrt_clk+inc_clk
                        if (vartype == nf_float) then
                          net_wrt_size=net_wrt_size + 4*size
                        elseif (vartype == nf_double) then
                          net_wrt_size=net_wrt_size + 8*size
                        endif
#endif
                      else
                        write(*,'(/1x,6A/12x,A/)')       '### ERROR: ',
     &                          'Cannot write second derivative of ''',
     &                                   varname(1:lvar), ''' into ''',
     &                                        trg_fname(1:ltrg), '''.',
     &                                               nf_strerror(ierr)
                      endif
                    else
                      write(*,'(/1x,6A/12x,A/)')  '### ERROR: Cannot ',
     &                   'find variable ID for ''d2_', varname(1:lvar),
     &                                   ''' in ''', trg_fname(1:ltrg),
     &                                        '''.', nf_strerror(ierr)
                    endif
                  endif  !<-- quintic
                else
                  write(*,'(/1x,6A/12x,A/)')      '### ERROR: Cannot ',
     &               'read variable ''', varname(1:lvar), ''' from ''',
     &                     src_fname(1:lsrc), '''.', nf_strerror(ierr)
                endif
                if (ierr /= nf_noerr) goto 98              !--> BREXIT
              enddo  !<-- ichunk=0,num_chunks-1
            enddo  !<-- m=0,net_ext_count-1
          elseif(.not. timedep) then

! Copy single-record variables: basically read as the whole and write
! into the target file. This operation is completely passive in sense
! that no computation or reshaping takes place.

            size=1
            do j=1,vardims
              size=size*dsize(vardimids(j))
            enddo

            if (vartype == nf_char .and. size > alloc_cbfr) then
              if (allocated(cbfr)) deallocate(cbfr)
              allocate(cbfr(size)) ;   alloc_cbfr=size
              write(*,*) 'allocated nf_char workspace',size
            elseif (vartype == nf_byte .and. size>alloc_ibfr1) then
              if (allocated(ibfr1)) deallocate(ibfr1)
              allocate(ibfr1(size)) ; alloc_ibfr1=size
              write(*,*) 'allocated nf_byte workspace',size
            elseif (vartype == nf_short .and. size>alloc_ibfr2) then
              if (allocated(ibfr2)) deallocate(ibfr2)
              allocate(ibfr2(size)) ; alloc_ibfr2=size
              write(*,*) 'allocated nf_int2 workspace',size
            elseif (vartype == nf_int .and. size > alloc_ibffr) then
              if (allocated(ibffr)) deallocate(ibffr)
              allocate(ibffr(size)) ; alloc_ibffr=size
              write(*,*) 'allocated nf_int workspace',size
            elseif (vartype == nf_float .and. size>alloc_bffr4) then
              if (allocated(bffr4)) deallocate(bffr4)
              allocate(bffr4(size)) ; alloc_bffr4=size
              write(*,*) 'allocated nf_real workspace',size
            elseif (vartype==nf_double .and. size>alloc_bffr8) then
              if (allocated(bffr8)) deallocate(bffr8)
              allocate(bffr8(size)) ; alloc_bffr8=size
              write(*,*) 'allocated nf_double workspace',size
            endif

            ierr=nf_inq_varid(nctarg, varname(1:lvar), vidtrg)
            if (ierr == nf_noerr) then
              if (vartype == nf_char) then
                ierr=nf_get_var_text  (ncsrc, vid, cbfr)
              elseif (vartype == nf_byte) then
                ierr=nf_get_var_int1  (ncsrc, vid, ibfr1)
              elseif (vartype == nf_short) then
                ierr=nf_get_var_int2  (ncsrc, vid, ibfr2)
              elseif (vartype == nf_int) then
                ierr=nf_get_var_int   (ncsrc, vid, ibffr)
              elseif (vartype == nf_float) then
                ierr=nf_get_var_real  (ncsrc, vid, bffr4)
              elseif (vartype == nf_double) then
                ierr=nf_get_var_double(ncsrc, vid, bffr8)
              else
                write(*,'(/8x,4A/)') '### ERROR: Unrecognized type ',
     &                        'of variable ''', varname(1:lvar), '''.'
                ierr=nf_noerr+1
              endif
              if (ierr == nf_noerr) then
#ifdef TIMING
                nclk=3-nclk
                call system_clock(iclk(nclk), clk_rate,clk_max)
                inc_clk=iclk(nclk)-iclk(3-nclk)
                if (inc_clk < 0) inc_clk=inc_clk+clk_max
                net_clk=net_clk+inc_clk
                net_read_clk=net_read_clk+inc_clk

                if (vartype == nf_char .or. vartype == nf_byte) then
                  net_read_size=net_read_size+size
                elseif (vartype == nf_short) then
                  net_read_size=net_read_size+2*size
                elseif (vartype==nf_int .or. vartype==nf_float) then
                  net_read_size=net_read_size+4*size
                elseif (vartype == nf_double) then
                  net_read_size=net_read_size+8*size
                endif
#endif
                if (vartype == nf_char) then
                  ierr=nf_put_var_text  (nctarg, vidtrg, cbfr)
                elseif (vartype == nf_byte) then
                  ierr=nf_put_var_int1  (nctarg, vidtrg, ibfr1)
                elseif (vartype == nf_short) then
                  ierr=nf_put_var_int2  (nctarg, vidtrg, ibfr2)
                elseif (vartype == nf_int) then
                  ierr=nf_put_var_int   (nctarg, vidtrg, ibffr)
                elseif (vartype == nf_float) then
                  ierr=nf_put_var_real  (nctarg, vidtrg, bffr4)
                elseif (vartype == nf_double) then
                  ierr=nf_put_var_double(nctarg, vidtrg, bffr8)
                endif
                if (ierr == nf_noerr) then
                  write(*,'(1x,5A)')   'wrote timeless variable ''',
     &                                varname(1:lvar), ''' into ''',
     &                                      trg_fname(1:ltrg), '''.'
#ifdef TIMING
                  nclk=3-nclk
                  call system_clock(iclk(nclk), clk_rate,clk_max)
                  inc_clk=iclk(nclk)-iclk(3-nclk)
                  if (inc_clk < 0) inc_clk=inc_clk+clk_max
                  net_clk=net_clk+inc_clk
                  net_wrt_clk=net_wrt_clk+inc_clk

                  if (vartype == nf_char .or. vartype == nf_byte) then
                    net_wrt_size=net_wrt_size+size
                  elseif (vartype == nf_short) then
                    net_wrt_size=net_wrt_size+2*size
                  elseif (vartype==nf_int .or. vartype==nf_float) then
                    net_wrt_size=net_wrt_size+4*size
                  elseif (vartype == nf_double) then
                    net_wrt_size=net_wrt_size+8*size
                  endif
#endif
                else
                  write(*,'(/1x,6A/12x,A/)')      '### ERROR: Cannot ',
     &              'write variable ''', varname(1:lvar), ''' into ''',
     &                     trg_fname(1:ltrg), '''.', nf_strerror(ierr)
                endif
              else
                write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot read ',
     &                    'variable ''', varname(1:lvar), ''' from ''',
     &                     src_fname(1:lsrc), '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(/1x,6A/12x,A/)')     '### ERROR: Cannot find ',
     &               'variable ID for ''', varname(1:lvar), ''' in ''',
     &                     trg_fname(1:ltrg), '''.', nf_strerror(ierr)
            endif
          endif !<-- 3-way switch, time-var, time-dependent, timeless
        else
          write(*,'(/1x,2A,I3/12x,4A/12x,A/)')    '### ERROR: Cannot ',
     &              'make general inquiry for variable # ', vid, 'in ',
     &                      'netCDF file ''', src_fname(1:lsrc), '''.',
     &                                               nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) goto 98                      !--> BREXIT
      enddo  !<-- vid=1,nvars

  98  i=nf_close(nctarg)
      if (ierr == nf_noerr) then
        write(*,'(4x,3A)') 'Closed file ''', trg_fname(1:ltrg),
     &                                                           '''.'
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot close netCDF ',
     &         'file ''', trg_fname(1:ltrg), '''.', nf_strerror(ierr)
      endif
      i=nf_close(ncsrc)
      if (ierr == nf_noerr) then
        write(*,'(4x,3A)')  'Closed file ''', src_fname(1:lsrc),
     &                                                           '''.'
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot close netCDF ',
     &         'file ''', src_fname(1:lsrc), '''.', nf_strerror(ierr)
      endif

#ifdef TIMING
      call cpu_time(run_time) ; run_time=run_time-tstart
      write(*,'(/2x,A,F11.2,1x,A/)') 'cpu time =', run_time, 'sec'

      nclk=3-nclk
      call system_clock(iclk(nclk), clk_rate, clk_max)
      inc_clk=iclk(nclk)-iclk(3-nclk)
      if (inc_clk < 0) inc_clk=inc_clk+clk_max
      net_clk=net_clk+inc_clk
      net_sync_clk=net_sync_clk+inc_clk

      if (clk_rate > 0) then
        ReadSize=1.0D-6*net_read_size
        WrtSize=1.0D-6*net_wrt_size
        ReadTime=net_read_clk/dble(clk_rate)
        CompTime=net_comp_clk/dble(clk_rate)
        WrtTime = net_wrt_clk/dble(clk_rate)
        SyncTime=net_sync_clk/dble(clk_rate)

        write(*,'(6x,A,F12.2,1x,A,F12.2,1x,A,F8.2,1x,A)')
     &           'Total data read :', ReadSize, 'MBytes in', ReadTime,
     &                           'sec (', ReadSize/ReadTime, 'MB/sec)'
        write(*,'(14x,A,F12.2,1x,A,F12.2,1x,A,F8.2,1x,A)') 'written :',
     &          WrtSize, 'MBytes in',  WrtTime, 'sec (',
     &                                      WrtSize/WrtTime, 'MB/sec)'
        write(*,'(12x,A,22x,F12.2,1x,A)') 'Computing :',CompTime,'sec'
        write(*,'(6x,A,22x,F12.2,1x,A)') 'File operations :', SyncTime,
     &                                                           'sec'
        write(*,'(44x,A/10x,A,12x,F12.2,1x,A/)')  '------------------',
     &  'Elapsed wall-clock time', dble(net_clk)/dble(clk_rate), 'sec'
      endif
#endif
      end
