#ifdef BAROTR_PV
      program bpvort
#else
      program vort
#endif

! Compute relative vorticity, barotropic potential vorticity, or wind
! stress curl. Takes a sequence of ROMS-style netCDF files, reads them
! sequentially [only variables "ocean_time", "u","v" (or "ubar", "vbar"
! if "u","v" are not available, or "sustr","svstr", or wind components
! "uwnd" or "vwnd", etc), and, in the case of PV also "zeta" for each
! snapshot, computes vorticity or PV, and puts the outcome into netCDF
! file called "vort.nc" or "bpvort.nc". U,V-components can be located
! at either U- and V-points on a C-grid, or co-located at RHO-points
! (as it is often the case for wind forcing files).   The type of grid
! staggering is identified by the dimension names of U,V-components and
! the appropriate algorithm is accordingly selected.  The first file in
! the sequence is expected to contain relevant environmental variables
! "pm", "pn", "h", "f", and (if available) "mask_rho".

# define VERBOSE

      implicit none
      integer iarg, nargs, arg_to_start, ncid, ndims, nvars, natts,
     &        vid_t_in, vid_u,vid_v, vid_vert_coord, max_recs, irec,
     &        nctarg, numdims, vid_t_out,vid_vrt, unlm_id_out, ierr,
     &        irec_out, Lm,Mm,N, size, vid, vtype, i,j, dimids(8),
     &                   dimsize(8), udim(8), vdim(8),  q2dgrd(8),
     &                   lsrc, ltrg, ltim, lvar, ldm, iargc, lenstr
      character(len=128) fname, trg_fname
      character(len=16) vert_coord_name, time_name, u_name,v_name,
     &                       vort_name, vname, dname, dimname(8)
      logical normalize, fnd_rmask, fnd_u_spv, non_stagg,
     &                             cgrd_stagg, new_dim

      character(len=128), allocatable, dimension(:) :: args
      real(kind=4), allocatable, dimension(:) :: vert_coord_var
      real(kind=4), allocatable, dimension(:,:) :: rmask, pm,pn,
     &                                  f, f_q, dm_u,dn_v, iA_q
#ifdef BAROTR_PV
     &                                                , h, zeta
      real(kind=4) hmdv, fmdv
#endif
      real(kind=8) time
      real(kind=4), allocatable, dimension(:,:,:) :: u,v,q
      real(kind=4) qnorm, spv
      real(kind=4), parameter :: FillValue=1.E+33

#include "netcdf.inc"
                                       ! Reset everything:
      spv=FillValue     ;  qnorm=0.    !------ --------------
      fnd_rmask=.false. ;  irec_out=0  ! horizontal and vertical
      fnd_u_spv=.false. ;  Lm=0 ; N=0  ! dimensions, flags to signal
      numdims=0         ;  Mm=0        ! presence of land mask or
      do i=1,8                         ! special values for input
        dimname(i)='              '    ! velocity components; counter
        dimsize(i)=0                   ! for output records; number of
        udim(i)=0                      ! dimensions beyond XI-, ETA-
        vdim(i)=0                      ! in the target file, and their
      enddo                            ! names, and sizes;

      time_name='none'                 ! Initialize input variable 
#ifdef BAROTR_PV
      u_name='ubar'    ; v_name='vbar' ! names to their first-guess 
      vort_name='bar_pv'               ! default values: if variables 
      trg_fname='bpvort.nc'            ! with these names will not be 
#else
      u_name='u'       ; v_name='v'    ! found in netCDF files, the
      vort_name='rvort'                ! names set here will be 
      trg_fname='vort.nc'              ! changed during second 
#endif
      normalize=.false.                ! and third attempts.

                                       ! Sort out number of arguments
      nargs=iargc()                    ! given in command line and save
      if (nargs > 0) then              ! the actual arguments (names of
        allocate(args(nargs))          ! netCDF files) into array.
        do iarg=1,nargs
          call getarg(iarg,args(iarg)) ; lsrc=lenstr(args(iarg))
        enddo

        iarg=1 ; arg_to_start=1
        do while(args(iarg)(1:1) == '-' .and. iarg < nargs)
          lsrc=lenstr(args(iarg))
          if (lsrc == 2 .and. args(iarg)(1:lsrc) == '-n' .or.
     &        lsrc == 6 .and. args(iarg)(1:lsrc) == '--norm') then
            normalize=.true. ;  arg_to_start=iarg+1
#ifndef BAROTR_PV
          elseif (lsrc > 9 .and. args(iarg)(1:7) == '--vars=') then
            i=8
            write(*,*) args(iarg)(1:lsrc)
            do while(args(iarg)(i:i) /= ',' .and. i < lsrc) 
              i=i+1
            enddo                         ! manually specify netCDF
            if (i < lsrc) then            ! variable names for vector
              u_name=args(iarg)(8:i-1)    ! components
              v_name=args(iarg)(i+1:lsrc)
              arg_to_start=iarg+1
              write(*,'(1x,A,2(2x,3A))')  'manually specified ',
     &              'u_name=''', u_name(1:lenstr(u_name)), '''',
     &              'v_name=''', v_name(1:lenstr(v_name)), ''''
            else
              write(*,'(/1x,4A/)') '### ERROR: Two vector components ',
     &         'must be specified in --vars=', args(iarg)(8:lsrc), '.'
              stop
            endif
#endif
          else
            write(*,'(/2A/)') '### ERROR: The optional argument ',
     &                           'should be either -n or --norm.'
            stop
          endif
          iarg=iarg+1
        enddo

c**     write(*,*) 'arg_to_start =', arg_to_start
      else
        write(*,'(/1x,A/ /12x,2A/1x,A/12x,2A/)') 'Usage should be: ',
#ifdef BAROTR_PV
     &          'bpvort [--norm] ',
#else
     &          'vort [--norm] [--vars=u_name,v_name] ',
#endif
     &          '[grid_file.nc] file1.nc [file2.nc ...]',
     &          'or',
#ifdef BAROTR_PV
     &          'bpvort [-n] ',
#else
     &          'vort [-n] [--vars=u_name,v_name] ',
#endif
     &          '[grid_file.nc] file1.nc [file2.nc ...]'
        stop
      endif

! Since the first file in the sequence may be either a pure grid file
! or file containing both grid data and vector components, and because
! the names of the components are not known in advance, open the second
! file in the sequence (unless there is only one file name given) and
! verify that all necessary variables are available. If not found,
! make an attempt to find alternatives: first try barotropic velocity
! components to compute 2D relative vorticity, if still not, try wind
! stress component to compute wind stress curl.  If these secondary
! attempts are successful, adjust the names of the target variable
! and output file name.  Once the suitable pair of variables is found,
! find dimensions, check and retrieve special values (if any).  Based
! on this, set dimensions for the target netCDF file.

      fname=args(min(arg_to_start+1,nargs)) ; lsrc=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        write(*,*) 'Examining ''', fname(1:lsrc), '''...'
        ierr=nf_inq_varid(ncid, u_name, vid_u)
#ifndef BAROTR_PV
        if (ierr /= nf_noerr) then                ! Because BAROTR_PV
          u_name='ubar' ; v_name='vbar'           ! needs strictly ubar
          ierr=nf_inq_varid(ncid, u_name, vid_u)  ! and vbar, there is
          if (ierr /= nf_noerr) then              ! no attempt to check
                                                  ! for alternative
            u_name='sustr' ; v_name='svstr'       ! pairs of variables.
            vort_name='curl'
            trg_fname='curl.nc'
            ierr=nf_inq_varid(ncid, u_name, vid_u)
            if (ierr /= nf_noerr) then
              u_name='uwnd'    ;  v_name='vwnd'
              vort_name='curl' ; trg_fname='curl.nc'
              ierr=nf_inq_varid(ncid, u_name, vid_u)
              if (ierr /= nf_noerr) then
                write(*,'(/2(1x,A)/)')   '### ERROR :: No suitable ',
     &          'pair of vector components is found in netCDF file.'
                stop
              endif
            endif
          endif
        endif
#endif
        ierr=nf_inq_nvars(ncid, nvars)
        if (ierr == nf_noerr) then
          do vid=1,nvars
            ierr=nf_inq_varname(ncid, vid, vname)
            if (ierr == nf_noerr) then
              lvar=lenstr(vname)
              if (lvar > 3) then
                if ( vname(1:4)  == 'time' .or.
     &               vname(lvar-3:lvar) == 'time' ) then
                  time_name=vname(1:lvar)
                  write(*,*) 'identified time variable ''',
     &                                vname(1:lvar), '''.'
                endif
              endif
            else
              write(*,'(/1x,2A,I4,1x,3A/12x,A)') '### ERROR: Cannot ',
     &                   'determine name of variable #', vid, 'in ''',
     &                        fname(1:lsrc), '''.', nf_strerror(ierr)
              stop
            endif
          enddo
        else
          write(*,'(/1x,4A/12x,A)')    '### ERROR: Cannot determine ',
     &                     'number of variables in ''', fname(1:lsrc),
     &                                       '''.', nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) stop

! At this moment variable ID "vid_u" is expected to be known.
! Inspect U-component: check whether it has special value for land
! masked areas, save names and sizes of its dimensions -- these will
! be used to determine (i) type of grid staggering (C- vs. B-grid),
! and to select only relevant dimensions for the output file.

        ierr=nf_inq_var(ncid, vid_u, vname, vtype,  ndims,
     &                                      dimids, natts)
        if (ierr == nf_noerr) then
          numdims=ndims ; lvar=lenstr(vname)
          ierr=nf_get_att_real(ncid, vid_u, '_FillValue', spv)
          if (ierr == nf_noerr) then
            write(*,*) 'retrieved _FillValue =', spv, ' for ',
     &                     'variable ''', vname(1:lvar), '''.'
            fnd_u_spv=.true.
          else
            write(*,*)      'no _FillValue is specified for ',
     &                     'variable ''', vname(1:lvar), '''.'
            ierr=nf_noerr
          endif
          do i=1,numdims
            ierr=nf_inq_dim(ncid, dimids(i), dname, size)
            if (ierr == nf_noerr) then
              ldm=lenstr(dname)
              write(*,'(8x,3A,I5)') 'U-comp ',dname(1:ldm),' =',size
              udim(i)=i
              dimname(i)=dname
              dimsize(i)=size
            else
              write(*,'(/1x,2A,I3,1x,3A/12x,A/)')      '### ERROR: ',
     &              'Cannot make inquiry for dimension #', dimids(i),
     &              'in ''', fname(1:lsrc), '''.', nf_strerror(ierr)
              stop
            endif
          enddo
        else
          write(*,'(/1x,2A,I4,1x,5A/12x,A/)')    '### ERROR: Cannot ',
     &       'make general inquiry for variable #', vid_u, 'named ''',
     &            u_name(1:lenstr(u_name)), ''' in ''', fname(1:lsrc),
     &                                       '''.', nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) stop

! Inspect V-component and add its dimensions to the list initialized
! above: because U- and V-component are expected to have some (perhaps
! even all) of their dimensions be the same, expand the list only if
! the dimension of V-component does not match any of the ones already
! in the list.

        ierr=nf_inq_varid(ncid, v_name, vid_v)
        if (ierr == nf_noerr) then
          ierr=nf_inq_var (ncid, vid_v, vname, vtype,  ndims,
     &                                         dimids, natts)
          lvar=lenstr(vname)
          if (ierr == nf_noerr) then
            do i=1,ndims
              ierr=nf_inq_dim(ncid, dimids(i), dname, size)
              if (ierr == nf_noerr) then
                ldm=lenstr(dname)
                write(*,'(8x,3A,I5)') 'V-comp ',dname(1:ldm),' =',size
                new_dim=.true.
                do j=1,numdims
                  if (dimname(j) == dname) then
                    new_dim=.false. ; vdim(i)=j
                  endif
                enddo
                if (new_dim) then
                  numdims=numdims+1
                  dimname(numdims)=dname
                  dimsize(numdims)=size
                  vdim(i)=numdims
                endif
              else
                write(*,'(/1x,2A,I3,1x,3A/12x,A/)')    '### ERROR: ',
     &              'Cannot make inquiry for dimension #', dimids(i),
     &              'in ''', fname(1:lsrc), '''.', nf_strerror(ierr)
                stop
              endif
            enddo
          else
            write(*,'(/1x,2A,I4,1x,5A/12x,A/)')  '### ERROR: Cannot ',
     &       'make general inquiry for variable #', vid_v, 'named ''',
     &            v_name(1:lenstr(v_name)), ''' in ''', fname(1:lsrc),
     &                                       '''.', nf_strerror(ierr)
          endif
        else
          write(*,'(/1x,6A/12x,A/)')       '### ERROR: Cannot find ',
     &           'variable ''', v_name(1:lenstr(v_name)), ''' in ''',
     &                      fname(1:lsrc), '''.',  nf_strerror(ierr)
        endif
        if (ierr /= nf_noerr) stop

! Sort out dimensions: horizontal sizes should be consistent between
!----- --- ----------  U- and V-components in a specific way depending 
! on grid staggering: C- or non-staggered  Vertical dimension (if any)
! and time (if exists) should be the same (by dimension ID) for both
! components; Verify,complain about inconsistency and stop, if any of
! the tests fails.

        cgrd_stagg=.false. ; non_stagg=.false.
        if ( dimsize(udim(1))+1 == dimsize(vdim(1)) .and.
     &       dimsize(udim(2))   == dimsize(vdim(2))+1 ) then
          cgrd_stagg=.true.
          write(*,*) 'identified C-grid staggered U,V components'
        elseif ( dimsize(udim(1)) == dimsize(vdim(1)) .and.
     &           dimsize(udim(2)) == dimsize(vdim(2)) ) then
          non_stagg=.true.
          write(*,*) 'identified non-staggered U,V components' 
        endif

        if (cgrd_stagg .or. non_stagg) then
          Lm=dimsize(vdim(1))-2 ; Mm=dimsize(udim(2))-2
        else
          write(*,'(/1x,2A,2(/14x,2(2x,2A,I5))/)')     '### ERROR: ',
     &   'Horizontal dimensions of U- and V-components do not match',
     &                   dimname(udim(1)), ' =', dimsize(udim(1)),
     &                   dimname(udim(2)), ' =', dimsize(udim(2)),
     &                   dimname(vdim(1)), ' =', dimsize(vdim(1)),
     &                   dimname(vdim(2)), ' =', dimsize(vdim(2))
          stop
        endif

        if (udim(3) > 0) then
          if (udim(3) == vdim(3)) then
            N=dimsize(udim(3)) ;  ldm=lenstr(dimname(udim(3)))
            if (udim(4) > 0) then
              if (udim(4) /= vdim(4)) then
                write(*,'(/1x,3A,/14x,2(2x,2A,I5)/)')  '### ERROR: ',
     &               'The fourth dimension must be the same for U- ',
     &                                           'and V-components:',
     &                  dimname(udim(4)), ' =', dimsize(udim(4)),
     &                  dimname(vdim(4)), ' =', dimsize(vdim(4))
                stop
              endif
            else      !--> 3rd-dim is either vertical or time

! Since vertical dimension may or may not exist, and since its name is
! not known in advance, assume that it is the next after ETA-dimension,
! but is not the time dimension identified here by presence of suffix
! or prefix "time" within its name.

              if (ldm > 3) then
                if ( dimname(udim(3))(1:4)  == 'time' .or.
     &               dimname(udim(3))(ldm-3:ldm) == 'time' ) then
                  N=1  !<-- cancel N as vertical coordinate
                endif
              endif
            endif

! If N > 1 survived the tests above and udim(3) is therefore identified
! as vertical dimension, then there is a chance of file having vertical
! coordinate named the same way (compliant with "CF-conventions") so
! attempt to read it: it may or may not succeed. It found, then it will
! will be copied into output vorticity file as is; if not just skip it.
! Allocation status of "vert_coord_var" will be used as logical flag
! for its presence.

            if (N > 1) then
              ierr=nf_inq_varid(ncid, dimname(udim(3))(1:ldm), vid)
              if (ierr == nf_noerr) then
                allocate(vert_coord_var(N))
                ierr=nf_get_var_real(ncid, vid, vert_coord_var)
                if (ierr == nf_noerr) then
                  vert_coord_name=dimname(udim(3))(1:ldm)
                  write(*,*) 'read vertical coordinate variable ''',
     &                               dimname(udim(3))(1:ldm), '''.'
                else
                  deallocate(vert_coord_var)
                endif
              endif
            endif
          else
            write(*,'(/1x,2A,/14x,2(2x,2A,I5)/)')   '### ERROR: The ',
     &     'third dimension must be the same for U- and V-components',
     &                  dimname(udim(3)), ' =', dimsize(udim(3)),
     &                  dimname(vdim(3)), ' =', dimsize(vdim(3))
            stop
          endif
        else
          N=1  !<-- neither vertical, nor time dimension is present
        endif
        ierr=nf_close(ncid)
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot open  netCDF ',
     &  'file ''',fname(1:lsrc),''' for reading.', nf_strerror(ierr)
      endif

#ifdef VERBOSE
      write(*,'(1x,A,2(3x,A,I5),2(3x,A,I3))')  'found dimensions:',
     &          'Lm =',Lm, 'Mm =',Mm, 'N =',N, 'numdims=',numdims
#endif
      if (u_name == 'ubar') then
        N=1
#ifdef VERBOSE
        write(*,'(3x,2A)') 'Changed N to 1 because only barotropic ',
     &                                  ' velocities are available.'
#endif
      endif

! Open the first file in the sequence (it should be either grid
! file or history containing grid data), determine size of arrays,
! allocate them, read geometric data, compute metric coefficients.

      fname=args(arg_to_start) ; lsrc=lenstr(fname)
      ierr=nf_open(fname, nf_nowrite, ncid)
      if (ierr == nf_noerr) then
        write(*,*) 'reading grid data from ''', fname(1:lsrc), '''.'
        if (N > 1) then
          call roms_check_dims(ncid, fname, Lm,Mm,N)
        else
          call roms_check_dims(ncid, fname, Lm,Mm,0)
        endif
        allocate( pm(0:Lm+1,0:Mm+1), dm_u(1:Lm+1,0:Mm+1),
     &            pn(0:Lm+1,0:Mm+1),  dn_v(0:Lm+1,1:Mm+1),
     &          iA_q(1:Lm+1,1:Mm+1), rmask(0:Lm+1,0:Mm+1),
#ifdef BAROTR_PV
     &              h(0:Lm+1,0:Mm+1), zeta(0:Lm+1,0:Mm+1),
#endif
     &             f_q(1:Lm+1,1:Mm+1), f(0:Lm+1,0:Mm+1))

        call get_var_by_name_real(ncid, 'pm', pm)
        call get_var_by_name_real(ncid, 'pn', pn)
        call get_var_by_name_real(ncid,  'f',  f)

#ifdef BAROTR_PV
        call get_var_by_name_real(ncid, 'h', h)
        if (normalize) then
          if (mod(Mm,2) == 0) then
            hmdv=0.5*(h(2,Mm/2)+h(2,Mm/2+1))
            fmdv=0.5*(f(2,Mm/2)+f(2,Mm/2+1))
          else
            hmdv=h(2,(Mm+1)/2)
            fmdv=f(2,(Mm+1)/2)
          endif
          write(*,*) 'h_at_middle =',hmdv,'  f_at_middle =', fmdv
          qnorm=hmdv/fmdv
        endif
#endif

        ierr=nf_inq_varid(ncid, 'mask_rho', i)
        if (ierr == nf_noerr) then
          ierr=nf_get_var_real(ncid, i, rmask)
          if (ierr == nf_noerr) then
            fnd_rmask=.true.
            write(*,'(8x,A)') 'retrieved ''rmask'''
          endif
        else
          write(*,'(/1x,4A/14x,2A/)')    '### WARNING: Cannot find ',
     &            'land mask from file ''',   fname(1:lsrc),   '''.',
     &            'Assuming rmask=1 everywhere. ', nf_strerror(ierr)
          ierr=nf_noerr
        endif

! Finally, check whether the first file is a pure grid file or a file
! containing both grid data and dynamic variables: advance the starting
! argument first, then retract it back if "u,v"-vector components are
! present:

        arg_to_start=arg_to_start+1
        ierr=nf_inq_varid(ncid, u_name, vid_u)
        if (ierr == nf_noerr) then
          ierr=nf_inq_varid(ncid, v_name, vid_v)
          if (ierr == nf_noerr) then
            arg_to_start=arg_to_start-1
          endif
        endif
        ierr=nf_close(ncid)

C$OMP PARALLEL SHARED(Lm,Mm, pm,pn, f,f_q, fnd_rmask,rmask, dm_u,dn_v,
C$OMP&                                                           iA_q)
        call init_vort_thread(Lm,Mm, pm,pn, f,f_q, fnd_rmask,rmask,
     &                                                dm_u,dn_v, iA_q)
C$OMP END PARALLEL
        deallocate(f)
      else
        write(*,'(/1x,4A/12x,A/)') '### ERROR: Cannot open  netCDF ',
     &  'file ''',fname(1:lsrc),''' for reading.', nf_strerror(ierr)
      endif

! Process the entire sequence:
!-------- --- ------ ---------

      do iarg=arg_to_start,nargs
        fname=args(iarg) ; lsrc=lenstr(fname)
        ierr=nf_open(fname, nf_nowrite, ncid)
        if (ierr == nf_noerr) then
          if (N > 1) then
            call roms_check_dims(ncid, fname, Lm,Mm,N)
          else
            call roms_check_dims(ncid, fname, Lm,Mm,0)
          endif
          write(*,'(1x,A,I4,1x,3A)',advance='no') 'iarg =', iarg,
     &                       'Processing ''', fname(1:lsrc), ''','

          if (.not.allocated(q)) then  !<-- do it only once
            allocate(q(1:Lm+1,1:Mm+1,N))
            if (cgrd_stagg) then
              allocate(u(1:Lm+1,0:Mm+1,N), v(0:Lm+1,1:Mm+1,N))
            elseif(non_stagg) then
              allocate(u(0:Lm+1,0:Mm+1,N), v(0:Lm+1,0:Mm+1,N))
            endif

! Create output netCDF file: note that dimension names are inherited
!------- ------ ------ ----- from the input file rather than specified
! explicitly; also note that in both cases - C-grid or non-staggered -
! vorticity variable inherits its first dimension from first dimension
! of U- and the second from the second of V-component.  The sizes of
! vorticity dimensions are the same for C- or non-staggered grid, and
! are one less than the first dimension of V- and the second of
! U-component respectively. 

            ltrg=lenstr(trg_fname)
            ierr=nf_create(trg_fname, nf_64bit_offset, nctarg)
            if (ierr == nf_noerr) then
              ldm=lenstr(dimname(udim(1)))
              ierr=nf_def_dim(nctarg, dimname(udim(1))(1:ldm), Lm+1,
     &                                                    q2dgrd(1))
              ldm=lenstr(dimname(vdim(2)))
              ierr=nf_def_dim(nctarg, dimname(vdim(2))(1:ldm), Mm+1,
     &                                                    q2dgrd(2))
              ndims=2
              if (udim(3) > 0) then
                ndims=3 ; ldm=lenstr(dimname(udim(3)))
                if (N > 1) then
                  ierr=nf_def_dim(nctarg, dimname(udim(3))(1:ldm), N,
     &                                                    q2dgrd(3))
                  if (udim(4) > 0) then
                    ndims=4 ; ldm=lenstr(dimname(udim(4)))
                    ierr=nf_def_dim(nctarg, dimname(udim(4))(1:ldm),
     &                                      nf_unlimited, q2dgrd(4))
                    unlm_id_out=q2dgrd(4)
                  endif
                else
                  ierr=nf_def_dim(nctarg, dimname(udim(3))(1:ldm),
     &                                   nf_unlimited, q2dgrd(3))
                  unlm_id_out=q2dgrd(3)
                endif
              endif

! Create vertical coordinate variable (only if found in input file):

              if (allocated(vert_coord_var)) then
                ierr=nf_def_var(nctarg, vert_coord_name, nf_real,
     &                              1, q2dgrd(3), vid_vert_coord)
              endif

! Create time variable and copy its attributes from the original file:

              if (time_name(1:4) /= 'none' .and. unlm_id_out > 0) then
                ierr=nf_def_var(nctarg, time_name, nf_real, 1,
     &                                 unlm_id_out, vid_t_out)
                ierr=nf_inq_varid(ncid, time_name, vid_t_in)
                if (ierr == nf_noerr) then
                  ierr=nf_copy_att(ncid, vid_t_in, 'units',
     &                                        nctarg, vid_t_out)
                  if (ierr == nf_noerr) then
                    write(*,*)  'copied attribute ''units'' for ',
     &                                                 time_name
                  else
                    write(*,*) 'cannot find attribute ''units'' ',
     &                            'for time variable --> skipping'
                  endif
                else
                  write(*,*) 'cannot find time variable'
                endif
              endif

! Create vorticity variable:

              ierr=nf_def_var(nctarg, vort_name, nf_real, ndims,
     &                                          q2dgrd, vid_vrt)
#ifdef BAROTR_PV
              if (normalize) then
                ierr=nf_put_att_text(nctarg, vid_vrt, 'long_name', 52,
     &         'Normalized (centered) barotropic potential vorticity')
              else
                ierr=nf_put_att_text(nctarg, vid_vrt, 'long_name', 30,
     &                               'Barotropic potential vorticity')
              endif
#else
              if (normalize) then
                ierr=nf_put_att_text(nctarg, vid_vrt, 'long_name', 57,
     &    'Relative vorticity normalized by local Coriolis parameter')
              else
                ierr=nf_put_att_text(nctarg, vid_vrt, 'long_name', 18,
     &                                           'Relative vorticity')
              endif
#endif
              ierr=nf_put_att_real(nctarg, vid_vrt, '_FillValue',
     &                                    nf_real, 1, FillValue)
              ierr=nf_enddef(nctarg)
              if (allocated(vert_coord_var)) then
                call put_var_by_name_real(nctarg, vert_coord_name,
     &                                            vert_coord_var)
              endif
            else
              write(*,'(/1x,4A/12x,A)')  '### ERROR: Cannot create ',
     &                           'netCDF file ''', trg_fname(1:ltrg),
     &                                      '''.', nf_strerror(ierr)
              stop
            endif
          endif  !<-- .not.allocated(q)


!      *****    *********    ******   *******    *********
!    ***   ***  *  ***  *   *   ***   ***   ***  *  ***  *
!    ***           ***     **   ***   ***   ***     ***
!      *****       ***    ***   ***   ***   **      ***
!          ***     ***    *********   ******        ***
!    ***   ***     ***    ***   ***   ***  **       ***
!      *****       ***    ***   ***   ***   ***     ***


! Note: The number of records to be processed is determined by the
! length of dimension of the timing variable.  However, in some files
! (notably initial conditions) the timing variable may not even have
! dimension if the file contains is just a single snapshot. If the
! actual variable in the file is a scalar, reading it by
! nf_get_vara(.., start=irec,count=1,..) returns NO_ERROR retrieving
! its only scalar value regardless of the value of argument "start"
! (i.e., nf_get_vara behaves like nf_get_var in this case). Similarly,
! if "u" and "v" in the file do not have time dimension, an attempt to
! retrieve them by nf_get_vara(...,start,count,...) with start at the
! index corresponding to would be time dimension set to non-existing
! record nevertheless returns NO_ERROR status because netCDF ignores
! the content of start,count for indices beyond the actual number of
! dimensions of the variable in the file.  This behavior causes an
! algorithm relying solely on non-zero error status to detect when
! running out of records (like the previous version of this algorithm)
! end up executing endless loop if there is no time dimension (reading
! the same data again with NO_ERROR status, despite the fact that
! record number "irec" is incremented at every iteration of the loop).
! To address this, the following code is designed to positively
! identify the number of records "max_recs", assigning it to one, in
! the case if no time dimension is found.

          if (time_name(1:4) /= 'none') then
            ierr=nf_inq_varid(ncid, time_name, vid_t_in)
            if (ierr == nf_noerr) then
              ierr=nf_inq_var(ncid, vid_t_in, vname, vtype,  ndims,
     &                                               dimids, natts)
              if (ierr == nf_noerr) then
                ltim=lenstr(vname)
                max_recs=0
                if (ndims == 0) then
                  max_recs=1   ! <-- no records, just a single snapshot
                elseif (ndims == 1) then
                  ierr=nf_inq_dimlen(ncid, dimids, max_recs)
                  if (ierr /= nf_noerr) then
                    write(*,'(/1x,5A/)')     '### ERROR: Cannot find ',
     &                        'number of records for time variable ''',
     &                 vname(1:ltim), ''' in ''', fname(1:lsrc), '''.'
                  endif
                else
                  write(*,'(/1x,5A/)')   '### ERROR: Time variable ''',
     &                    vname(1:ltim), ''' in ''', fname(1:lsrc),
     &                               ''' has more than one dimension.'
                endif
              else
                write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot make ',
     &                        'general inquiry about time variable ''',
     &                        vname(1:ltim), ''' in ''', fname(1:lsrc),
     &                                        '''.', nf_strerror(ierr)
              endif
            else
              write(*,'(1x,6A/12x,A/)')     '### ERROR: Cannot find ',
     &                  'variable ''', time_name(1:lenstr(time_name)),
     &      ''' in file ''', fname(1:lsrc), '''.', nf_strerror(ierr)
            endif
          else
            max_recs=1
          endif
          if (ierr /= nf_noerr) stop

          do irec=1,max_recs
            if (time_name(1:4) /= 'none') then
              ierr=nf_get_vara_double(ncid, vid_t_in, irec, 1, time)
              if (ierr == nf_noerr) then
                write(*,*) time_name(1:ltim),' =',time,'  rec =',irec

              else
                write(*,'(/1x,6A/12x,A/)')   '### ERROR: Cannot read ',
     &              'time variable ''', time_name(1:lenstr(time_name)),
     &            ''' in ''', fname(1:lsrc), '''.', nf_strerror(ierr)
              endif
            endif

#ifdef BAROTR_PV
            call get_rec_by_name_real(ncid, fname, 'zeta', Lm+2,Mm+2,
     &                                                 0, irec, zeta)
#endif

!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
! MAIN COMPUTATIONAL PART: Note that the two logical branches below,
! "cgrd_stagg" vs. "non_stagg", differ only by (i) the dimensions of
! incoming "u,v",  and (ii) _C_ vs _A_ in the middle of the name of
! working subroutine.

            if (cgrd_stagg) then
              call get_rec_by_name_real(ncid, fname, u_name,
     &                                    Lm+1,Mm+2,N, irec, u)
              call get_rec_by_name_real(ncid, fname, v_name,
     &                                    Lm+2,Mm+1,N, irec, v)


              if (normalize) then
                if (fnd_rmask) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                          , rmask )

                  call comp_vort_CNmsk_thread(    Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                          , rmask, FillValue )
C$OMP END PARALLEL

                elseif (fnd_u_spv) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                                        , spv )

                  call comp_vort_CNspv_thread(    Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                               , spv, FillValue
     &                                                              )
C$OMP END PARALLEL

                else
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                                              )
                  call comp_vort_CN_thread(      Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                                              )
C$OMP END PARALLEL

                endif
              else
               if (fnd_rmask) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                          , rmask )

                  call comp_vort_Cmsk_thread(     Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                          , rmask, FillValue )
C$OMP END PARALLEL

                elseif (fnd_u_spv) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                                        , spv )

                  call comp_vort_Cspv_thread(     Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                               , spv, FillValue
     &                                                              )
C$OMP END PARALLEL

                else
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                                              )
                  call comp_vort_C_thread(       Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                                              )
C$OMP END PARALLEL
                endif
              endif
            elseif (non_stagg) then
              call get_rec_by_name_real(ncid, fname, u_name,
     &                                    Lm+2,Mm+2,N, irec, u)
              call get_rec_by_name_real(ncid, fname, v_name,
     &                                    Lm+2,Mm+2,N, irec, v)
              if (normalize) then
                if (fnd_rmask) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                          , rmask )

                  call comp_vort_ANmsk_thread(    Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                          , rmask, FillValue )
C$OMP END PARALLEL

                elseif (fnd_u_spv) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                                        , spv )

                  call comp_vort_ANspv_thread(    Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                               , spv, FillValue
     &                                                              )
C$OMP END PARALLEL


                else
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta, qnorm
#else
C$OMP&                                          , f_q
#endif
C$OMP&                                                              )
                  call comp_vort_AN_thread(      Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta, qnorm
#else
     &                                          , f_q
#endif
     &                                                              )
C$OMP END PARALLEL

                endif
              else
               if (fnd_rmask) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                          , rmask )

                  call comp_vort_Amsk_thread(     Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                          , rmask, FillValue )
C$OMP END PARALLEL

                elseif (fnd_u_spv) then
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                                        , spv )

                  call comp_vort_Aspv_thread(     Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                               , spv, FillValue
     &                                                              )
C$OMP END PARALLEL

                else
C$OMP PARALLEL SHARED(          Lm,Mm,N, dm_u,dn_v, iA_q, u,v, q
#if defined BAROTR_PV
C$OMP&                                          , f_q, h, zeta
#endif
C$OMP&                                                              )
                  call comp_vort_A_thread(       Lm,Mm,N, dm_u,dn_v,
     &                                                   iA_q, u,v, q
#if defined BAROTR_PV
     &                                          , f_q, h, zeta
#endif
     &                                                              )
C$OMP END PARALLEL
                endif
              endif   !<-- normalize
            endif   !<-- non-stagg vs. C-grid
!<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

            irec_out=irec_out+1
            if (time_name(1:4) /= 'none') then
              ierr=nf_put_vara_double(nctarg, vid_t_out, irec_out, 1,
     &                                                         time)
            endif
            call put_rec_by_name_real(nctarg, trg_fname, vort_name,
     &                                    Lm+1,Mm+1,N, irec_out, q)
            ierr=nf_sync(nctarg)
          enddo
          ierr=nf_close(ncid)
        else
          write(*,'(1x,3A/12x,A/)')  '### ERROR: Cannot open file ''',
     &            fname(1:lsrc), ''' for reading.', nf_strerror(ierr)
        endif
      enddo !<-- iarg
      ierr=nf_close(nctarg)
      stop
      end


      subroutine init_vort_thread(Lm,Mm, pm,pn, f,f_q, fnd_rmask,
     &                                   rmask,  dm_u,dn_v, iA_q)
      implicit none
      integer Lm,Mm
      real(kind=4), dimension(0:Lm+1,0:Mm+1) :: pm,pn,f, rmask
      real(kind=4), dimension(1:Lm+1,0:Mm+1) :: dm_u
      real(kind=4), dimension(0:Lm+1,1:Mm+1) :: dn_v
      real(kind=4), dimension(1:Lm+1,1:Mm+1) :: iA_q, f_q
      logical fnd_rmask

      integer numthreads,trd, range, tile, my_first,my_last,
     &                   nsub_x,nsub_y, istr,iend,jstr,jend
C$    integer omp_get_num_threads, omp_get_thread_num
      numthreads=1 ; trd=0
C$    numthreads=omp_get_num_threads() ; trd=omp_get_thread_num()
      call set_tiles(Lm,Mm, nsub_x,nsub_y)
      range=(nsub_x*nsub_y +numthreads-1)/numthreads
      my_first=trd*range
      my_last=min(my_first + range-1, nsub_x*nsub_y-1)

      do tile=my_first,my_last
        call comp_tile_bounds(tile, Lm,Mm, nsub_x,nsub_y,
     &                               istr,iend,jstr,jend)

        call init_vort_tile(istr,iend,jstr,jend, Lm,Mm, pm,pn,
     &              f,f_q, fnd_rmask, rmask, dm_u,dn_v, iA_q)
      enddo
      end


      subroutine init_vort_tile(istr,iend,jstr,jend, Lm,Mm, pm,pn,
     &                  f,f_q, fnd_rmask, rmask, dm_u,dn_v, iA_q)
      implicit none
      integer istr,iend,jstr,jend, Lm,Mm, i,j
      real(kind=4), dimension(0:Lm+1,0:Mm+1) :: pm,pn, f, rmask
      real(kind=4), dimension(1:Lm+1,0:Mm+1) :: dm_u
      real(kind=4), dimension(0:Lm+1,1:Mm+1) :: dn_v
      real(kind=4), dimension(1:Lm+1,1:Mm+1) :: iA_q, f_q
      logical fnd_rmask

#include "compute_extended_bounds.h"

      if (fnd_rmask) then
        do j=jstrR,jendR
          do i=istr,iendR
            dm_u(i,j)=2.*rmask(i,j)*rmask(i-1,j)/(pm(i,j)+pm(i-1,j))
          enddo
        enddo
        do j=jstr,jendR
          do i=istrR,iendR
            dn_v(i,j)=2.*rmask(i,j)*rmask(i,j-1)/(pn(i,j)+pn(i,j-1))
          enddo
        enddo
      else
        do j=jstrR,jendR
          do i=istrR,iendR
            rmask(i,j)=1.D0
          enddo
        enddo
        do j=jstrR,jendR
          do i=istr,iendR
            dm_u(i,j)=2./(pm(i,j)+pm(i-1,j))
          enddo
        enddo
        do j=jstr,jendR
          do i=istrR,iendR
            dn_v(i,j)=2./(pn(i,j)+pn(i,j-1))
          enddo
        enddo
      endif
      do j=jstr,jendR
        do i=istr,iendR
          iA_q(i,j)=0.0625*(pm(i,j)+pm(i-1,j)+pm(i,j-1)+pm(i-1,j-1))
     &                    *(pn(i,j)+pn(i,j-1)+pn(i-1,j)+pn(i-1,j-1))

          f_q(i,j)=0.25*(f(i,j)+f(i-1,j)+f(i,j-1)+f(i-1,j-1))

! In the case of computing normalized relative vorticity (however
! NOT barotropic PV) convert Coriolis parameter into normalization
! multiplier -- basically inverse of itself.

#ifndef BAROTR_PV
          f_q(i,j)=1./f_q(i,j)
#endif
        enddo
      enddo
      end

! The following part permutates 3 switches: (i) NON_STAGG (undef/def),
! (ii) NORMALIZE(undef/def), and (iii) MASK_RHO/SPV_UV/neither (3-way),
! resulting in overall 12 variants of vorticity computation, which is
! further multiplied by 2 due to BAROTR_PV switch defined externally.

#include "compute_vorticity.F"
#define MASK_RHO
#include "compute_vorticity.F"
#undef MASK_RHO
#define SPV_UV
#include "compute_vorticity.F"
#undef SPV_UV

#define NORMALIZE
#include "compute_vorticity.F"
#define MASK_RHO
#include "compute_vorticity.F"
#undef MASK_RHO
#define SPV_UV
#include "compute_vorticity.F"
#undef SPV_UV
#undef NORMALIZE

#define NON_STAGG
#include "compute_vorticity.F"
#define MASK_RHO
#include "compute_vorticity.F"
#undef MASK_RHO
#define SPV_UV
#include "compute_vorticity.F"
#undef SPV_UV

#define NORMALIZE
#include "compute_vorticity.F"
#define MASK_RHO
#include "compute_vorticity.F"
#undef MASK_RHO
#define SPV_UV
#include "compute_vorticity.F"
